/+
 ' Description     : Diva (Dynamically Interconnected Variable Arguments)
 ' D Class         : Diva.d v1.0
 '                 : Allows the ability to "get" and "set" data, plus alter 
 '                 : D's variadic arguments function parameter layout with 
 '                 : methods like "prepend, append, insert, splice, and remove." 
 '                 : 
 '                 : Class implicitly converts stored values of non-exact types 
 '                 : whenever possible, allowing for a polymorphic-like value 
 '                 : passing to variables, otherwise it returns a default value,
 '                 : avoiding errors if all possible. The Diva Class is more focused
 '                 : toward flexablity (using automatic value conversion) than on 
 '                 : opitmal (speed) peformance (operating on only exact/near-exact 
 '                 : types i.e. std.boxer).
 '                 : 
 ' Author          : David L. 'SpottedTiger' Davis
 ' Source Platform : Intel x86 / WinXP SP2 
 ' Target Platform : Should work on any Intel x86 based OS, but not tested.
 '                 :
 ' Created Date    : 02.Aug.05 Initial base compiled with dmd v0.128.
 ' Modified Date   : 07.Aug.05 Tons of changes and compiled with dmd v0.129.
 '                 : 17.Aug.05 Tons of more changes, and solve my problem 
 '                 :           with using "set" template with numerics and 
 '                 :           strings, plus added the isNumeric.d module 
 '                 :           to the project.
 '                 : 23.Aug.05 Added a version of copy(), remove(), and
 '                 :           splice() that uses typeid() as the parameter.
 '                 : 24.Aug.05 Added the counttype() function.      
 '                 : 25.Aug.05 Added a copyall(), removeall(), and
 '                 :           spliceall() which uses one of more typeid()s 
 '                 :           as the parameter to gather multi-typeid()s at
 '                 :           once in one function. 
 '                 : 01.Sep.05 Added setFromComplex() to handle all Complex to 
 '                 :           Complex, Complex.re to Floats, and Complex.im to 
 '                 :           Imaginary Floats convertions.
 '                 : 05.Sep.05 Improved error trapping within many of the functions.
 '                 : 06.Sep.05 Compiled and tested with dmd v0.130.
 '                 : 10.Sep.05 Added inRange.d to handle datatype range compares,
 '                 :           adjusted some of the code, also code was compiled 
 '                 :           and tested with dmd v0.131.
 '                 : 24.Oct.05 Compiled and tested with dmd v0.137, plus now that 
 '                 :           the implicit conversions between imaginary floats 
 '                 :           and regular floats has been disallowed...I've had 
 '                 :           to add even more templates to correct the code.
 '                 : 26.Oct.05 Added more code to handle pulling imaginary floats 
 '                 :           from complex floats.
 '                 : 30.Dec.05 Compiled and tested with dmd v0.142.
 '                 : 23.Mar.06 Compiled and tested with dmd v0.150.
 '                 : 28.May.06 Recompiled and Tested with dmd v0.159.
 '                 : 19.Jul.06 All "Imports" are now private by default (which 
 '                 :           shouldn't  be a problem, since I always use the 
 '                 :           private keyword) in dmd v0.163, recompiled and tested.
 '                 : 26.Aug.06 Minor correction done, recompiled and tested
 '                 :           dmd v0.165. Also had to make a bug fix in the 
 '                 :           copy(TypeInfo, uint = 0U, uint = uint.max) function,
 '                 :           and in the set!(char[]) template finction.
 '                 : 08.Sep.06 Compiled and tested with dmd v0.166. Plus added
 '                 :           2 more files for unittesting in diva_unittest1.d
 '                 :           and diva_unittest2.d.
 '                 : 23.Oct.06 Recompiled and Tested with dmd v0.172.
 '                 : 05.Nov.06 Recompiled and Tested with dmd v0.173.
 '                 : 09.Dec.06 Recompiled and Tested with dmd v0.177.
 '                 : 05.Jan.07 Recompiled and Tested with dmd v1.0.
 '                 : 09.Feb.07 Recompiled and Tested with dmd v1.005 (a major D update).
 '                 : 16.Feb.07 Recompiled and Tested with dmd v1.006 (a major D update).
 '                 : 27.Apr.07 Recompiled and Tested with dmd v1.014 (a major D update).
 '                 : 03.Jan.08 Recompiled and Tested with dmd v1.025. 
 '                 : 20.Mar.08 Recompiled and Tested with dmd v1.028.  
 '                 : 20.Jun.08 Recompiled and Tested with dmd v1.031. 
 '                 : 17.Jul.08 Recompiled and Tested with dmd v1.033.  
 '                 : 13.May.09 Recompiled and Tested with dmd v1.045.
 '                 :           
 ' Requirements    : std.stdarg, std.stdio, std.utf, std.string, std.conv,
 '                 : plus both inNumRange.d and isNumeric.d which are included.
 '                 :
 ' Licence         : Public Domain 
 ' ---------------------------------------------------------------------------
 ' 
 ' To complie for a unittest:
 ' dmd diva.d isnumeric.d innumrange.d -debug=diva -unittest
 '
 ' To compile a D program that uses diva.d:
 ' dmd MySource.d diva.d isnumeric.d innumrange.d
 +/
module diva;

//debug = diva;

debug( diva ) private import scomp = std.compiler;
private import io = std.stdio;
private import sa = std.stdarg; // for argument gathering
private import ss = std.string;
private import su = std.utf;
private import sc = std.conv;
private import sm = std.math;
/+
 ' DLD 06.Sep.05
 ' -----------------------------------------------------
 ' Once std.string is updated with the newer version of  
 ' isNumeric(), I'll remove this import as a requirement.
 +/
private import isnumeric;
private import innumrange;

extern (C) 
{
    void *memmove(void*, void*, uint);
    void *malloc(uint);
    void *realloc(void *, uint);
    void free(void *);

    // string, char, length of string
    void *memchr (void *, int, size_t); 
}

enum : ubyte 
{ 
    t_null       = 0b00000000,
    t_integer    = 0b00000001,
    t_uinteger   = 0b00000010, 
    t_longint    = 0b00000100,
    t_ulongint   = 0b00001000,
    t_float      = 0b00010000,  
    t_imaginary  = 0b00100000,
    t_complex    = 0b01000000, 
    t_string     = 0b10000000,
    t_unknown    = 0b11111111  
}

public const uint SIZEONVALUES = 0x0;

class Diva
{
    TypeInfo[] ArgType;
    void[]     ArgValueBuf; // GC controlled memory
    sa.va_list ArgValue;
    bool       bInit = false;

    enum
    {
        ISINTEGER   = 0,
        ISNEGATIVE  = 1,
        ISIMAGINARY = 2,
        ISCOMPLEX   = 3
    }
    
    this() { dedicate(512U); }

    this(in uint uiInitSize, ...) 
    { 
        // When no arguments are passed in, then use the 
        // uiInitSize to set the size of the diva data buffer.
        if (_arguments.length <= 0)
            dedicate(uiInitSize);
            
        // With one or more arguments, and if uiInitSize
        // is larger than the total arguments data size 
        // needed...then use it, otherwise use the size of  
        // the total size of the arguments being passed in.
        else if (_arguments.length > 0)
        {    
            uint uiSize = getoffset(_arguments.length, _arguments);
            if (uiInitSize >= uiSize)
                dedicate(uiInitSize);
            else
                dedicate(uiSize);

            append(_arguments, _argptr); 
        }    
    }

    ~this() { free(); bInit = false; }
 
    // Allots a locally GC controlled memory buffer
    private void dedicate(in uint uiInitSize = 512U)
    {
        if (bInit == false)
        {
            ArgValueBuf.length = uiInitSize;
            ArgValue = ArgValueBuf.ptr;
            bInit = true;
        }   
    }
    
    // Free locally controlled memory buffer
    public bool free()
    {
        if (ArgValueBuf.length > 0)
        {
            ArgType.length = 0;
            ArgValueBuf.length = 0;
            ArgValue = null;
            bInit = false;
            return true;
        }    
        
        return false;
    }   

    // If reserve increases the available space, then adjust to the new size,
    // else if it decreases available space to less than than space utilized
    // ...then set the reverse space to equal the utilized space size. 
    public bool reserve(in uint uiSize = 64U)
    {
        uint uiDataSize   = getoffset(ArgType.length);
        uint uiArgBufSize = ArgValueBuf.length;

        if (uiArgBufSize == 0)
        {
            dedicate(uiSize);       
            return true;
        }
        else if (uiSize > uiArgBufSize) 
        {
            ArgValueBuf.length = uiSize; 
            return true;
        }
        else if (uiSize < uiArgBufSize && uiSize <= uiDataSize)
        {
            ArgValueBuf.length = uiDataSize;
            return true;
        }
        else if (uiSize < uiArgBufSize && uiSize > uiDataSize)
        {
            ArgValueBuf.length = uiSize; 
            return true;
        }
        else
            return false;
    }

    // If the capacity size is greater the space utilized, 
    // squeeze the space size down to the utilized space size.
    public bool squeeze()
    {
        uint uiDataSize = getoffset(ArgType.length);

        if (ArgValueBuf.length > uiDataSize) 
        {
            ArgValueBuf.length = uiDataSize; 
            return true;
        }
        else
            return false;
    }

    // Return the current bytes used in the buffer
    public uint getUtilized() { return getoffset(ArgType.length); }

    // Returns the current max number of bytes the buffer is set for.
    public uint getCapacity() { return cast(uint)ArgValueBuf.length; }

    // Returns the current total number of arguments stored in the buffer
    public uint getNumberArgs() { return cast(uint)ArgType.length; }

    public TypeInfo[] getArgType()  { return ArgType;  }
    public sa.va_list getArgValue() { return ArgValue; }

    // Can get a total count for multiple typeid()s passed in. 
    public uint counttype(in TypeInfo[] ti_type ...)
    {
        if (ArgType.length == 0U)
            return 0U;
 
        uint ui = 0U;

        for (uint uiOuter = 0; uiOuter < ArgType.length; uiOuter++)
            for(uint uiInner = 0; uiInner < ti_type.length; uiInner++)
                if (ArgType[uiOuter] is ti_type[uiInner])
                    ui++; 

        return ui;
    }

    // Find a type from the starting position,
    // return a whole number if the type is found
    // else it returns a -1 for not found.
    public int findtype(in TypeInfo ti_type, in uint uiStartPos = 0U)
    {
        if (uiStartPos >= ArgType.length)
            return -1;
 
        for (int i = uiStartPos; i < ArgType.length; i++)
            if (ArgType[i] is ti_type)
                return i; 

        return -1;
    } 

    // Returns the offset bytes to the type's value
    private uint getoffset(in uint uiToPos)
    {
        if (uiToPos > ArgType.length || ArgType.length == 0)
            return 0U;

         return getoffset(uiToPos, ArgType);
    }

    // Returns the offset bytes to the type's value
    private uint getoffset(in uint uiToPos, in TypeInfo[] Args)
    {
        uint uiPosBytes = 0U;
    
        if (uiToPos > Args.length || Args.length == 0)
            return 0U;
 
        for (int i = 0; i < uiToPos; i++)
            uiPosBytes += ((Args[i].tsize  + int.sizeof - 1) & ~(int.sizeof - 1));   

        return uiPosBytes;
    } 

    // Returns the type of the value. 
    public TypeInfo gettype(...)
    {
        // Same as typeid(typeof(value)), but then gettype() would 
        // then need to be a template like: gettype!(T)(value)
        // which is clearly too much work for some simple 
        // information.
        TypeInfo ti = _arguments[0];
        _arguments.length = 0;
        _argptr = null;
        return ti;
    }

    // Returns the typeid at position passed in. (note 0 is the first position)
    public TypeInfo gettypeid(in uint uiPos)
    {
        if (uiPos >= ArgType.length || ArgType.length == 0)
            return null;

        //io.writefln("uiPos=%d, ArgType.length=%d", uiPos, ArgType.length);
        return ArgType[uiPos];
    }

    // Handles setting a char[] value into a numeric argument.
    template setToNumeric(T : char[])
    {
        private bool setToNumeric(in uint uiPos, in T s)
        {
            if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            TypeInfo   ti = ArgType[uiPos];
            sa.va_list vl = ArgValue + getoffset(uiPos);;
            bool[]     bflags;

            if (isnumeric.isNumeric(s, bflags))
            {
                // Whole Positive Integer
                if (bflags[ISINTEGER] == true && bflags[ISNEGATIVE] == false)
                {
                    if (ti is typeid(ulong))
                    {
                        ulong v = sc.toUlong(s);
                        memmove(&vl[0], &v, 8); 
                        return true;
                    }
                    else if (ti is typeid(bit) || ti is typeid(bool))
                    {
                        if (s == "0"c || s == "1"c)
                        {
                            bool v = (s == "1"c ? true : false);
                            memmove(&vl[0], &v, 4); 
                            return true;
                        }
                        else if ((ss.tolower(s) == "true"c ||
                                  ss.tolower(s) == "false"c) &&
                                  s.length <= 5)
                        {
                            bool v = (ss.tolower(s) == "true"c ? true : false);
                            memmove(&vl[0], &v, 4); 
                            return true;
                        }
                    }             
                    else if (ti is typeid(ubyte))
                    {
                        ubyte v = sc.toUbyte(s);
                        memmove(&vl[0], &v, 4); 
                        return true;
                    }
                    else if (ti is typeid(ushort))
                    {
                        ushort v = sc.toUshort(s);
                        memmove(&vl[0], &v, 4); 
                        return true;
                    }
                    else if (ti is typeid(uint))
                    {
                        uint v = sc.toUint(s);
                        memmove(&vl[0], &v, 4); 
                        return true;
                    }
                }
                // Whole Negative Integer
                else if (bflags[ISINTEGER] == true && bflags[ISNEGATIVE] == true)
                {
                    if (ti is typeid(long))
                    {
                        long v = sc.toLong(s);
                        memmove(&vl[0], &v, 8); 
                        return true;
                    }
                    else if (ti is typeid(byte))
                    {
                        byte v = sc.toByte(s);
                        memmove(&vl[0], &v, 4); 
                        return true;
                    }
                    else if (ti is typeid(short))
                    {
                        short v = sc.toShort(s);
                        memmove(&vl[0], &v, 4); 
                        return true;
                    }
                    else if (ti is typeid(int))
                    {
                        int v = sc.toInt(s);
                        memmove(&vl[0], &v, 4); 
                        return true;
                    }
                }
                // Imaginary Floats
                else if (bflags[ISIMAGINARY] == true && bflags[ISCOMPLEX] == false)
                {
                    if (ti is typeid(ifloat))
                    {
                        ifloat v = cast(ifloat)(sc.toFloat(s) * 1i);
                        memmove(&vl[0], &v, 4); 
                        return true;
                    }
                    else if (ti is typeid(idouble))
                    {
                        idouble v = cast(idouble)(sc.toDouble(s) * 1i);
                        memmove(&vl[0], &v, 8); 
                        return true;
                    }
                    else if (ti is typeid(ireal))
                    {
                        ireal v = cast(ireal)(sc.toReal(s) * 1i);
                        memmove(&vl[0], &v, (typeid(ireal).tsize + int.sizeof - 1) & ~(int.sizeof - 1)); 
                        return true;
                    }
                }
                // Complex Floats (currently limited)
                else if (bflags[ISCOMPLEX] == true)
                {
                    char[] s1;
                    char[] s2;
                    real   r1;
                    real   r2;
                    creal  cr;
                    
                    .splitComplexStrings(s, s1, s2);
                    r1 = sc.toReal(s1);

                    if (s2[s2.length - 1..s2.length] == "i"c || 
                        s2[s2.length - 1..s2.length] == "I"c)
                        r2 = sc.toReal(s2[0..s2.length - 1]);
                    else
                        r2 = sc.toReal(s2);

                    cr = cast(creal)(r1 + (r2 * 1.0i));
                 
                    if (ti is typeid(cfloat))
                    {
                        cfloat v = cast(cfloat)(r1 + (r2 * 1.0i));
                        if ( v.re == cr.re && v.im == cr.im )
                        {
                            memmove(&vl[0], &v, 8); 
                            return true;
                        }
                    }
                    else if (ti is typeid(cdouble))
                    {
                        cdouble v = cast(cdouble)(r1 + (r2 * 1.0i));
                        if ( v.re == cr.re && v.im == cr.im )
                        {
                            memmove(&vl[0], &v, 16); 
                            return true;
                        }
                    }
                    else if (ti is typeid(creal))
                    {
                        creal v = cast(creal)(r1 + (r2 * 1.0i));
                        memmove(&vl[0], &v, (typeid(creal).tsize + int.sizeof - 1) & ~(int.sizeof - 1));
                        return true;
                    }      

                    return false;
                }
                // Normal Floats
                else
                {
                    if (ti is typeid(float))
                    {
                        float v = sc.toFloat(s);
                        memmove(&vl[0], &v, 4); 
                        return true;
                    }
                    else if (ti is typeid(double))
                    {
                        double v = sc.toDouble(s);
                        memmove(&vl[0], &v, 8); 
                        return true;
                    }
                    else if (ti is typeid(real))
                    {
                        real v = sc.toReal(s);
                        memmove(&vl[0], &v, (typeid(real).tsize + int.sizeof - 1) & ~(int.sizeof - 1)); 
                        return true;
                    }
                }
            }

            return false; 
        }
    }

    template set(T : char[])
    {
        public bool set(in uint uiPos, in T s)
        {
            if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            sa.va_list vl = ArgValue + getoffset(uiPos);

            if (ArgType[uiPos] is typeid(char[]))
            {
                //char[] sv = sa.sa.va_arg!(char[])(vl);
                //io.writefln("Before: s.ptr=0x%X, buffer.ptr=0x%X, sv=%s", s.ptr, sv.ptr, sv);
                //vl = ArgValue + getoffset(uiPos);
                memmove(&vl[0], &s, 8); 
                //vl = ArgValue + getoffset(uiPos);
                //sv = sa.sa.va_arg!(char[])(vl);
                //io.writefln("After: s.ptr=0x%X, buffer.ptr=0x%X, sv=%s", s.ptr, sv.ptr, sv);
                return true;
            }
            else if (ArgType[uiPos] is typeid(wchar[]))
            {
                wchar[] ws = su.toUTF16(s);
                memmove(&vl[0], &ws, 8); 
                return true;
            }
            else if (ArgType[uiPos] is typeid(dchar[]))
            {
                dchar[] ds = su.toUTF32(s);
                memmove(&vl[0], &ds, 8); 
                return true;
            }
            else if (ArgType[uiPos] is typeid(void[]))
            {
                // uses the char[] s above
                memmove(&vl[0], &s, 8); 
                return true;
            }
            else
                return setToNumeric!(char[])(uiPos, s);

            return false;
        }
    }

    template set(T : wchar[])
    {
        public bool set(in uint uiPos, in T ws)
        {
            if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            sa.va_list vl = ArgValue + getoffset(uiPos);

            if (ArgType[uiPos] is typeid(wchar[]))
            {
                memmove(&vl[0], &ws, 8); 
                return true;
            }
            else if (ArgType[uiPos] is typeid(char[]))
            {
                char[] s = su.toUTF8(ws);
                memmove(&vl[0], &s, 8); 
                return true;
            }
            else if (ArgType[uiPos] is typeid(dchar[]))
            {
                dchar[] ds = su.toUTF32(ws);
                memmove(&vl[0], &ds, 8); 
                return true;
            }
            else if (ArgType[uiPos] is typeid(void[]))
            {
                char[] s = su.toUTF8(ws);
                memmove(&vl[0], &s, 8); 
                return true;
            }
            else
            {
                char[] s = su.toUTF8(ws); 
                return setToNumeric!(char[])(uiPos, s);
            }

            return false;
        }
    }

    template set(T : dchar[])
    {
        public bool set(in uint uiPos, in T ds)
        {
            if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            sa.va_list vl = ArgValue + getoffset(uiPos);

            if (ArgType[uiPos] is typeid(dchar[]))
            {
                memmove(&vl[0], &ds, 8); 
                return true;
            }
            else if (ArgType[uiPos] is typeid(char[]))
            {
                char[] s = su.toUTF8(ds);
                memmove(&vl[0], &s, 8); 
                return true;
            }
            else if (ArgType[uiPos] is typeid(wchar[]))
            {
                wchar[] ws = su.toUTF16(ds);
                memmove(&vl[0], &ws, 8); 
                return true;
            }
            else if (ArgType[uiPos] is typeid(void[]))
            {
                char[] s = su.toUTF8(ds);
                memmove(&vl[0], &s, 8); 
                return true;
            }
            else
            {
                char[] s = su.toUTF8(ds); 
                return setToNumeric!(char[])(uiPos, s);
            }

            return false;
        }
    }

    template set(T : void[])
    {
        public bool set(in uint uiPos, in T va)
        {
            if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            sa.va_list vl = ArgValue + getoffset(uiPos);

            if (ArgType[uiPos] is typeid(dchar[]))
            {
                memmove(&vl[0], &va, 8); 
                return true;
            }
            else if (ArgType[uiPos] is typeid(char[]))
            {
                char[] s = cast(char[])va;
                memmove(&vl[0], &s, 8); 
                return true;
            }
            else if (ArgType[uiPos] is typeid(wchar[]))
            {
                wchar[] ws = su.toUTF16(cast(char[])va);
                memmove(&vl[0], &ws, 8); 
                return true;
            }
            else if (ArgType[uiPos] is typeid(dchar[]))
            {
                dchar[] ds = su.toUTF32(cast(char[])va);
                memmove(&vl[0], &ds, 8); 
                return true;
            }
            else
            {
                char[] s = cast(char[])va; 
                return setToNumeric!(char[])(uiPos, s);
            }

            return false;
        }
    }

    // Handles Complex to Complex, Complex.re to Floats, and Complex.im to Imaginary Floats
    template setFromComplex(T)
    {
        bool setFromComplex(in uint uiPos, in T vValue)
        {
            if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            //io.writefln("ArgValue[%d].type=%s", uiPos, ArgType[uiPos]);
            TypeInfo   ti = ArgType[uiPos];
            sa.va_list vl = ArgValue;
            vl += getoffset(uiPos);
            
            if (ti is typeid(cfloat))
            {
                creal  cr = cast(creal)vValue;
                cfloat v  = cast(cfloat)vValue; 
                if (cr == v)
                {
                    if ((v.re >= cfloat.re.min && v.re <= cfloat.re.max) && 
                        (v.im >= cfloat.im.min && v.im <= cfloat.im.max))
                    {
                       memmove(&vl[0], &v, 8); 
                       return true;
                    }
                }
            }
            else if (ti is typeid(cdouble))
            {
                creal   cr = cast(creal)vValue;
                cdouble v  = cast(cdouble)vValue; 
                if (cr == v)
                {
                    if ((v.re >= cdouble.re.min && v.re <= cdouble.re.max) && 
                        (v.im >= cdouble.im.min && v.im <= cdouble.im.max))
                    {
                        memmove(&vl[0], &v, 16); 
                        return true;
                    }
                }
            }
            else if (ti is typeid(creal))
            {
                creal v = cast(creal)vValue;
                memmove(&vl[0], &v, (typeid(creal).tsize + int.sizeof - 1) & ~(int.sizeof - 1)); 
                return true;
            }
            else if (ti is typeid(float))
            {
                float v = cast(float)vValue.re;
                if (v == vValue.re)
                {
                    memmove(&vl[0], &v, 4);
                    return true;
                }
            }
            else if (ti is typeid(double))
            {
                double v = cast(double)vValue.re;
                if (v == vValue.re)
                {
                    memmove(&vl[0], &v, 8);
                    return true;
                }
            }
            else if (ti is typeid(real))
            {
                real v = cast(real)vValue.re;
                memmove(&vl[0], &v, (typeid(real).tsize + int.sizeof - 1) & ~(int.sizeof - 1));
                return true;
            }
            else if (ti is typeid(ifloat))
            {
                ifloat v = cast(ifloat)vValue.im;
                if (v == vValue.im)
                {
                    memmove(&vl[0], &v, 4);
                    return true;
                }
            }
            else if (ti is typeid(idouble))
            {
                idouble v = cast(idouble)vValue.im;
                if (v == vValue.im)
                {
                    memmove(&vl[0], &v, 8);
                    return true;
                }
            }
            else if (ti is typeid(ireal))
            {
                 ireal v = cast(ireal)vValue.im;
                 memmove(&vl[0], &v, (typeid(ireal).tsize + int.sizeof - 1) & ~(int.sizeof - 1));
                 return true;
            }    

            return false; 
        }
    }

    template set(T : cfloat)
    {
        public bool set(in uint uiPos, in T vValue)
        {
            if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);
            else
                return setFromComplex!(cfloat)(uiPos, vValue);
        }
    }

    template set(T : cdouble)
    {
        public bool set(in uint uiPos, in T vValue)
        {
            if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);
            else
                return setFromComplex!(cdouble)(uiPos, vValue);
        }
    }

    template set(T : creal)
    {
        public bool set(in uint uiPos, in T vValue)
        {
            if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);
            else
                return setFromComplex!(creal)(uiPos, vValue);
        }
    }
    
    // Handles Whole Positive and Negative Integers, and
    // Normal and Imaginary Floats numeric values
    template set(T)
    {
        public bool set(in uint uiPos, in T vValue)
        {
            ubyte ubytSetMethod = t_null;

            if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            //io.writefln("ArgValue[%d].type=%s", uiPos, ArgType[uiPos]);
            TypeInfo   ti = ArgType[uiPos];
            sa.va_list vl = ArgValue;
            vl += getoffset(uiPos);

            if (ti is typeid(bit) || ti is typeid(bool))  
            {
                if (isNumValueInRange(vValue, TF_BOOL))
                    ubytSetMethod = t_integer;
            }
            else if (ti is typeid(byte)) 
            {
                if (isNumValueInRange(vValue, TF_BYTE))
                    ubytSetMethod = t_integer;
            }
            else if (ti is typeid(ubyte))
            {
                if (isNumValueInRange(vValue, TF_UBYTE))
                    ubytSetMethod = t_uinteger;
            }
            else if (ti is typeid(short))
            {
                if (isNumValueInRange(vValue, TF_SHORT))
                    ubytSetMethod = t_integer;
            }
            else if (ti is typeid(ushort)) 
            {
                if (isNumValueInRange(vValue, TF_USHORT))
                    ubytSetMethod = t_uinteger;
            }
            else if (ti is typeid(int))
            {
                if (isNumValueInRange(vValue, TF_INT))
                    ubytSetMethod = t_integer;
            }
            else if (ti is typeid(uint)) 
            {
                if (isNumValueInRange(vValue, TF_UINT))
                    ubytSetMethod = t_uinteger;
            }
            else if (ti is typeid(long))
            {
                if (isNumValueInRange(vValue, TF_LONG))
                {
                    long v = cast(long)vValue;
                    memmove(&vl[0], &v, 8); 
                    return true;
                }
            }
            else if (ti is typeid(ulong)) 
            {
                if (isNumValueInRange(vValue, TF_ULONG))
                {
                    ulong v = cast(ulong)vValue;
                    memmove(&vl[0], &v, 8); 
                    return true;
                }
            }
            else if (ti is typeid(char))
            {
                if (isNumValueInRange(vValue, TF_CHAR))
                    ubytSetMethod = t_uinteger; //t_integer;
            }
            else if (ti is typeid(wchar))
            {
                if (isNumValueInRange(vValue, TF_WCHAR))
                    ubytSetMethod = t_uinteger; //t_integer;
            }
            else if (ti is typeid(dchar))
            {
                if (isNumValueInRange(vValue, TF_DCHAR))
                    ubytSetMethod = t_uinteger; //t_integer;
            }
            else if (ti is typeid(float))
            {
                //if (vValue >= float.min && vValue <= float.max)
                if (isNumValueInRange(vValue, TF_FLOAT))
                {
                    float v = cast(float)vValue; 
                    memmove(&vl[0], &v, 4); 
                    return true;
                }
            }
            else if (ti is typeid(double))
            {
                //if (vValue >= double.min && vValue <= double.max)
                if (isNumValueInRange(vValue, TF_DOUBLE))
                {
                    double v = cast(double)vValue; 
                    memmove(&vl[0], &v, 8); 
                    return true;
                }
            }
            else if (ti is typeid(real))
            {
                //if (vValue >= real.min && vValue <= real.max)
                if (isNumValueInRange(vValue, TF_REAL))
                {
                    real v = cast(real)vValue;
                    
                    memmove(&vl[0], &v, (typeid(real).tsize + int.sizeof - 1) & ~(int.sizeof - 1)); 
                    return true;
                }
            }
            else if (ti is typeid(ifloat))
            {
                //if (vValue >= ifloat.min && vValue <= ifloat.max)
                if (isNumValueInRange(vValue, TF_IFLOAT))
                {
                    ifloat v = cast(ifloat)vValue; 
                    memmove(&vl[0], &v, 4); 
                    return true;
                }
            }
            else if (ti is typeid(idouble))
            {
                //if (vValue >= idouble.min && vValue <= idouble.max)
                if (isNumValueInRange(vValue, TF_IDOUBLE))
                {
                    idouble v = cast(idouble)vValue; 
                    memmove(&vl[0], &v, 8); 
                    return true;
                }
            }
            else if (ti is typeid(ireal))
            {
                //if (vValue >= ireal.min && vValue <= ireal.max)
                if (isNumValueInRange(vValue, TF_IREAL))
                {
                    ireal v = cast(ireal)vValue;
                    memmove(&vl[0], &v, (typeid(ireal).tsize + int.sizeof - 1) & ~(int.sizeof - 1));
                    return true;
                }
            }

            switch(ubytSetMethod)
            {
              // Handles bool, byte, short, int 
              case t_integer:
              {
                  int v = cast(int)vValue;
                  memmove(&vl[0], &v, 4); 
                  return true;
              }
              // Handles ubyte, char, ushort, 
              // uint, wchar, dchar
              case t_uinteger:
              {
                  uint v = cast(uint)vValue;
                  memmove(&vl[0], &v, 4); 
                  return true;
              }
              default: break;
            }

            return false;
        }  
    }

    template get(T : char[])
    {
        public T get(in uint uiPos, in T vDefault = ""c)
        {
            if (ArgType.length == 0)
                EmptyObjectError("request"c);
            else if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);
                
            TypeInfo   ti = ArgType[uiPos];
            sa.va_list vl = ArgValue + getoffset(uiPos);

            if (ti is typeid(char[]))
                return sa.va_arg!(T)(vl);
            else if (ti is typeid(wchar[]))
                return su.toUTF8(sa.va_arg!(wchar[])(vl));
            else if (ti is typeid(dchar[]))
                return su.toUTF8(sa.va_arg!(dchar[])(vl));
            else if (ti is typeid(void[]))
                return cast(char[])sa.va_arg!(void[])(vl);
            else if (ti is typeid(char))
            {
                char[] s;
                s.length = 1;
                s[0..1]  = sa.va_arg!(char)(vl);
                return s;
            }
            else if (ti is typeid(wchar))
            {
                wchar[] ws;
                ws.length = 1;
                ws[0..1]  = sa.va_arg!(wchar)(vl);
                return su.toUTF8(ws);
            }
            else if (ti is typeid(dchar))
            {
                dchar[] ds;
                ds.length = 1;
                ds[0..1]  = sa.va_arg!(dchar)(vl);
                return su.toUTF8(ds);
            }
            else if (ti is typeid(bool)) 
                return (sa.va_arg!(bool)(vl) == true ? "true"c : "false"c );
            else if (ti is typeid(byte)) 
                return ss.toString(sa.va_arg!(byte)(vl));
            else if (ti is typeid(ubyte))
                return ss.toString(sa.va_arg!(ubyte)(vl));
            else if (ti is typeid(short))
                return ss.toString(sa.va_arg!(short)(vl));
            else if (ti is typeid(ushort))
                return ss.toString(sa.va_arg!(ushort)(vl));
            else if (ti is typeid(int))
                return ss.toString(sa.va_arg!(int)(vl));
            else if (ti is typeid(uint))
                return ss.toString(sa.va_arg!(uint)(vl));
            else if (ti is typeid(long))
                return ss.toString(sa.va_arg!(long)(vl));
            else if (ti is typeid(ulong))
                return ss.toString(sa.va_arg!(ulong)(vl));
            else if (ti is typeid(float))
                return ss.toString(sa.va_arg!(float)(vl));
            else if (ti is typeid(double))
                return ss.toString(sa.va_arg!(double)(vl));
            else if (ti is typeid(real))
                return ss.toString(sa.va_arg!(real)(vl));
            else if (ti is typeid(ifloat))
                return ss.toString(sa.va_arg!(ifloat)(vl));
            else if (ti is typeid(idouble))
                return ss.toString(sa.va_arg!(idouble)(vl));
            else if (ti is typeid(ireal))
                return ss.toString(sa.va_arg!(ireal)(vl));
            else if (ti is typeid(cfloat))
                return ss.toString(sa.va_arg!(cfloat)(vl));
            else if (ti is typeid(cdouble))
                return ss.toString(sa.va_arg!(cdouble)(vl));
            else if (ti is typeid(creal))
                return ss.toString(sa.va_arg!(creal)(vl));
            else if (ti is typeid(Object))
                return sa.va_arg!(Object)(vl).toString();
            else
                return vDefault;
        }
    }

    template get(T : wchar[])
    {
        public T get(in uint uiPos, in T vDefault = ""w)
        {
            if (ArgType.length == 0)
                EmptyObjectError("request"c);
            else if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            TypeInfo   ti = ArgType[uiPos];
            sa.va_list vl = ArgValue + getoffset(uiPos);
            char[]     s;

            if (ti is typeid(wchar[]))
                return sa.va_arg!(wchar[])(vl);
            else if (ti is typeid(char[]))
                return su.toUTF16(sa.va_arg!(char[])(vl));
            else if (ti is typeid(dchar[]))
                return su.toUTF16(sa.va_arg!(dchar[])(vl));
            else if (ti is typeid(void[]))
                s = cast(char[])sa.va_arg!(void[])(vl);
            else if (ti is typeid(char))
            {
                s.length = 1;
                s[0..1]  = sa.va_arg!(char)(vl);
                return su.toUTF16(s);
            }
            else if (ti is typeid(wchar))
            {
                wchar[] ws;
                ws.length = 1;
                ws[0..1]  = sa.va_arg!(wchar)(vl);
                return ws;
            }
            else if (ti is typeid(dchar))
            {
                dchar[] ds;
                ds.length = 1;
                ds[0..1]  = sa.va_arg!(dchar)(vl);
                return su.toUTF16(ds);
            }
            else if (ti is typeid(bool)) 
                return (sa.va_arg!(bool)(vl) == true ? "true"w : "false"w );
            else if (ti is typeid(byte)) 
                s = ss.toString(sa.va_arg!(byte)(vl));
            else if (ti is typeid(ubyte))
                s = ss.toString(sa.va_arg!(ubyte)(vl));
            else if (ti is typeid(short))
                s = ss.toString(sa.va_arg!(short)(vl));
            else if (ti is typeid(ushort))
                s = ss.toString(sa.va_arg!(ushort)(vl));
            else if (ti is typeid(int))
                s = ss.toString(sa.va_arg!(int)(vl));
            else if (ti is typeid(uint))
                s = ss.toString(sa.va_arg!(uint)(vl));
            else if (ti is typeid(long))
                s = ss.toString(sa.va_arg!(long)(vl));
            else if (ti is typeid(ulong))
                s = ss.toString(sa.va_arg!(ulong)(vl));
            else if (ti is typeid(float))
                s = ss.toString(sa.va_arg!(float)(vl));
            else if (ti is typeid(double))
                s = ss.toString(sa.va_arg!(double)(vl));
            else if (ti is typeid(real))
                s = ss.toString(sa.va_arg!(real)(vl));
            else if (ti is typeid(ifloat))
                s = ss.toString(sa.va_arg!(ifloat)(vl));
            else if (ti is typeid(idouble))
                s = ss.toString(sa.va_arg!(idouble)(vl));
            else if (ti is typeid(ireal))
                s = ss.toString(sa.va_arg!(ireal)(vl));
            else if (ti is typeid(cfloat))
                s = ss.toString(sa.va_arg!(cfloat)(vl));
            else if (ti is typeid(cdouble))
                s = ss.toString(sa.va_arg!(cdouble)(vl));
            else if (ti is typeid(creal))
                s = ss.toString(sa.va_arg!(creal)(vl));
            else if (ti is typeid(Object))
                s = sa.va_arg!(Object)(vl).toString();
            else
            {
                // Works again...need to test. --This should've worked, but it's giving me too much grief!
                //s = get!(char[])(uiPos, su.toUTF8(cast(char[])s));
                //return s;  
                return vDefault;
            }

            return su.toUTF16(s);
        }
    }

    template get(T : dchar[])
    {
        public T get(in uint uiPos, in T vDefault = ""d)
        {
            if (ArgType.length == 0)
                EmptyObjectError("request"c);
            else if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            TypeInfo   ti = ArgType[uiPos];
            sa.va_list vl = ArgValue + getoffset(uiPos);
            char[]     s;

            if (ti is typeid(dchar[]))
                return sa.va_arg!(dchar[])(vl);
            else if (ti is typeid(char[]))
                return su.toUTF32(sa.va_arg!(char[])(vl));
            else if (ti is typeid(wchar[]))
                return su.toUTF32(sa.va_arg!(wchar[])(vl));
            else if (ti is typeid(void[]))
                s = cast(char[])sa.va_arg!(void[])(vl);
            else if (ti is typeid(char))
            {
                s.length = 1;
                s[0..1]  = sa.va_arg!(char)(vl);
                return su.toUTF32(s);
            }
            else if (ti is typeid(wchar))
            {
                wchar[] ws;
                ws.length = 1;
                ws[0..1]  = sa.va_arg!(wchar)(vl);
                return su.toUTF32(ws);
            }
            else if (ti is typeid(dchar))
            {
                dchar[] ds;
                ds.length = 1;
                ds[0..1]  = sa.va_arg!(dchar)(vl);
                return ds;
            }
            else if (ti is typeid(bool)) 
                return (sa.va_arg!(bool)(vl) == true ? "true"d : "false"d );
            else if (ti is typeid(byte)) 
                s = ss.toString(sa.va_arg!(byte)(vl));
            else if (ti is typeid(ubyte))
                s = ss.toString(sa.va_arg!(ubyte)(vl));
            else if (ti is typeid(short))
                s = ss.toString(sa.va_arg!(short)(vl));
            else if (ti is typeid(ushort))
                s = ss.toString(sa.va_arg!(ushort)(vl));
            else if (ti is typeid(int))
                s = ss.toString(sa.va_arg!(int)(vl));
            else if (ti is typeid(uint))
                s = ss.toString(sa.va_arg!(uint)(vl));
            else if (ti is typeid(long))
                s = ss.toString(sa.va_arg!(long)(vl));
            else if (ti is typeid(ulong))
                s = ss.toString(sa.va_arg!(ulong)(vl));
            else if (ti is typeid(float))
                s = ss.toString(sa.va_arg!(float)(vl));
            else if (ti is typeid(double))
                s = ss.toString(sa.va_arg!(double)(vl));
            else if (ti is typeid(real))
                s = ss.toString(sa.va_arg!(real)(vl));
            else if (ti is typeid(ifloat))
                s = ss.toString(sa.va_arg!(ifloat)(vl));
            else if (ti is typeid(idouble))
                s = ss.toString(sa.va_arg!(idouble)(vl));
            else if (ti is typeid(ireal))
                s = ss.toString(sa.va_arg!(ireal)(vl));
            else if (ti is typeid(cfloat))
                s = ss.toString(sa.va_arg!(cfloat)(vl));
            else if (ti is typeid(cdouble))
                s = ss.toString(sa.va_arg!(cdouble)(vl));
            else if (ti is typeid(creal))
                s = ss.toString(sa.va_arg!(creal)(vl));
            else if (ti is typeid(Object))
                s = sa.va_arg!(Object)(vl).toString();
            else
            {
                // This should've worked, but it's giving me too much grief!
                //s = get!(char[])(uiPos, cast(char[])su.toUTF8(vDefault));
                return vDefault;
            }

            return su.toUTF32(s);
        }
    }

    template get(T : void[])
    {
        public T get(in uint uiPos, in T vDefault = ""c)
        {
            if (ArgType.length == 0)
                EmptyObjectError("request"c);
            else if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            TypeInfo   ti = ArgType[uiPos];
            sa.va_list vl = ArgValue + getoffset(uiPos);

            if (ti is typeid(void[]))
                return sa.va_arg!(void[])(vl);
            else if (ti is typeid(char[]))
                return sa.va_arg!(T)(vl);
            else if (ti is typeid(wchar[]))
                return su.toUTF8(sa.va_arg!(wchar[])(vl));
            else if (ti is typeid(dchar[]))
                return su.toUTF8(sa.va_arg!(dchar[])(vl));
            else if (ti is typeid(char))
            {
                char[] s;
                s.length = 1;
                s[0..1]  = sa.va_arg!(char)(vl);
                return s;
            }
            else if (ti is typeid(wchar))
            {
                wchar[] ws;
                ws.length = 1;
                ws[0..1]  = sa.va_arg!(wchar)(vl);
                return su.toUTF8(ws);
            }
            else if (ti is typeid(dchar))
            {
                dchar[] ds;
                ds.length = 1;
                ds[0..1]  = sa.va_arg!(dchar)(vl);
                return su.toUTF8(ds);
            }
            else if (ti is typeid(bool)) 
                return (sa.va_arg!(bool)(vl) == true ? "true"c : "false"c );
            else if (ti is typeid(byte)) 
                return ss.toString(sa.va_arg!(byte)(vl));
            else if (ti is typeid(ubyte))
                return ss.toString(sa.va_arg!(ubyte)(vl));
            else if (ti is typeid(short))
                return ss.toString(sa.va_arg!(short)(vl));
            else if (ti is typeid(ushort))
                return ss.toString(sa.va_arg!(ushort)(vl));
            else if (ti is typeid(int))
                return ss.toString(sa.va_arg!(int)(vl));
            else if (ti is typeid(uint))
                return ss.toString(sa.va_arg!(uint)(vl));
            else if (ti is typeid(long))
                return ss.toString(sa.va_arg!(long)(vl));
            else if (ti is typeid(ulong))
                return ss.toString(sa.va_arg!(ulong)(vl));
            else if (ti is typeid(float))
                return ss.toString(sa.va_arg!(float)(vl));
            else if (ti is typeid(double))
                return ss.toString(sa.va_arg!(double)(vl));
            else if (ti is typeid(real))
                return ss.toString(sa.va_arg!(real)(vl));
            else if (ti is typeid(ifloat))
                return ss.toString(sa.va_arg!(ifloat)(vl));
            else if (ti is typeid(idouble))
                return ss.toString(sa.va_arg!(idouble)(vl));
            else if (ti is typeid(ireal))
                return ss.toString(sa.va_arg!(ireal)(vl));
            else if (ti is typeid(cfloat))
                return ss.toString(sa.va_arg!(cfloat)(vl));
            else if (ti is typeid(cdouble))
                return ss.toString(sa.va_arg!(cdouble)(vl));
            else if (ti is typeid(creal))
                return ss.toString(sa.va_arg!(creal)(vl));
            else if (ti is typeid(Object))
                return sa.va_arg!(Object)(vl).toString();
            else
            {
                // This should've worked, but it's giving me too much grief!
                //return get!(char[])(uiPos, vDefault);
                return vDefault;
            }
        }        
    }

    template get(T : cfloat)
    {
        public T get(in uint uiPos, in T vDefault = 0.0e0f+0.0fi)
        {
            if (ArgType.length == 0)
                EmptyObjectError("request"c);
            else if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            TypeInfo   ti = ArgType[uiPos];
            sa.va_list vl = ArgValue + getoffset(uiPos);

            if (ti is typeid(cfloat))
                return sa.va_arg!(cfloat)(vl);
            else if (ti is typeid(cdouble))
            {
                cdouble cd = sa.va_arg!(cdouble)(vl);
                if ((cd.re >= T.re.min && cd.re <= T.re.max) &&
                    (cd.im >= T.im.min && cd.im <= T.im.max))
                    return cast(T)cd;
            }
            else if (ti is typeid(creal))
            {
                creal cr = sa.va_arg!(creal)(vl);
                if ((cr.re >= T.re.min && cr.re <= T.re.max) &&
                    (cr.im >= T.im.min && cr.im <= T.im.max))
                    return cast(T)cr;
            }
            else 
                return vDefault;
        }
    }
    
    template get(T : cdouble)
    {
        public T get(in uint uiPos, in T vDefault = 0.0e0+0.0i)
        {
            if (ArgType.length == 0)
                EmptyObjectError("request"c);
            else if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            TypeInfo   ti = ArgType[uiPos];
            sa.va_list vl = ArgValue + getoffset(uiPos);
        
            if (ti is typeid(cfloat))
                return cast(cdouble)sa.va_arg!(cfloat)(vl);
            else if (ti is typeid(cdouble))
                return sa.va_arg!(cdouble)(vl);
            else if (ti is typeid(creal))
            {
                creal cr = sa.va_arg!(creal)(vl);
                if ((cr.re >= T.re.min && cr.re <= T.re.max) &&
                    (cr.im >= T.im.min && cr.im <= T.im.max))
                    return cast(T)cr;
            }
            else 
                return vDefault;        }
    }
    
    template get(T : creal)
    {
        public T get(in uint uiPos, in T vDefault = 0.0e0L+0.0Li)
        {
            if (ArgType.length == 0)
                EmptyObjectError("request"c);
            else if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            TypeInfo   ti = ArgType[uiPos];
            sa.va_list vl = ArgValue + getoffset(uiPos);

            if (ti is typeid(cfloat))
                return cast(creal)sa.va_arg!(cfloat)(vl);
            else if (ti is typeid(cdouble))
                return cast(creal)sa.va_arg!(cdouble)(vl);
            else if (ti is typeid(creal))
                return sa.va_arg!(creal)(vl);
            else 
                return vDefault;
        }
    }
 
    template get(T : ifloat)
    {
        public T get(in uint uiPos, in T vDefault = 0.0fi)
        {
            if (ArgType.length == 0)
                EmptyObjectError("request"c);
            else if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            TypeInfo   ti = ArgType[uiPos];
            sa.va_list vl = ArgValue + getoffset(uiPos);

            if (ti is typeid(ifloat))
                return sa.va_arg!(ifloat)(vl);
            else if (ti is typeid(idouble))
            {
                idouble v = sa.va_arg!(idouble)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(ireal))
            {
                ireal v = sa.va_arg!(ireal)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(cfloat))
            {   
                cfloat v = sa.va_arg!(cfloat)(vl);
                return cast(ifloat)(v.im * 1.0i);
            }
            else if (ti is typeid(cdouble))
            {   
                cdouble v = sa.va_arg!(cdouble)(vl);
                if (isNumValueInRange(cast(idouble)v.im, vDefault))
                    return cast(ifloat)(v.im * 1.0i);
                else
                    return vDefault; 
            }
            else if (ti is typeid(creal))
            {   
                creal v = sa.va_arg!(creal)(vl);
                //io.writefln("get(ifloat) creal=", v, " creal.im=", cast(ifloat)v.im);
                if (isNumValueInRange(cast(ireal)v.im, vDefault))
                    return cast(ifloat)(v.im * 1.0i);
                else
                    return vDefault; 
            }
            else 
                return vDefault;        
        }
    }

    template get(T : idouble)
    {
        public T get(in uint uiPos, in T vDefault = 0.0i)
        {
            if (ArgType.length == 0)
                EmptyObjectError("request"c);
            else if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            TypeInfo   ti = ArgType[uiPos];
            sa.va_list vl = ArgValue + getoffset(uiPos);

            if (ti is typeid(ifloat))
                return cast(idouble)sa.va_arg!(ifloat)(vl);
            else if (ti is typeid(idouble))
                return sa.va_arg!(idouble)(vl);
            else if (ti is typeid(ireal))
            {
                ireal v = sa.va_arg!(ireal)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(cfloat))
            {   
                cfloat v = sa.va_arg!(cfloat)(vl);
                return cast(idouble)(v.im * 1.0i);
            }
            else if (ti is typeid(cdouble))
            {   
                cdouble v = sa.va_arg!(cdouble)(vl);
                return cast(idouble)(v.im * 1.0i);
            }
            else if (ti is typeid(creal))
            {   
                creal v = sa.va_arg!(creal)(vl);
                if (isNumValueInRange(cast(ireal)v.im, vDefault))
                    return cast(idouble)(v.im * 1.0i);
                else
                    return vDefault; 
            }
            else 
                return vDefault;        
        }        

    }
   
    template get(T : ireal)
    {
        public T get(in uint uiPos, in T vDefault = 0.0Li)
        {
            if (ArgType.length == 0)
                EmptyObjectError("request"c);
            else if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            TypeInfo   ti = ArgType[uiPos];
            sa.va_list vl = ArgValue + getoffset(uiPos);

            if (ti is typeid(ifloat))
                return cast(ireal)sa.va_arg!(ifloat)(vl);
            else if (ti is typeid(idouble))
                return cast(ireal)sa.va_arg!(idouble)(vl);
            else if (ti is typeid(ireal))
                return sa.va_arg!(ireal)(vl);
            else if (ti is typeid(cfloat))
            {   
                cfloat v = sa.va_arg!(cfloat)(vl);
                return cast(ireal)v;
            }
            else if (ti is typeid(cdouble))
            {   
                cdouble v = sa.va_arg!(cdouble)(vl);
                return cast(ireal)v;
            }
            else if (ti is typeid(creal))
            {   
                creal v = sa.va_arg!(creal)(vl);
                return cast(ireal)v;
            }
            return vDefault;
        }
    }

    template get(T)
    {
        public T get(in uint uiPos, in T vDefault = 0)
        {
            if (ArgType.length == 0)
                EmptyObjectError("request"c);
            else if (uiPos >= ArgType.length)
                OutOfBoundsError(uiPos);

            TypeInfo   ti = ArgType[uiPos];
            sa.va_list vl = ArgValue + getoffset(uiPos);
            char[]     s;
            ubyte      ubytMethod = t_unknown;
            bool[]     bflags;

            if (ti is typeid(char[]))
            {
                s = sa.va_arg!(char[])(vl);
                ubytMethod = t_string;
            }
            else if (ti is typeid(wchar[]))
            {
                s = su.toUTF8(sa.va_arg!(wchar[])(vl));
                ubytMethod = t_string;
            }
            else if (ti is typeid(dchar[]))
            {
                s = su.toUTF8(sa.va_arg!(dchar[])(vl));
                ubytMethod = t_string;
            }
            else if (ti is typeid(void[]))
            {
                s = cast(char[])sa.va_arg!(void[])(vl);
                ubytMethod = t_string;
            }
            else if (ti is typeid(bool))  
            {
                bool v = sa.va_arg!(bool)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(byte)) 
            {
                byte v = sa.va_arg!(byte)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(ubyte))
            {
                ubyte v = sa.va_arg!(ubyte)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(short))
            {
                short v = sa.va_arg!(short)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(ushort)) 
            {
                ushort v = sa.va_arg!(ushort)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(int))
            {
                int v = sa.va_arg!(int)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(uint)) 
            {
                uint v = sa.va_arg!(uint)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(long))
            {
                long v = sa.va_arg!(long)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(ulong)) 
            {
                ulong v = sa.va_arg!(ulong)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(char))
            {
                char v = sa.va_arg!(char)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(wchar))
            {
                wchar v = sa.va_arg!(wchar)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(dchar))
            {
                dchar v = sa.va_arg!(dchar)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(float))
            {
                float v = sa.va_arg!(float)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(double))
            {
                double v = sa.va_arg!(double)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(real))
            {
                real v = sa.va_arg!(real)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(ifloat))
            {
                ifloat v = sa.va_arg!(ifloat)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(idouble))
            {
                idouble v = sa.va_arg!(idouble)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(ireal))
            {
                ireal v = sa.va_arg!(ireal)(vl);
                if (isNumValueInRange(v, vDefault))
                    return cast(T)v;  
                else
                    return vDefault;
            }
            else if (ti is typeid(cfloat))
            {
                TypeInfo tio = typeid(T);
                cfloat v = sa.va_arg!(cfloat)(vl);

                if (tio is typeid(float))
                    ubytMethod = t_float; 
                else if (tio is typeid(double))
                    ubytMethod = t_float; 
                else if (tio is typeid(real))
                    ubytMethod = t_float;
                else if (tio is typeid(ifloat))
                    ubytMethod = t_imaginary;
                else if (tio is typeid(idouble))
                    ubytMethod = t_imaginary;
                else if (tio is typeid(ireal))
                    ubytMethod = t_imaginary;
                
                if ( ubytMethod == t_float)
                {
                    if (v.re >= T.min && v.re <= T.max)
                        return cast(T)v.re;  
                }
                else if (ubytMethod == t_imaginary)
                {
                    if (cast(T)(v.im * 1.0i) >= T.min && 
                        cast(T)(v.im * 1.0i) <= T.max)
                        return cast(T)(v.im * 1.0i);  
                }

                return vDefault;
            }
            else if (ti is typeid(cdouble))
            {
                TypeInfo tio = typeid(T);
                cdouble v = sa.va_arg!(cdouble)(vl);

                if (tio is typeid(float))
                    ubytMethod = t_float; 
                else if (tio is typeid(double))
                    ubytMethod = t_float; 
                else if (tio is typeid(real))
                    ubytMethod = t_float;
                else if (tio is typeid(ifloat))
                    ubytMethod = t_imaginary;
                else if (tio is typeid(idouble))
                    ubytMethod = t_imaginary;
                else if (tio is typeid(ireal))
                    ubytMethod = t_imaginary;
                
                if ( ubytMethod == t_float)
                {
                    if (v.re >= T.min && v.re <= T.max)
                        return cast(T)v.re;  
                }
                else if (ubytMethod == t_imaginary)
                {
                    if (cast(T)(v.im * 1.0i) >= T.min && 
                        cast(T)(v.im * 1.0i) <= T.max)
                        return cast(T)(v.im * 1.0i);  
                }

                return vDefault;
            }
            else if (ti is typeid(creal))
            {
                TypeInfo tio = typeid(T);
                creal v = sa.va_arg!(creal)(vl);

                if (tio is typeid(float))
                    ubytMethod = t_float; 
                else if (tio is typeid(double))
                    ubytMethod = t_float; 
                else if (tio is typeid(real))
                    ubytMethod = t_float;
                else if (tio is typeid(ifloat))
                    ubytMethod = t_imaginary;
                else if (tio is typeid(idouble))
                    ubytMethod = t_imaginary;
                else if (tio is typeid(ireal))
                    ubytMethod = t_imaginary;
                
                if ( ubytMethod == t_float)
                {
                    if (v.re >= T.min && v.re <= T.max)
                        return cast(T)v.re;  
                }
                else if (ubytMethod == t_imaginary)
                {
                    if (cast(T)(v.im * 1.0i) >= T.min && 
                        cast(T)(v.im * 1.0i) <= T.max)
                        return cast(T)(v.im * 1.0i);  
                }

                return vDefault;
            }

            if (ubytMethod == t_string)
            {
                if (isnumeric.isNumeric(s, bflags)) 
                {
                    // Whole Positive Integer
                    if (bflags[ISINTEGER] == true && bflags[ISNEGATIVE] == false)
                    {
                        ulong v = sc.toUlong(s);
                        if (isNumValueInRange(v, vDefault))
                        //if (v >= T.min && v <= T.max)
                            return cast(T)v;
                    }
                    // Whole Negative Integer
                    else if (bflags[ISINTEGER] == true && bflags[ISNEGATIVE] == true)
                    {
                        long v = sc.toLong(s);
                        if (isNumValueInRange(v, vDefault))
                        //if (v >= T.min && v <= T.max)
                            return cast(T)v;
                    }
                    // Imaginary Floats
                    else if (bflags[ISIMAGINARY] == true && bflags[ISCOMPLEX] == false)
                    {
                        ireal v = cast(ireal)(sc.toReal(s) * 1i);
                        //if (v >= T.min && v <= T.max)
                        if (isNumValueInRange(v, vDefault))
                            return cast(T)v;
                    }
                    // Complex Floats
                    else if (bflags[ISCOMPLEX] == true)
                    {
                        /+
                        char[] sre;
                        char[] sim;
                        splitComplexStrings(s, sre, sim)
                        creal v = cast(creal)(sc.toReal(sre) + sc.toReal(sim) * 1i);
                        if (v.re >= T.min && v.re <= T.max)
                            return cast(T)v;
                        +/
                        return vDefault;
                    }
                    // Normal Floats
                    else
                    {
                        real v = sc.toReal(s); 
                        //if (v >= T.min && v <= T.max)
                        if (isNumValueInRange(v, vDefault))
                            return cast(T)v;
                    }
                }
            }

            return vDefault;
        }        
    }

    // Removes the argument(s) at the given position(s) 
    public void remove(in uint uiStartPos, in uint uiEndPos = 0U)
    {
        uint uiArgLen = ArgType.length;

        if (uiArgLen == 0)
            EmptyObjectError("remove"c);
        else if (uiStartPos >= uiArgLen)
            OutOfBoundsError(uiStartPos);

        /*********************************
         * -- Handling special cases. -- *
         *********************************/
        // If start equals the ArgType.length delete the very last position.
        if (uiStartPos == uiArgLen)
        {
            ArgType.length = uiArgLen - 1U;
            return;
        }

        if (uiEndPos >= uiArgLen)
            uiEndPos = uiArgLen - 1;

        /********************************
         * -- Handling normal cases. -- *
         ********************************/
        if (uiEndPos < uiStartPos)
            uiEndPos = uiStartPos;

        sa.va_list vl1 = ArgValue; // Points to the deleted entry
        sa.va_list vl2 = ArgValue; // Points to next entry to move
        uint uiPos1    = getoffset(uiStartPos);
        uint uiPos2    = getoffset(uiEndPos + 1U);
        uint uiTotal   = getoffset(uiArgLen);
        uint uiDelCnt  = 0U;

        vl1 += uiPos1;
        vl2 += uiPos2;
        memmove(&vl1[0], &vl2[0], uiTotal - uiPos2);    

        uiDelCnt = ((uiEndPos + 1U) - (uiStartPos + 1U)) + 1U;
        for (uint ui = uiStartPos; ui < uiArgLen - uiDelCnt; ui++)
            ArgType[ui] = ArgType[ui + uiDelCnt];

        ArgType.length = uiArgLen - uiDelCnt;
    }

    // Remove all arguments of the same type, optional from a start position and a end position.
    public void remove(in TypeInfo ti, in uint uiStartPos = 0U, in uint uiEndPos = uint.max)
    {
        uint uiArgLen = ArgType.length;

        if (uiArgLen == 0)
            EmptyObjectError("remove"c);
        else if (uiStartPos >= uiArgLen)
            OutOfBoundsError(uiStartPos);

        /*********************************
         * -- Handling special cases. -- *
         *********************************/
        if (uiStartPos == uiArgLen)
            uiStartPos = uiArgLen - 1;

        if (uiEndPos == uiArgLen)
            uiEndPos = uiArgLen - 1;

        /********************************
         * -- Handling normal cases. -- *
         ********************************/
        if (uiEndPos < uiStartPos)
            uiEndPos = uiStartPos;

        uint uix = uiStartPos;

        LOOPSTART:
            if (ArgType[uix] == ti)
                remove(uix);
            else
                uix++;

            if (uix >= ArgType.length || ArgType.length == 0 || uix > uiEndPos)
                goto EXITLOOP;

        goto LOOPSTART;

        EXITLOOP:
        return;
    }
 
    // Remove all of the typeid()s passed in. 
    public void removeall(in TypeInfo[] ti_type ...)
    {
        if (ArgType.length == 0)
            EmptyObjectError("removeall"c);
 
        for (uint ui = 0U; ui < ti_type.length; ui++)
            remove(ti_type[ui]); 
    }   

    // Copy an argument from a position, or optional copy all 
    // from a start position to an end position.
    public Diva copy(in uint uiStartPos, in uint uiEndPos = 0U)
    {
        uint uiArgLen = ArgType.length;

        if (uiArgLen == 0)
            EmptyObjectError("copy"c);
        else if (uiStartPos >= uiArgLen)
            OutOfBoundsError(uiStartPos);

        sa.va_list vlsrc   = ArgValue; 
        uint       uiPos1  = 0U;
        uint       uiPos2  = 0U;
        uint       uiTotal = 0U;

        /*********************************
         * -- Handling special cases. -- *
         *********************************/
        if (uiStartPos >= uiArgLen)
            uiStartPos = uiArgLen - 1;

        if (uiEndPos >= uiArgLen)
            uiEndPos = uiArgLen - 1;

        /********************************
         * -- Handling normal cases. -- *
         ********************************/
        if (uiEndPos < uiStartPos)
            uiEndPos = uiStartPos;
        //else if (uiStartPos >= uiArgLen || uiEndPos >= uiArgLen)
        //    return null;

        uiPos1  = getoffset(uiStartPos);
        uiPos2  = getoffset(uiEndPos + 1U);
        uiTotal = ((uiPos2 - uiPos1) == 0U ? uiPos1 + ((ArgType[uiPos1].tsize + int.sizeof - 1) & ~(int.sizeof - 1))
                                           : uiPos2 - uiPos1);
                                           
        Diva oDiva = new Diva(uiTotal);
        sa.va_list vldst = oDiva.getArgValue(); 
        uint uiCnt = ((uiEndPos + 1U) - (uiStartPos + 1U)) + 1U;
        uint uij = 0U;
            
        oDiva.ArgType.length = uiCnt;
        vlsrc += uiPos1;
        memmove(&vldst[0], &vlsrc[0], uiTotal);
            
        for (uint ui = uiStartPos; ui <= uiEndPos; ui++)
        {
            oDiva.ArgType[uij] = ArgType[ui];
            uij++;
        }
        
        return oDiva;    
    }

    // Copy one or all arguments of the same type, optional from a start position 
    // and an end position.
    public Diva copy(in TypeInfo ti, in uint uiStartPos = 0U, in uint uiEndPos = uint.max)
    {
        uint uiArgLen = ArgType.length;

        if (uiArgLen == 0)
            EmptyObjectError("copy"c);
        else if (uiStartPos > uiArgLen)
            OutOfBoundsError(uiStartPos);

        /*********************************
         * -- Handling special cases. -- *
         *********************************/
         // In case getNumberArgs() was used for passing
         // in a the start position for copying the very
         // last element, just adjust it with a -1.
         if (uiStartPos == uiArgLen)
            uiStartPos = uiArgLen - 1;

        // Set the end position to the last element.
        if (uiEndPos >= uiArgLen)
            uiEndPos = uiArgLen - 1;
        /* ***************************** */

        if (uiEndPos < uiStartPos)
        {
            OtherError("The ending index of "c ~ ss.toString(uiStartPos) ~
                       " is less than the starting index which is "c ~ ss.toString(uiStartPos));

            uiEndPos = uiStartPos;
        }
         
         // Test Display
        //io.writefln("copy (1st): uiStartPos=%d, uiEndPos=%d", uiStartPos, uiEndPos);
        // ------------
  
        int iz = findtype(ti, uiStartPos);
        
        if (iz == -1)
            return null;
        
        // Test Display
        //io.writefln("copy (1st): iz=%d, ti=%s", iz, ti);
        // ------------
        
        // default size of the buffer is set to the size of
        // the source data minus the first known position
        // of the data type we're copying.
        Diva oDiva       = new Diva(getUtilized() - getoffset(iz));
        sa.va_list vlsrc = ArgValue; 
        sa.va_list vldst = oDiva.getArgValue();
        uint uiTypeSize  = 0U;
        uint uiy         = 0U;

        do
        {
            //io.writefln("Copy() iz=%d, uiy=%d, iz + 1=%d", iz, uiy, iz + 1);
            oDiva.ArgType.length = uiy + 1;
            oDiva.ArgType[uiy] = ti;
            uiTypeSize = ((oDiva.ArgType[uiy].tsize  + int.sizeof - 1) & ~(int.sizeof - 1));
            vlsrc += getoffset(iz);
            //io.writefln("vlsrc=0x%X, vldst=0x%X", vlsrc, vldst);
            memmove(&vldst[0], &vlsrc[0], uiTypeSize); 
            uiy++;
            vlsrc  = ArgValue;
            vldst += uiTypeSize; 
            
            iz = findtype(ti, iz + 1); 
            // Test Display
            //io.writefln("copy (loop): iz=%d, ti=%s", iz, ti);
            // ------------

        } while (iz != -1 && iz <= uiEndPos)
        
        oDiva.reserve(oDiva.getUtilized());
        
        return oDiva;
    }

    // Copy all of the typeid()s passed in. 
    public Diva copyall(in TypeInfo[] ti_type ...)
    {
        if (ArgType.length == 0)
            EmptyObjectError("copyall"c);

        Diva oDiva = new Diva(getUtilized());

        for(int i = 0; i < ti_type.length; i++)
            oDiva.append(copy(ti_type[i])); 

        if (oDiva !is null)
            oDiva.reserve(oDiva.getUtilized());

        return oDiva;
    }   

    // Splice an argument of a given type from a given position, 
    // or optional all arguments of the same type from a start 
    // position to an end position.
    public Diva splice(in uint uiStartPos, in uint uiEndPos = 0U)
    {
        if (ArgType.length == 0)
            EmptyObjectError("splice"c);
        else if (uiStartPos >= ArgType.length)
            OutOfBoundsError(uiStartPos);

        // First copy enties from the original to the copy
        Diva oDiva = copy(uiStartPos, uiEndPos);
        
        // Then remove those entries from the original
        if (oDiva !is null)
            remove(uiStartPos, uiEndPos);
     
        return oDiva;       
    }
    
    // Splice an argument for a given position, or optional all arguments of the same type, 
    // from a given start position to an end position.
    public Diva splice(in TypeInfo ti, in uint uiStartPos = 0U, in uint uiEndPos = uint.max)
    {
        if (ArgType.length == 0)
            EmptyObjectError("splice"c);
        else if (uiStartPos >= ArgType.length)
            OutOfBoundsError(uiStartPos);

        // First copy enties from the original to the copy
        Diva oDiva = copy(ti, uiStartPos, uiEndPos);

        // Then remove those entries from the original
        if (oDiva !is null)
            remove(ti, uiStartPos, uiEndPos);
     
        return oDiva;       
    }

    // Splice all of the typeid()s passed in. 
    public Diva spliceall(in TypeInfo[] ti_type ...)
    {
        if (ArgType.length == 0)
            EmptyObjectError("spliceall"c);
 
        Diva oDiva = new Diva(getUtilized());

        for(int i = 0; i < ti_type.length; i++)
        {   
            // Test Display
            /+
            io.writefln("spliceall: i=%d, ti_type[i]=%s", i, ti_type[i]);
            Diva oDivaCopy = new Diva;
            oDivaCopy = copyall(ti_type[i]);
            io.writefln("spliceall: Display oDivaCopy's entries");
            displayArguments(oDivaCopy);
            +/
            // ------------
            
            oDiva.append(copyall(ti_type[i])); 
        }
        
        if (oDiva !is null)
        {
            for(int i = 0; i < ti_type.length; i++)
                removeall(ti_type[i]); 

            oDiva.reserve(oDiva.getUtilized);
        }

        return oDiva;
    }   
    
    // Insert at the beginning
    public void prepend(...) 
    { 
        insert(0U, _arguments, _argptr); 
        _arguments.length = 0U;
        _argptr = null;
    }
    
    public void prepend(in TypeInfo[] _arguments, in sa.va_list _argptr) 
    { insert(0U, _arguments, _argptr); }

    public void prepend(in Diva oDiva) 
    { insert(0U, oDiva.ArgType, oDiva.ArgValue); }

    // Insert at the end
    public void append(...) 
    { 
        insert(ArgType.length, _arguments, _argptr); 
        _arguments.length = 0U;
        _argptr = null;
    }
    
    public void append(in TypeInfo[] _arguments, in sa.va_list _argptr) 
    { insert(ArgType.length, _arguments, _argptr); }

    public void append(in Diva oDiva) 
    { insert(ArgType.length, oDiva.ArgType, oDiva.ArgValue); }

    public void insert(in uint uiDstPos, ...)
    { 
        insert(uiDstPos, _arguments, _argptr); 
        _arguments.length = 0U;
        _argptr = null;
    }    

    public void insert(in uint uiDstPos, in Diva oDiva)
    { insert(uiDstPos, oDiva.ArgType, oDiva.ArgValue); }

    public void insert(in uint uiDstPos, in TypeInfo[] _arguments, in sa.va_list _argptr)
    {
        // handles using perpend() and append() to add data after a free() call,
        // plus this will set uiDstPos to zero for an insert() only in this case.
        if (ArgType.length == 0)
        {
            if (_arguments.length > 0)
            {
                // Allot the exact amount of bytes needed to store the arguments into.
                dedicate(getoffset(_arguments.length, _arguments));
                uiDstPos = 0;
            }
            else
                OtherError("Nothing to insert, no arguments were passed in."c);
        }

        // Test Display
        //io.writefln("insert: uiDstPos=%d, _arguments.length=%d", uiDstPos, _arguments.length);
        // ------------
        
        sa.va_list vl1  = ArgValue; // Points to the insert position
        sa.va_list vl2  = ArgValue; // Points at move to position
        uint uiDstPos1  = 0U;
        uint uiDstPos2  = 0U;
        uint uiDstLen   = ArgType.length;
        uint uiSrcLen   = _arguments.length;
        uint uiTotalLen = uiDstLen + uiSrcLen;
        // -----------------------------------------------------------
        uint uiDstTotalBytes = getoffset(ArgType.length);
        uint uiSrcTotalBytes = getoffset(uiSrcLen, _arguments);
        uint uiReqTotalBytes = uiDstTotalBytes + uiSrcTotalBytes;

        // In this case uiDstPos can be equal to ArgType.length
        // for appending new data to the tail end arguments, but
        // not greater than.
        if (uiDstPos > uiDstLen) 
            OutOfBoundsError(uiDstPos);
        else if (uiReqTotalBytes > ArgValueBuf.length)
            ArgValueBuf.length = ArgValueBuf.length + (uiReqTotalBytes - ArgValueBuf.length);
 
        // If appending a value or a set of values to the very end.
        if (uiDstPos == uiDstLen)
        {
            vl1 += uiDstTotalBytes;
            uiDstPos1 = ArgType.length;
            ArgType.length = uiTotalLen; 

            for (int i = 0; i < uiSrcLen; i++)
                ArgType[uiDstPos1 + i] = _arguments[i];

            memmove(&vl1[0], &_argptr[0], uiSrcTotalBytes);
        }
        // Inserting a value or set of values
        else
        { 
            void[] voBuf;

            uiDstPos1 = uiDstLen; 
            uiDstPos2 = getoffset(uiDstPos);
            voBuf.length = uiDstTotalBytes - uiDstPos2;
            vl1 += uiDstPos2;
            vl2 += uiSrcTotalBytes + uiDstPos2;
            ArgType.length = uiTotalLen; 

            memmove(&voBuf[0], &vl1[0], uiDstTotalBytes - uiDstPos2);
            memmove(&vl1[0], &_argptr[0], uiSrcTotalBytes);
            memmove(&vl2[0], &voBuf[0], voBuf.length);
            voBuf.length = 0U;

            // Move current data to its new position
            for (int i = uiTotalLen; i > uiTotalLen - (uiDstLen - uiDstPos); i--)
                ArgType[i - 1] = ArgType[i - uiSrcLen - 1];

            // Insert the data into position
            for (int i = 0; i < uiSrcLen; i++)
                ArgType[uiDstPos + i] = _arguments[i];
        }
    }

    private void EmptyObjectError(in char[] s)
    {
        throw new Exception("Can not "c ~ s ~ " data from an emtpy object."c);
    }

    private void OutOfBoundsError(in uint uiPos)
    {
        throw new Exception("Position index "c ~ ss.toString(uiPos) ~
                            " is out of bounds with the current valid indexs, which are 0 to "c ~
                            ss.toString(getNumberArgs() - 1));
    }

    private void OtherError(in char[] s)
    {
        throw new Exception(s);
    }
}

/+
 ' Splits a complex floats (cfloat, cdouble, and creal) into two workable strings.
 ' Grammar:
 ' (for cfloat, cdouble, and creal)
 ' ['+'|'-']digit(s)[.][digit(s)][[e-|e+]digit(s)][+]
 '         [digit(s)[.][digit(s)][[e-|e+]digit(s)][i|f|L|Li|fi]]
 '      or [nan|nani|nan+nani|inf|-inf]
 '
 ' creal Range: from 3.362103E-4932+3.362103E-4932i (min) 
 '                to 1.189731E+4932+1.189731E+4932i (max)
 +/
void splitComplexStrings(in char[] s, out char[] s1, out char[] s2)
{
    char[] sx   = s;
    int    iLen = s.length;
    int    plus = 0;
    
    s1 = ""c;
    s2 = ""c;
    
    if (!iLen)
        throw new Exception("spiitComplexStrings(): "c ~
                            " input string is an empty string."c);

    if (iLen >= 2)
    {
        sx = ss.tolower(s);   
        // When "nan" or "nani" just return them.
        if (sx == "nan"c || sx == "nani"c ||
            sx == "nan+nani"c || sx =="nan+0i"c)
        { s1 = "nan"c; s2 = "nani"c; return; }
    
        if (sx == "inf"c)
        { s1 = "inf"c; s2 = "infi"c; return; }    
    
        if (sx == "-inf"c)
        { s1 = "-inf"c; s2 = "-infi"c; return; }    
    }
    
    // Split the original string out into two strings.
    for (int i = 1; i < iLen; i++)
        if ((sx[i - 1] != 'e' && sx[i - 1] != 'E') && sx[i] == '+')
        {
            s1 = sx[0..i];
            if (i + 1 < iLen - 1)
                s2 = sx[i + 1..iLen]; // - 1];
            else 
                s2 = "0e+0"c; //"0e+0i";
            
            break;
        }   

    // Handle the case when there's only a single value 
    // to work with, and set the other string to zero.
    if (!s1.length)
    { s1 = sx; s2 = "0e+0"c; } //"0e+0i"; }
 
    //io.writefln( "splitComplexStrings() s=\"%s\", s1=\"%s\", s2=\"%s\", len=%d", 
    //           s, s1, s2, len );
}

// Displays on a (tsize) 4, 8, 12, etc boundary
void displayArguments(...)
{
    displayArguments(_arguments, _argptr);
}

void displayArguments
(
    in Diva oDiva, 
    in uint uiStartPos = 0U, 
    in uint uiEndPos = uint.max
)
{
    //io.writefln("void displayArguments(in Diva oDiva) called"); 
    if (oDiva is null)
        return;
        
    displayArguments(oDiva.ArgType, oDiva.ArgValue, uiStartPos, uiEndPos);
}

void displayArguments
(
    in TypeInfo[] _arguments, 
    in sa.va_list _argptr, 
    in uint uiStartPos = 0U, 
    in uint uiEndPos = uint.max
)
{
    uint uiArgLen = _arguments.length;   

    if (uiArgLen == 0U)
    {
        io.writefln("!!!!!! No arguments to display !!!!!!"c);
        return;
    }
    else if (uiStartPos >= uiArgLen)
        uiStartPos = uiArgLen - 1;
   
    if (uiEndPos > uiArgLen)
        uiEndPos = uiArgLen - 1;
    else if (uiEndPos < uiStartPos)
        uiEndPos = uiStartPos;

    sa.va_list vl1 = _argptr;
    sa.va_list vl2;
    sa.va_list vl3;
    short   sht = 0U;
    char[] sMask      = ", value size=%02d, elements=N/A, value="c;
    char[] sMaskArray = ", value size=%02d, elements=%03d, value=\"%s\" (0x%0X)"c;

    //io.writefln("args=%d", _arguments.length);    
    for (int i = uiStartPos; i <= uiEndPos; i++)
    {   
        vl3 = vl1;

        //_arguments[i].print();
        io.writef("typeid=%12s, addr=0x%0X, tsize=0x%02X"c,  
               _arguments[i], vl3, _arguments[i].tsize);
              
        if (_arguments[i] is typeid(real))
            io.writef(sMask ~ "%f"c, real.sizeof, sa.va_arg!(real)(vl1));
        else if (_arguments[i] is typeid(char)) 
            io.writef(sMask ~ "'%s'"c, char.sizeof, sa.va_arg!(char)(vl1));   
        else if (_arguments[i] is typeid(wchar)) 
            io.writef(sMask ~ "'%s'"c, wchar.sizeof, sa.va_arg!(wchar)(vl1));   
        else if (_arguments[i] is typeid(dchar)) 
            io.writef(sMask ~ "'%s'"c, dchar.sizeof, sa.va_arg!(dchar)(vl1));   
        else if (_arguments[i] is typeid(byte))  
            io.writef(sMask ~ "%d"c, byte.sizeof, sa.va_arg!(byte)(vl1));
        else if (_arguments[i] is typeid(ubyte)) 
            io.writef(sMask ~ "%d"c, ubyte.sizeof, sa.va_arg!(ubyte)(vl1));
        else if (_arguments[i] is typeid(short)) 
            io.writef(sMask ~ "%d"c, short.sizeof, sa.va_arg!(short)(vl1));
        else if (_arguments[i] is typeid(ushort)) 
            io.writef(sMask ~ "%d"c, ushort.sizeof, sa.va_arg!(ushort)(vl1));
        else if (_arguments[i] is typeid(int))  
            io.writef(sMask ~ "%d"c, int.sizeof, sa.va_arg!(int)(vl1)); 
        else if (_arguments[i] is typeid(uint)) 
            io.writef(sMask ~ "%d"c, uint.sizeof, sa.va_arg!(uint)(vl1)); 
        else if (_arguments[i] is typeid(long))  
            io.writef(sMask ~ "%d"c, long.sizeof, sa.va_arg!(long)(vl1));  
        else if (_arguments[i] is typeid(ulong)) 
            io.writef(sMask ~ "%d"c, ulong.sizeof, sa.va_arg!(ulong)(vl1));
        else if (_arguments[i] is typeid(float)) 
            io.writef(sMask ~ "%f"c, float.sizeof, sa.va_arg!(float)(vl1));
        else if (_arguments[i] is typeid(double))
            io.writef(sMask ~ "%f"c, double.sizeof, sa.va_arg!(double)(vl1));
        else if (_arguments[i] is typeid(ifloat)) 
            io.writef(sMask ~ "%g"c, ifloat.sizeof, sa.va_arg!(ifloat)(vl1));
        else if (_arguments[i] is typeid(idouble))
            io.writef(sMask ~ "%g"c, idouble.sizeof, sa.va_arg!(idouble)(vl1));
        else if (_arguments[i] is typeid(cfloat)) 
            io.writef(sMask ~ "%g"c, cfloat.sizeof, sa.va_arg!(cfloat)(vl1));
        else if (_arguments[i] is typeid(cdouble))
            io.writef(sMask ~ "%g"c, cdouble.sizeof, sa.va_arg!(cdouble)(vl1));
        else if (_arguments[i] is typeid(ireal))
            io.writef(sMask ~ "%g"c, ireal.sizeof, sa.va_arg!(ireal)(vl1));
        else if (_arguments[i] is typeid(creal))
            io.writef(sMask ~ "%g"c, creal.sizeof, sa.va_arg!(creal)(vl1));
        else if (_arguments[i] is typeid(bool))
            io.writef(sMask ~ "%b"c, bool.sizeof, sa.va_arg!(bool)(vl1));
        else if (_arguments[i] is typeid(Object))
            io.writef(sMask ~ "\"%s\""c, Object.sizeof, 
                   sa.va_arg!(Object)(vl1).toString());
        else if (_arguments[i] is typeid(void*))
            io.writef(sMask ~ "0x%X"c, void.sizeof, sa.va_arg!(void*)(vl1));
            
        // char[] { short, int* } 
        else if (_arguments[i] is typeid(char[]))
        {
            vl2 = vl1;
            sht = sa.va_arg!(short)(vl2);
            uint sp = sa.va_arg!(uint)(vl2);
            io.writef(sMaskArray, sht * char.sizeof, sht, sa.va_arg!(char[])(vl1), sp);
        }    
        else if (_arguments[i] is typeid(wchar[]))
        {
            vl2 = vl1;
            sht = sa.va_arg!(short)(vl2);
            uint sp = sa.va_arg!(uint)(vl2);
            io.writef(sMaskArray, sht * wchar.sizeof, sht, sa.va_arg!(wchar[])(vl1), sp);
        }    
        else if (_arguments[i] is typeid(dchar[]))
        {
            vl2 = vl1;
            sht = sa.va_arg!(short)(vl2);
            uint sp = sa.va_arg!(uint)(vl2);
            io.writef(sMaskArray, sht * dchar.sizeof, sht, sa.va_arg!(dchar[])(vl1), sp);
        }  
        else if (_arguments[i] is typeid(void[]))
        {
            vl2 = vl1;
            sht = sa.va_arg!(short)(vl2);
            uint sp = sa.va_arg!(uint)(vl2);
            io.writef(sMaskArray, sht * void.sizeof, sht, cast(char[])sa.va_arg!(void[])(vl1), sp);
        }  
        else
            vl1 += _arguments[i].tsize;

        io.writefln(); 
    }  
    
    io.writefln("ending addr=%X"c, vl1);
}

//alias Diva.get!(byte).get     get;
//alias Diva.get!(ubyte).get    get;

debug( diva )
{
/****************************************
 * (Note: Copied here from std.math's mfeq() function for unittesting)
 * Simple function to compare two floating point values
 * to a specified precision.
 * Returns:
 *  1   match
 *  0   nomatch
 */
private bool feq(in real r1, in real r2, in real precision = 0.000001L)
{
    if (r1 == r2)
        return 1;
    
    if (sm.isnan(r1))
        return cast(bool)sm.isnan(r2);

    if (sm.isnan(r2))
        return 0;
        
    return cast(bool)(sm.fabs(r1 - r2) <= precision);
} 
}

unittest
{
    Diva oDiva = new Diva;
    uint ui = 0UL;

    oDiva.append('C', 123L, 456UL, true, 987.56f);
    assert(oDiva.getNumberArgs() == 5);
    assert(oDiva.getArgType()  != null);
    assert(oDiva.getArgValue() != null);
    assert(oDiva.getUtilized() == 28);
    assert(oDiva.getCapacity() == 512);
    oDiva.insert(3, "Test string"c);
    assert(oDiva.getNumberArgs() == 6);
    assert(oDiva.getUtilized() == 36);
    oDiva.squeeze();
    assert(oDiva.getCapacity() == 36);

    ui = oDiva.get!(uint)(1);
    assert(ui == 123);
    oDiva.set!(bool)(1, true);
    assert(oDiva.get!(bool)(1) == true);
    assert(oDiva.get!(char)(0) == 'C');
    assert(oDiva.get!(ulong)(2) == 456UL);
    assert(oDiva.get!(char[])(3) == "Test string"c);
    assert(oDiva.get!(bool)(4) == true);
    assert(oDiva.get!(float)(5) == 987.56f);
    oDiva.set!(float)(5, 345.99e-2f);
    assert(oDiva.get!(float)(5) == 345.99e-2f);

    // these over boundary get()s will cause an error. 
    //assert(oDiva.get!(float)(6) == 0.0f);
    //assert(oDiva.get!(char[])(6) == ""c);

    oDiva.remove(4);
    assert(oDiva.getNumberArgs() == 5);
    assert(oDiva.getUtilized() == 32);
    oDiva.free();
    assert(oDiva.getNumberArgs() == 0);
    assert(oDiva.getArgType()  == null);
    assert(oDiva.getArgValue() == null);
    assert(oDiva.getUtilized() == 0);
    assert(oDiva.getCapacity() == 0);

    // Needed some negative and float value testing.
    oDiva.append(-123L, 456123454L, 78U, -9565.57f, -0, 0xFF, "-3456");
    assert(oDiva.get!(int)(0) == -123);
    assert(oDiva.get!(ulong)(0) == 0L); 
    assert(oDiva.get!(real)(1) == 456123454L);
    assert(oDiva.get!(long)(3) == 0); 
    debug( diva ) assert(.feq(cast(real)oDiva.get!(double)(3), cast(real)-9565.570313, 0.001L));
    assert(oDiva.get!(byte)(5) == 0);
    assert(oDiva.get!(ubyte)(5) == 255);
    debug( diva ) assert(.feq(cast(real)oDiva.get!(ireal)(2), cast(real)78.0Li));
    assert(oDiva.get!(bool)(4) == false); // -0 causes the default false
    assert(oDiva.get!(long)(3) == 0L);
    assert(oDiva.get!(char[])(3) == "-9565.57"c);
    assert(oDiva.get!(long)(6) == -3456L);
}

debug( diva )
{
int main()
{
    io.writefln();
    io.writefln("Compiled and tested with: \"%s v%d.%03d\"", scomp.name, scomp.version_major, scomp.version_minor);
    io.writefln("diva.unittest finished!"c); 
	
    return 0;
}
}