/+
 ' Module          : dynstrstack.d - General Dynamic String (char[]) Stack Class
 ' Version         : v1.0
 ' Author          : David L. 'SpottedTiger' Davis
 ' Source Platform : Intel x86 / WinXP SP2 
 ' Target Platform : Should work on any CPU / OS combo that D compiles on.
 '                 :
 ' Created Date    : 01.Nov.04 Complied and Tested with dmd v0.105.
 ' Modified Date   : 04.Nov.04 Added unittest code, recompiled and tested.
 '                 : 08.Nov.04 Added import convext for converting strings
 '                 :           back to a floating numbers. 
 '                 : 09.Nov.04 Recomplied and Tested with dmd v0.106.
 '                 : 20.Jan.05 Recomplied and Tested with dmd v0.111.
 '                 : 20.Mar.05 Recomplied and Tested with dmd v0.119.
 '                 : 09.Apr.05 Recomplied and Tested with dmd v0.120.
 '                 : 06.Jun.05 Recomplied and Tested with dmd v0.126, removed convext.d.
 '                 : 17.Sep.05 Recomplied and Tested with dmd v0.131.
 '                 : 26.Oct.05 Recompiled and Tested with dmd v0.137.    
 '                 : 30.Dec.05 Recompiled and Tested with dmd v0.142.  
 '                 : 19.Mar.06 Recompiled and Tested with dmd v0.150. 
 '                 : 28.May.06 Recompiled and Tested with dmd v0.159. 
 '                 : 19.Jul.06 All "Imports" are now private by default 
 '                 :           (which shouldn't be a problem, since I
 '                 :           always use the private keyword) in 
 '                 :           dmd v0.163, recompiled and tested.
 '                 : 23.Oct.06 Recompiled and Tested with dmd v0.172.
 '                 : 11.Dec.06 Recompiled and Tested with dmd v0.177.
 '                 : 05.Jan.07 Recompiled and Tested with dmd v1.0.
 '                 : 09.Feb.07 Recompiled and Tested with dmd v1.005.
 '                 :           Replaced t_bit to t_bool to reflect something 
 '                 :           that changed in D some time ago. 
 '                 : 09.Jan.08 Recompiled and Tested with dmd v1.025. 
 '                 : 18.Jul.08 Recompiled and Tested with dmd v1.033. 
 '                 : 13.May.09 Recompiled and Tested with dmd v1.045.
 '                 :           
 '                 :
 ' Requirements    : std.string for the toString() conversion functions,
 '                 : std.utf for the toUTF16() and toUTF32() functions,
 '                 : std.conv for all the whole number and floating point 
 '                 : number conversions.
 '                 :
 ' Licence         : Public Domain
 ' ---------------------------------------------------------------------------
 ' Note: The purpose of this class, is to allow for storing any of D's basic data 
 '       types into one char[] array stack without the need for using a template approach.  
 '       Values pop-off the stack as a char[] values, but can be easily converted 
 '       back as needed to their original data type (just look at the example code  
 '       written in the unittest on how I did it for testing). Also the stack does 
 '       allow for pushing and popping of wchar[]s and dchar[]s on and off the stack
 '       without any data lost.
 '
 ' To complie for a unittest:
 ' dmd dynstrstack.d -debug=dynstrstack -unittest
 '
 ' To compile a D program that use dynstrstack.d:
 ' dmd MySource.d dynstrstack.d 
 '
 ' Public class member functions:
 ' ------------------------------
 ' 01) void push( in char[] )
 ' 02) void pushW( in wchar[] ) 
 ' 03) void pushD( in dchar[] )
 ' 04) void push( in char )
 ' 05) void pushW( in wchar )
 ' 06) void pushD( in dchar )
 ' 07) void push( in bool )
 ' 08) void push( in byte )
 ' 09) void push( in int )
 ' 10) void push( in short )
 ' 11) void push( in long )
 ' 12) void push( in ubyte )
 ' 13) void push( in uint ) 
 ' 14) void push( in ushort )
 ' 15) void push( in ulong )
 ' 16) void push( in float )
 ' 17) void push( in double )
 ' 18) void push( in real )
 ' 19) void push( in ifloat )
 ' 20) void push( in idouble )
 ' 21) void push( in ireal )
 ' 22) void push( in cfloat )
 ' 23) void push( in cdouble )
 ' 24) void push( in creal ) 
 ' 25) char[] pop() // main pop() workhorse
 ' 26) char[] pop( out ubyte )
 ' 27) wchar[] popW()
 ' 28) wchar[] popW( out ubyte )
 ' 29) dchar[] popD()
 ' 30) dchar[] popD( out ubyte )
 ' 31) uint length()
 '
 ' Private class member functions:
 ' -------------------------------
 ' 01) void push( in char[], in ubyte ) // main push() workhorse
 ' 02) ubyte poptype()
 '
 +/
module dynstrstack;

private import std.string; // for the toString() conversion functions
private import std.utf;    // for the toUTF16() and toUTF32() functions
private import std.conv;   // for the whole number conversion functions
//private import convext;  // for the floating number conversion functions (these are now in std.conv)

debug( dynstrstack )
{
private import std.stdio;  // for the writefln() and writef() functions 
}

enum : ubyte 
{ 
    t_bool    = 0x00,
    t_byte    = 0x01, 
    t_short   = 0x02,    
    t_int     = 0x03,
    t_long    = 0x04,
    t_ubyte   = 0x05, 
    t_ushort  = 0x06,     
    t_uint    = 0x07,
    t_ulong   = 0x08,
    t_float   = 0x09,
    t_double  = 0x0A,
    t_real    = 0x0B,
    t_ifloat  = 0x0C,
    t_idouble = 0x0D,
    t_ireal   = 0x0E,
    t_cfloat  = 0x0F,
    t_cdouble = 0x10,
    t_creal   = 0x11,
    t_char    = 0x12,
    t_wchar   = 0x13,
    t_dchar   = 0x14,
    t_charA   = 0x15,
    t_wcharA  = 0x16,
    t_dcharA  = 0x17,
}

// Dynamic String (char[]) Stack Class       
class DynStrStack
{
    struct udtStack
    {
        char[] sValue;
        ubyte  ubytType;
    }

    private udtStack[] oStack;
    private static int iCurrPos;
 
    private ubyte poptype()
    {
        if (iCurrPos > 0)
            return oStack[iCurrPos - 1].ubytType;
        else
            return t_charA;    
    }
    
    private void push(in char[] s, in ubyte ubytType) 
    {
        if (s == null)
            return;
        
        iCurrPos = oStack[].length + 1;
        oStack.length = iCurrPos;
        oStack[iCurrPos - 1].sValue = s.dup;
        oStack[iCurrPos - 1].ubytType = ubytType;
        
        //writefln( "s.pushed %s, iCurrPos=%d, tiTypeID=%X", 
        //   oStack[][ iCurrPos - 1 ].sValue, iCurrPos - 1, oStack[ iCurrPos - 1 ].ubytType );
    }
    
    public char[] pop()
    {
        char[] s;
        ubyte  ubytType;

        if (iCurrPos == 0)
            return ""; 
            
        iCurrPos = iCurrPos - 1;
        
        s = oStack[][iCurrPos].sValue.dup;
        ubytType = oStack[iCurrPos].ubytType;

        if (ubytType < t_char)
            s = strip(s);
        
        //writefln("s.popped %s, iCurrPos=%d, tiTypeID=%X", oStack[][iCurrPos].sValue, iCurrPos, ubytType);
               
        oStack.length = iCurrPos;
        
        return s;
    }    

    public uint length() {return iCurrPos;}    
    public char[] pop(out ubyte ubytType) { ubytType = poptype(); return pop(); }

    public wchar[] popW() 
    { 
        ubyte ubytType = poptype(); 
        
        if (ubytType == t_wcharA || ubytType == t_wchar)
            return cast(wchar[])pop();
        else
            return cast(wchar[])std.utf.toUTF16(pop());

    } // end wchar[] popW()

    public wchar[] popW(out ubyte ubytType) 
    { 
        ubytType = poptype(); 
        
        if (ubytType == t_wcharA || ubytType == t_wchar)
            return cast(wchar[])pop();
        else
            return cast(wchar[])std.utf.toUTF16(pop());

    } // end wchar[] popW( out ubyte )

    public dchar[] popD() 
    { 
        ubyte ubytType = poptype(); 
        
        if (ubytType == t_dcharA || ubytType == t_dchar)
            return cast(dchar[])pop(); 
        else
            return cast(dchar[])std.utf.toUTF32(pop());

    } // end dchar[] popD()

    public dchar[] popD(out ubyte ubytType) 
    { 
        ubytType = poptype(); 
        
        if (ubytType == t_dcharA || ubytType == t_dchar)
            return cast(dchar[])pop(); 
        else
            return cast(dchar[])std.utf.toUTF32(pop());

    }  // end dchar[] popD( out ubyte )
    
    // Handle different character array and character types 
    public void push(in char[] s)    { push( s, t_charA ); }
    public void pushW(in wchar[] ws) { push( cast(char[])ws, t_wcharA ); }
    public void pushD(in dchar[] ds) { push( cast(char[])ds, t_dcharA ); }
    public void push(in char c)      { push( std.string.toString( c ), t_char ); }
    public void pushW(in wchar wc)   { push( std.string.toString( cast(char)wc ), t_wchar ); }
    public void pushD(in dchar dc)   { push( std.string.toString( cast(char)dc ), t_dchar ); }
   
    // Convert number types to char[]s
    public void push(in bool b)       { push( std.string.toString( cast(ubyte)b ), t_bool ); }
    public void push(in byte byt)    { push( std.string.toString( byt ),  t_byte ); }
    public void push(in short sht)   { push( std.string.toString( sht ),  t_short ); }
    public void push(in int i)       { push( std.string.toString( i ),    t_int ); }
    public void push(in long l)      { push( std.string.toString( l ),    t_long ); }
    public void push(in ubyte ubyt)  { push( std.string.toString( ubyt ), t_ubyte ); }
    public void push(in ushort usht) { push( std.string.toString( usht ), t_ushort ); }
    public void push(in uint ui)     { push( std.string.toString( ui ),   t_uint ); }
    public void push(in ulong ul)    { push( std.string.toString( ul ),   t_ulong ); }
    public void push(in float f)     { push( std.string.toString( f ),    t_float ); }
    public void push(in double d)    { push( std.string.toString( d ),    t_double ); }
    public void push(in real r)      { push( std.string.toString( r ),    t_real ); }
    public void push(in ifloat ift)  { push( std.string.toString( ift ),  t_ifloat ); }
    public void push(in idouble id)  { push( std.string.toString( id ),   t_idouble ); }
    public void push(in ireal ir)    { push( std.string.toString( ir ),   t_ireal ); }
    public void push(in cfloat cf)   { push( std.string.toString( cf ),   t_cfloat ); }
    public void push(in cdouble cd)  { push( std.string.toString( cd ),   t_cdouble ); }
    public void push(in creal cr)    { push( std.string.toString( cr ),   t_creal ); }

} // end class DynStrStack

unittest
{
    // Types allowed to store
    char[]  s    = "char[]"c;
    wchar[] ws   = "wchar[]"w;
    dchar[] ds   = "dchar[]"d;
    char    c    = 'C';
    wchar   wc   = 'W';
    dchar   dc   = 'D'; 
    bool    b    = true;
    byte    byt  = byte.max;
    short   sht  = short.max;
    int     i    = int.max;
    long    l    = long.max; 
    ubyte   ubyt = ubyte.max;
    ushort  usht = ushort.max;
    uint    ui   = uint.max;
    ulong   ul   = ulong.max;
    float   f    = float.max;
    double  d    = double.max;
    real    r    = real.max;
    ifloat  ift  = ifloat.max;
    idouble id   = idouble.max;
    ireal   ir   = ireal.max;
    cfloat  cf   = cfloat.max;
    cdouble cd   = cdouble.max;
    creal   cr   = creal.max;
    //------------------------
    char[]  sPopValue = "";
    ubyte   ubytType  = t_char;
    uint    uiLen;
    int     iPos1     = 0;
    int     iPos2     = 0;
    char[]  sXValue   = "";
    real    r1;
    real    r2;
    char[]  s1;
    char[]  s2;
    
    const char[][] sType = 
    [ 
        "bool"c, "byte"c, "short"c, "int"c, "long"c, "ubyte"c,
        "ushort"c, "uint"c, "ulong"c, "float"c, "double"c, "real"c,
        "ifloat"c, "idouble"c, "ireal"c, "cfloat"c, "cdouble"c, "creal"c, 
        "char"c, "wchar"c, "dchar"c, "char[]"c, "wchar[]"c, "dchar[]"c
    ];

    DynStrStack os = new DynStrStack;
    
    writefln( "dynstrstack.unittest started."c );
    os.push(s);
    os.pushW(ws);
    os.pushD(ds);
    os.push(c);
    os.pushW(wc);
    os.pushD(dc);
    os.push(b);
    os.push(byt);
    os.push(sht);
    os.push(i);
    os.push(l);
    os.push(ubyt);
    os.push(usht);
    os.push(ui);
    os.push(ul);
    os.push(f);
    os.push(d);
    os.push(r);
    os.push(ift);
    os.push(id);
    os.push(ir);
    os.push(cf);
    os.push(cd);
    os.push(cr);

    uiLen = os.length;
    assert(uiLen == 24);

    for (uint uix = 0; uix < uiLen; uix++)
    {
       sPopValue = os.pop( ubytType ); 

       if (sPopValue[sPopValue.length - 1] == 'i')
           sPopValue.length = sPopValue.length - 2;

       /+
        ' An example of how to use the ubytType variable
        ' in converting the data back into its original
        ' basic data type.
        +/
       switch (ubytType)
       {
         case t_bool : 
             b = (sPopValue == "1"c ? true : false); 
             writefln("Type=%s, Value=%b"c, sType[ubytType], b);
             assert(b == true);
             break;

         case t_byte : 
             byt = std.conv.toByte(sPopValue); 
             writefln("Type=%s, Value=0x%0X"c, sType[ubytType], byt);
             assert(byt == 0x7F);
             break;

         case t_short : 
             sht = std.conv.toShort(sPopValue); 
             writefln("Type=%s, Value=%d"c, sType[ubytType], sht);
             assert(sht == sht.max);
             break;

         case t_int : 
             i = std.conv.toInt(sPopValue); 
             writefln("Type=%s, Value=%d"c, sType[ubytType], i);
             assert(i == int.max);
             break;

         case t_long : 
             l = std.conv.toLong( sPopValue); 
             writefln("Type=%s, Value=%d"c, sType[ubytType], l);
             assert(l == long.max);
             break;

         case t_ubyte : 
             ubyt = cast(ubyte)std.conv.toUint(sPopValue); 
             writefln("Type=%s, Value=0x%0X"c, sType[ubytType], ubyt);
             assert(ubyt == 0xFF);
             break;

         case t_ushort : 
             usht = std.conv.toUshort(sPopValue); 
             writefln("Type=%s, Value=%d"c, sType[ubytType], usht);
             assert(usht == ushort.max);
             break;

         case t_uint : 
             ui   = std.conv.toUint(sPopValue); 
             writefln("Type=%s, Value=%d"c, sType[ubytType], ui);
             assert(ui == uint.max);
             break;

         case t_ulong : 
             ul = std.conv.toUlong(sPopValue); 
             writefln("Type=%s, Value=%d"c, sType[ubytType], ul);
             assert(ul == ulong.max);
             break;

         case t_float : 
             f = std.conv.toFloat(sPopValue);
             writefln("Type=%s, Value=%g, Max=%g"c, sType[ubytType], f, float.max);
             //assert( f == float.max );
             assert(sPopValue == std.string.toString(float.max));
             break;

         case t_double : 
             d = std.conv.toDouble(sPopValue);
             writefln("Type=%s, Value=%g, max=%g"c, sType[ubytType], d, double.max);
             //assert( d == double.max );
             assert(sPopValue == std.string.toString(double.max));
             break;

         case t_real : 
             r = std.conv.toReal(sPopValue);
             writefln("Type=%s, Value=\"%s\", r=%14.7g"c, sType[ubytType], sPopValue, r);
             assert(sPopValue == std.string.toString(real.max));
             //assert(std.math2.feq(r, real.max, 0.0000001) ); // 0.1e+4932));
             break;

         case t_ifloat :
             //ift = std.conv.toIfloat(sPopValue);

             //** A temp fix until the toIreal() is ready in std.conv **
             r1 = std.conv.toReal(sPopValue);
             ift = cast(ifloat)(r1 * 1.0i);

             writefln("Type=%s, Value=\"%s\", ift=%gi"c, sType[ubytType], sPopValue, ift);
             //writefln("toString()=\"%s\"", std.string.toString(ift));
             break;

         case t_idouble :
             //id = std.conv.toIdouble(sPopValue);

             //** A temp fix until the toIreal() is ready in std.conv **
             r1 = std.conv.toReal(sPopValue);
             id = cast(idouble)(r1 * 1.0i);
              
             writefln("Type=%s, Value=\"%s\", id=%gi"c, sType[ubytType], sPopValue, id);
             break;

         case t_ireal :
             //ir = std.conv.toIreal(sPopValue);
 
             //** A temp fix until the toIreal() is ready in std.conv **
             r1 = std.conv.toReal(sPopValue);
             ir = cast(ireal)(r1 * 1.0i);
             
             writefln("Type=%s, Value=\"%s\", ir=%gi"c, sType[ubytType], sPopValue, ir);
             break;
             
         case t_cfloat :
             //cf = std.conv.toCfloat(sPopValue);
 
             //** A temp fix until the toCreal() is ready in std.conv **
             .splitComplexStrings(sPopValue, s1, s2);
             //writefln("sx=\"%s\", s1=\"%s\", s2=\"%s\"", sx, s1, s2);
             r1 = std.conv.toReal(s1);
             r2 = std.conv.toReal(s2);
             cf = cast(cfloat)(r1 + (r2 * 1.0i));
             
             writefln("Type=%s, Value=\"%s\", cf=%g"c, sType[ubytType], sPopValue, cf);
             break;

         case t_cdouble :
             //cd = std.conv.toCdouble(sPopValue);
 
             //** A temp fix until the toCreal() is ready in std.conv **
             .splitComplexStrings(sPopValue, s1, s2);
             //writefln("sx=\"%s\", s1=\"%s\", s2=\"%s\"", sx, s1, s2);
             r1 = std.conv.toReal(s1);
             r2 = std.conv.toReal(s2);
             cd = cast(cdouble)(r1 + (r2 * 1.0i));
             
             writefln("Type=%s, Value=\"%s\", cd=%g"c, sType[ubytType], sPopValue, cd);
             break;

         case t_creal :
             //cr = std.conv.toCreal(sPopValue);
 
             //** A temp fix until the toCreal() is ready in std.conv **
             .splitComplexStrings(sPopValue, s1, s2);
             //writefln("sx=\"%s\", s1=\"%s\", s2=\"%s\"", sx, s1, s2);
             r1 = std.conv.toReal(s1);
             r2 = std.conv.toReal(s2);
             cr = cast(creal)(r1 + (r2 * 1.0i));
             
             writefln("Type=%s, Value=\"%s\", cr=%g"c, sType[ubytType], sPopValue, cr);
             break;

         case t_char : 
             c = sPopValue[0]; 
             assert(c == 'C');
             writefln("Type=%s, Value=\'%s\'"c, sType[ubytType], c);
             break;

         case t_wchar : 
             wc = cast(wchar)sPopValue[0]; 
             writefln("Type=%s, Value=\'%s\'"c, sType[ubytType], wc); 
             assert(wc == 'W');
             break;

         case t_dchar : 
             dc = cast(dchar)sPopValue[0]; 
             writefln("Type=%s, Value=\'%s\'"c, sType[ubytType], dc);  
             assert(dc == 'D');
             break;

         case t_charA : 
             s = sPopValue; 
             writefln("Type=%s, Value=\"%s\""c, sType[ubytType], s);
             assert(s == "char[]"c);
             break;

         case t_wcharA : 
             ws = cast(wchar[])sPopValue;
             writefln("Type=%s, Value=\"%s\""c, sType[ubytType], ws);
             assert(ws == "wchar[]"w);
             break;

         case t_dcharA : 
             ds = cast(dchar[])sPopValue; 
             writefln("Type=%s, Value=\"%s\""c, sType[ubytType], ds);
             assert(ds == "dchar[]"d);
             break;

         default : 
       }
    }

    writefln("dynstrstack.unittest done."c);
    writefln();
}

/+
 ' Splits a complex floats (cfloat, cdouble, and creal) into two workable strings.
 ' Grammar:
 ' (for cfloat, cdouble, and creal)
 ' ['+'|'-']digit(s)[.][digit(s)][[e-|e+]digit(s)][+]
 '         [digit(s)[.][digit(s)][[e-|e+]digit(s)][i|f|L|Li|fi]]
 '      or [nan|nani|nan+nani|inf|-inf]
 '
 ' creal Range: from 3.362103E-4932+3.362103E-4932i (min) 
 '                to 1.189731E+4932+1.189731E+4932i (max)
 +/
bool splitComplexStrings(in char[] s, out char[] s1, out char[] s2)
{
    char[] sx   = s;
    int    iLen = s.length;
    int    plus = 0;
    
    s1 = ""c;
    s2 = ""c;
    
    if (!iLen)
        throw new Exception("spiitComplexStrings(): "c ~
                            " input string is an empty string."c);

    if (iLen >= 2)
    {
        sx = std.string.tolower(s);   
        // When "nan" or "nani" just return them.
        if (sx == "nan"c || sx == "nani"c ||
            sx == "nan+nani"c || sx =="nan+0i"c)
        { s1 = "nan"c; s2 = "nani"c; return 1; }
    
        if (sx == "inf"c)
        { s1 = "inf"c; s2 = "infi"c; return 1; }    
    
        if (sx == "-inf"c)
        { s1 = "-inf"c; s2 = "-infi"c; return 1; }    
    }
    
    // Split the original string out into two strings.
    for (int i = 1; i < iLen; i++)
        if ((sx[i - 1] != 'e' && sx[i - 1] != 'E') && sx[i] == '+')
        {
            s1 = sx[0..i];
            if (i + 1 < iLen - 1)
                s2 = sx[i + 1..iLen]; // - 1];
            else 
                s2 = "0e+0"c; //"0e+0i";
            
            break;
        }   

    // Handle the case when there's only a single value 
    // to work with, and set the other string to zero.
    if (!s1.length)
    { s1 = sx; s2 = "0e+0"c; } //"0e+0i"; }
 
    //writefln( "splitComplexStrings() s=\"%s\", s1=\"%s\", s2=\"%s\", len=%d", 
    //           s, s1, s2, len );
   
    return true;
}

debug( dynstrstack )
{
int main()
{
    char[]  sR;
    char[]  s1 = "1st"c;
    wchar[] ws = "wchar[]"w;
    dchar[] ds = "dchar[]"d;
    wchar[] wsR;
    dchar[] dsR;
    ulong   ulVal;
    ubyte   ubytType;
    char[]  s2;
    wchar[] ws2;
    
    DynStrStack os = new DynStrStack;

    os.push(s1);
    os.push("2nd"c);
    os.push("3rd"c);
    dsR = os.popD();
    writefln( "dsR=%s"d, dsR );
    
    os.push("4th"c);
    os.push("5th"c);
    os.push(null);
    
    sR = os.pop();
    
    sR = os.pop();
    sR = os.pop();
    sR = os.pop();
    
    sR = os.pop();
    sR = os.pop();
    sR = os.pop();
    
    //writefln("os.length=%d", os.inStack);
 
    os.push(120.34);
    os.push(34567L);
    os.push(252);
    os.push(3.4567f);
    os.push(false);
    os.push(double.max);
    sR = os.pop();
    //os.push(real.max);
    os.pushW("Text"w);
    wsR = os.popW(ubytType);
    writefln( "sR=%s, type=%X"w, wsR, ubytType );
    sR = os.pop();
    sR = os.pop();
    sR = os.pop();
    sR = os.pop();
    dsR = os.popD(ubytType);
    writefln( "dsR=%s, type=%X"d, dsR, ubytType );
    //ulVal = os.popW();
    //os.pop();
    
    os.pushD(ds); 
    os.pushW(ws);
    wsR = os.popW();
    sR = os.pop(ubytType);

    if (ubytType == t_dcharA || ubytType == t_dchar)
        dsR = cast(dchar[])(sR);
    else
        dsR = ""d;

    writefln( "wsR=%s"w, wsR );
    writefln( "sR=%s, dsR=%s"d, sR, dsR );
    //sTypeID = typeid( double ).toString();
    //writefln("double typeid=%s", sTypeID);
       
    os.pushW('A');
    //ws2 = os.popW();
    os.pop();
    
    os.push('A');
    s2 = os.pop(ubytType);

    return 0;
    
} // end int main()
}