/+
 ' Source          : isBox.d - Set of "isFunc()" utility functions.
 ' Version         : v1.0 
 ' Author          : David L. 'SpottedTiger' Davis
 ' Source Platform : Intel x86 / WinXP SP2 
 ' Target Platform : Should work on any CPU / OS combo that D compiles on.
 '                 :
 ' Created  Date   : 22.Jun.05 Compiled and Tested with dmd v0.127.
 ' Modified Date   : 10.Sep.05 Recompiled and Tested with dmd v0.131.
 '                 : 26.Oct.05 Recompiled and Tested with dmd v0.137.    
 '                 : 30.Dec.05 Recompiled and Tested with dmd v0.142. 
 '                 : 23.Mar.06 Recompiled and Tested with dmd v0.150.  
 '                 : 23.Mar.06 Recompiled and Tested with dmd v0.172. 
 '                 : 05.Nov.06 Recompiled and Tested with dmd v0.173.
 '                 : 09.Dec.06 Because of "improved name mangling" changes done in dmd v0.176,
 '                 : had to change "extern (C) void assert_3std5boxer() { }" to
 '                 : "extern (C) int D3std5boxer8__assertFiZv() { return 0; }", then
 '                 : recompiled and unittested with dmd v0.177.
 '                 : recompiled and unittested with dmd v1.0.
 '                 : recompiled and unittested with dmd v1.025. 
 '                 : Commented extern (C) int D3std5boxer8__assertFiZv() { return 0; }
 '                 : 20.Mar.08 recompiled and unittested with dmd v1.028.
 '                 : 20.Jun.08 recompiled and unittested with dmd v1.031.
 '                 : 17.Jul.08 recompiled and unittested with dmd v1.033. 
 '                 : 13.May.09 recompiled and unittested with dmd v1.045.
 '                 :
 ' License         : Public Domain
 ' -----------------------------------------------------------
 ' To Debug:   dmd isbox.d -debug=isbox
 ' To Compile: dmd mycode.d isbox.d
 +/
module isbox;

debug( isbox ) private import scomp = std.compiler;
private import std.stdio;
private import std.string;
private import std.utf;
private import std.boxer;

// Burton Radons <burton-radons[at]smocky[dot]com>
// This entry allows -unittest for testing, and/or no need for the -release switch
// extern (C) void assert_3std5boxer() { } // dmd v0.175 and below
//extern (C) int D3std5boxer8__assertFiZv() { return 0; } // dmd v0.176+

// Returns a true if the Box stored value is a string 
// and it is an empty string value.
bool isEmpty(in Box boxx)
{
    TypeInfo ti = boxx.type();

    // Defined, but never assigned anything
    if (ti is null)
        return false;
    else if (ti is typeid(char[]))
        if (unbox!(char[])(boxx) == ""c)
            return true;
    else if (ti is typeid(wchar[]))
        if (unbox!(wchar[])(boxx) == ""w)
            return true;
    else if (ti is typeid(dchar[]))
        if (unbox!(dchar[])(boxx) == ""d)
            return true;

    return false;    
}

// Returns true is the Box has never been unassigned anything, 
// or if the value that the Box stores is a null.
bool isNull(in Box boxx)
{
    TypeInfo ti = boxx.type();
    
    // Defined, but never assigned anything
    if (ti is null)
        return true;

    // Checks now for if the value is a null
    // As in Box boxx = box(null);
    void[] va = boxx.data();
    char[] s  = cast(char[])va;
    
    if (s[0] == 0 &&
        s[1] == 0 &&
        s[2] == 0 &&
        s[3] == 0 &&
        ti is typeid(void *) && 
        s.length == 4)
        return true;
    else
        return false;
}

bool isSignedInteger(in Box boxx)
{ 
    TypeInfo ti = boxx.type();

    // Defined, but never assigned anything
    if (ti is null)
        return false;
    else if (ti is typeid(byte)  ||
             ti is typeid(short) ||
             ti is typeid(int)   ||
             ti is typeid(long))
        return true;
    else
        return false;
}

bool isUnsignedInteger(in Box boxx)
{
    TypeInfo ti = boxx.type();
 
    // Defined, but never assigned anything
    if (ti is null)
        return false;
    else if (ti is typeid(ubyte)  ||
             ti is typeid(ushort) ||
             ti is typeid(uint)   ||
             ti is typeid(ulong))
        return true;
    else
        return false;
} 

bool isFloatingPoint(in Box boxx)
{
    TypeInfo ti = boxx.type();

    // Defined, but never assigned anything
    if (ti is null)
        return false;
    else if (ti is typeid(float)  ||
             ti is typeid(double) ||
             ti is typeid(real))
        return true;
    else
        return false;
}

bool isImaginaryFloat(in Box boxx)
{
    TypeInfo ti = boxx.type();

    // Defined, but never assigned anything
    if (ti is null)
        return false;
    else if (ti is typeid(ifloat)  ||
             ti is typeid(idouble) ||
             ti is typeid(ireal))
        return true;
    else
        return false;
}

bool isComplexFloat(in Box boxx)
{
    TypeInfo ti = boxx.type();

    // Defined, but never assigned anything
    if (ti is null)
        return false;
    else if (ti is typeid(cfloat)  ||
             ti is typeid(cdouble) ||
             ti is typeid(creal))
        return true;
    else
        return false;
}

bool isStringUTF8(in Box boxx)
{
    TypeInfo ti = boxx.type();

    // Defined, but never assigned anything
    if (ti is null)
        return false;
    else if (ti is typeid(char[]))
        return true;
    else
        return false;
}

bool isCharUTF8(in Box boxx)
{
    TypeInfo ti = boxx.type();

    // Defined, but never assigned anything
    if (ti is null)
        return false;
    else if (ti is typeid(char))
        return true;
    else
        return false;
}

bool isStringUTF16(in Box boxx)
{
    TypeInfo ti = boxx.type();

    // Defined, but never assigned anything
    if (ti is null)
        return false;
    else if (ti is typeid(wchar[]))
        return true;
    else
        return false;
}

bool isCharUTF16(in Box boxx)
{
    TypeInfo ti = boxx.type();

    // Defined, but never assigned anything
    if (ti is null)
        return false;
    else if (ti is typeid(wchar))
        return true;
    else
        return false;
}

bool isStringUTF32(in Box boxx)
{
    TypeInfo ti = boxx.type();
    
    // Defined, but never assigned anything
    if (ti is null)
        return false;
    else if (ti is typeid(dchar[]))
        return true;
    else
        return false;
}

bool isCharUTF32(in Box boxx)
{
    TypeInfo ti = boxx.type();
    
    // Defined, but never assigned anything
    if (ti is null)
        return false;
    else if (ti is typeid(dchar))
        return true;
    else
        return false;
}

char[] toString(in Box boxx)
{
    TypeInfo ti = boxx.type();
    
    // Defined, but never assigned anything
    if (ti is null)
        return ""c;

    return boxx.toString();
}

wchar[] toStringW(in Box boxx)
{
    TypeInfo ti = boxx.type();

    // Defined, but never assigned anything
    if (ti is null)
        return ""w;

    return std.utf.toUTF16(boxx.toString());
}

dchar[] toStringD(in Box boxx)
{
    TypeInfo ti = boxx.type();

    // Defined, but never assigned anything
    if (ti is null)
        return ""d;

    return std.utf.toUTF32(boxx.toString());
}

int getDataSize(in Box boxx)
{
    TypeInfo ti = boxx.type();
    
    // Defined, but never assigned anything
    if (ti is null)
        return 0;
     
    return boxx.data().length;
}

private void IncorrectReturnError(in char[] sFunc, in char[] sValue, in char[] sType) 
{ 
    throw new Exception(sFunc ~ "() should return true for the value \"" ~ 
                        sValue ~ "\" of type (" ~ sType ~ ")."); 
}    

debug( isbox )
{
int main()
{
    Box   boxx;
    Box   boxy;
    Box[] boxz;
    
    TypeInfo ti;
    char[] s;

    writefln( "isbox.unittest started" );
    writefln( "isbox.isImaginaryFloat(in Box).unittest" );
    boxx = box(123i);
    if (!isImaginaryFloat(boxx))
        IncorrectReturnError("isImaginaryFloat", boxx.toString(), boxx.type().toString()); 

    writefln( "isbox.isComplexFloat(in Box).unittest" );
    boxx = box(56+0fi);
    if (!isComplexFloat(boxx))
        IncorrectReturnError("isComplexFloat", boxx.toString(), boxx.type().toString()); 

    //writefln("%f, isInt=%b, isUint=%b, isFloat=%b, isImaginaryFloat=%b, isComplexFloat=%b, size=%d", 
    //         123i, isSignedInteger(boxx), isUnsignedInteger(boxx), isFloatingPoint(boxx), 
    //         isImaginaryFloat(boxx), isComplexFloat(boxx), getDataSize(boxx)) ;

    writefln( "isbox.isFloatingPoint(in Box).unittest" );
    boxx = box(3.1416f);
    if (!isFloatingPoint(boxx))
        IncorrectReturnError("isFloatingPoint", boxx.toString(), boxx.type().toString()); 

    writefln( "isbox.isCharUTF8(in Box).unittest" );
    boxx = box('A');
    if (!isCharUTF8(boxx))
        IncorrectReturnError("isCharUTF8", boxx.toString(), boxx.type().toString());
    else if (getDataSize(boxx) != 1)
        throw new Exception("getDataSize(boxx) should be a length of 1.");
    else if (.toString(boxx) != "A"c)
        throw new Exception(".toString(boxx) should equal \"A\"");
        
    writefln( "isbox.isNull(in Box).unittest" );
    if (!isNull(boxy))
        writefln("isNull(boxy) != null, boxy=\"%s\"", ""); 

    boxy = box("12345678"c);  
    boxy = box(null);
    if (!isNull(boxy))
        IncorrectReturnError("isNull", boxy.toString(), boxy.type().toString());
    
    writefln( "isbox.isEmpty(in Box).unittest" );
    boxy = box(""c);
    if (!isEmpty(boxy))
        IncorrectReturnError("isEmpty", boxy.toString(), boxy.type().toString());

    writefln( "isbox.isStringUTF8(in Box).unittest" );
    boxy = box("ABC"c);
    if (!isStringUTF8(boxy))
        IncorrectReturnError("isStringUTF8", boxy.toString(), boxy.type().toString());

    boxz = boxArray(123U,456U,789U);
    //if (!isNormalArray(boxz))
    //    throw new Exception("isNormalArray() should've returned true.");
    if (!isUnsignedInteger(boxz[0]))
        IncorrectReturnError("isUnsignedInteger", boxz[0].toString(), boxz[0].type().toString());

    writefln( "isbox.boxArray(...).unittest" );
    boxz = boxArray(null,456L,789U);
    
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    writefln("isArray=", isArray(boxz));
    writefln("isAArray=", isAArray(boxz));
    dchar[] ds;
    writefln("isArray=", isArray(ds));
    writefln("isAArray=", isAArray(ds));
    int i;
    writefln("isArray=", isArray(i));
    writefln("isAArray=", isAArray(i));
    int[] iArray;
    iArray.length = 1;
    iArray[0] = -56;
    Box boxA = box(iArray);
    writefln("isArray=", isArray(boxA));
    //writefln("isAArray=", isAArray(i));
    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    if (!isNull(boxz[0]))
        IncorrectReturnError("isNull", boxz[0].toString(), boxz[0].type().toString());

    if (!isSignedInteger(boxz[1]))
        IncorrectReturnError("isSignedInteger", boxz[1].toString(), boxz[1].type().toString());

    if (!isUnsignedInteger(boxz[2]))
        IncorrectReturnError("isUnsignedInteger", boxz[2].toString(), boxz[2].type().toString());

    writefln();
    writefln("Compiled and tested with: \"%s v%d.%03d\"", scomp.name, scomp.version_major, scomp.version_minor);
    writefln("isbox.unittest finished!"c); 
    
    return 0;
}
}

// from apropos.meta by Andy Friesen 23.June.04
/** True for all array types. */
template IsArray  (T      ) {   const bool IsArray = false; }
template IsArray  (T : T[]) {   const bool IsArray = true;  }


bool isArray(...)
{
    TypeInfo ti     = _arguments[0]; 
    char[]   tiName = ti.classinfo.name;
    
    writefln("tiName=\"%s\", toString=%s", tiName, ti); 

    return (tiName[9] == 'A' || tiName == "TypeInfo_Array"c);
}

bool isBoxArray(in Box boxx)
{
    TypeInfo ti = boxx.type();

    // Defined, but never assigned anything
    if (ti is null)
        return false;
    else if (ti.classinfo.name == "TypeInfo_Array"c)
        return true;
    else
        return false;

    //return isArray(boxx);
}

//bool isAssociativeArray(...)
bool isAArray(...)
{
    TypeInfo ti     = _arguments[0];
    char[]   tiName = ti.classinfo.name;

    return (tiName == "TypeInfo_AssociativeArray"c);
}

/+
// These are not tested in any way...  
// may add them later if they're useful.
bool isNormalArray(in Box[] boxx)
{
    TypeInfo ti     = boxx[0].type();
    char[]   tiName = ti.classinfo.name;
    
    char[]   tiName = boxx.type.classinfo.name;
    writefln("tiName=\"%s\", toString=%s", tiName, boxx[].toString());

    return (tiName.length >= 10 && 
            tiName[9] == 'A' && 
            tiName != "TypeInfo_AssociativeArray");
}

bool isAssociativeArray(in Box[] boxx)
{
    TypeInfo ti     = boxx[0].type();
    char[]   tiName = ti.classinfo.name;

    return (tiName == "TypeInfo_AssociativeArray");
}
+/

/+
// Homegrown external toString() function 
// that uses / tests the new isFunc()s.
char[] toString(in Box boxx)
{
    // Empty
    if (isEmpty(boxx))
        return cast(char[])"";
    // Numerics    
    else if (isSignedInteger(boxx))
        return std.string.toString(unbox!(long)(boxx));
    else if (isUnsignedInteger(boxx))
        return std.string.toString(unbox!(ulong)(boxx));
    else if (isFloatingPoint(boxx))
        return std.string.toString(unbox!(real)(boxx));
    else if (isImaginaryFloat(boxx))
        return std.string.toString(unbox!(ireal)(boxx));
    else if (isComplexFloat(boxx))
        return std.string.toString(unbox!(creal)(boxx));
    // Other Strings    
    else if (isStringUTF16(boxx))
        return std.utf.toUTF8(unbox!(wchar[])(boxx));     
    else if (isStringUTF32(boxx))
        return std.utf.toUTF8(unbox!(dchar[])(boxx));     
    // Chars
    else if (isCharUTF8(boxx))
    {
    char[] s;
    s.length = s.length + 1;
        s[0] = unbox!(char)(boxx);
        return s;     
    }    
    else if (isCharUTF16(boxx))
    {
        wchar[] ws;
        ws.length = ws.length + 1;
        ws[0] = unbox!(wchar)(boxx);
        return std.utf.toUTF8(ws);     
    }    
    else if (isCharUTF32(boxx))
    {
        dchar[] ds;
        ds.length = ds.length + 1;
        ds[0] = unbox!(dchar)(boxx);
        return std.utf.toUTF8(ds);     
    }    
    // unbox and return   
    else if (isStringUTF8(boxx))
        return unbox!(char[])(boxx);  
    // Others     
    else
        return cast(char[])"";     

    return boxx.toString();
}
+/

