/+
 ' Module          : mathext.d - Addtional Math Functions
 ' Version         : v1.0  
 ' Author          : David L. 'SpottedTiger' Davis
 ' Source Platform : Intel x86 / WinXP SP2 
 ' Target Platform : Should work on any CPU / OS combo that D compiles on.
 '                 :
 ' Created Date    : 03.Feb.05 Compiled and Tested with dmd v0.111.
 ' Modified Date   : 20.Mar.05 Recompiled and Tested with dmd v0.119.
 '                 : 10.Jun.06 Recompiled and Tested with dmd v0.126.
 '                 : 17.Sep.06 Recompiled and Tested with dmd v0.131.
 '                 : 26.Oct.05 Recompiled and Tested with dmd v0.137.
 '                 : 30.Dec.05 Recompiled and Tested with dmd v0.142.
 '                 : 19.Mar.06 Recompiled and Tested with dmd v0.150.
 '                 : 28.May.06 Recompiled and Tested with dmd v0.159.
 '                 : 19.Jul.06 All "Imports" are now private by default  
 '                 :           (which shouldn't be a problem, since I  
 '                 :           always use the private keyword) in 
 '                 :           dmd v0.163, recompiled and tested.
 '                 : 23.Oct.06 Recompiled and Tested with dmd v0.172.
 '                 : 05.Nov.06 Recompiled and Tested with dmd v0.173.
 '                 : 09.Dec.06 Recompiled and Tested with dmd v0.177.
 '                 : 09.Feb.07 Recompiled and Tested with dmd v1.005. 
 '                 : 03.Jan.08 Recompiled and Tested with dmd v1.025.
 '                 : 20.Mar.08 Recompiled and Tested with dmd v1.028. 
 '                 : 19.Jun.08 Recompiled and Tested with dmd v1.031.  
 '                 : 17.Jul.08 Recompiled and Tested with dmd v1.033.  
 '                 : 13.May.09 Recompiled and Tested with dmd v1.045.  
 '                 :
 ' Requirements    : std.math, and (for debug only) std.stdio.
 ' Licence         : Public Domain
 ' ---------------------------------------------------------------------------
 '
 ' To complie for a unittest:
 ' dmd mathext.d -debug=mathext -unittest
 '
 ' To compile a D program that use mathext.d:
 ' dmd MySource.d mathext.d
 '
 ' Public functions:
 ' ------------------------------
 ' 01) bool isprime( in ulong )
 ' 02) bool isprimepower( in ulong )
 ' 03) ulong gcd( in ulong, in ulong )
 ' 04) ulong factorial( in ulong ) 
 ' 05) ulong fib( in ulong )
 ' 06) ulong gcf( in ulong, in ulong )
 ' 07) ulong lcm( in ulong, in ulong )
 +/
module mathext;

debug( mathext ) private import scomp = std.compiler; 
private import std.math;
debug( mathext ) private import std.stdio;

// Greatest Common Divisor of ula, ulb using the Euclidean algorithm
ulong gcd(in ulong ula, in ulong ulb)
{
  ulong ulc;
  
  while (ula % ulb != 0UL)
  {
      ulc = ula % ulb;
      ula = ulb;  
      ulb = ulc;
  }
  
  return ulb;
}

bool isprime(in ulong ulz) 
{
    if (ulz <= 1UL)
        return false;

    for (ulong ulx = 2UL; ulx <= std.math.floor(std.math.sqrt(cast(double)ulz)); ulx++)
        if ( ( ulz % ulx ) == 0UL ) 
            return false;
            
    return true;
}

bool isprimepower(in ulong ulz) 
{
    ulong ulx = 2UL;
    ulong uly;
    
    while (ulx <= std.math.floor(std.math.sqrt(cast(double)ulz)) && uly == 0UL) 
    {
        if ((ulz % ulx) == 0UL) 
            uly = ulx;
            
        ulx++;
    }

    for (ulx = 2UL; ulx <= std.math.floor(std.math.log(ulz) / std.math.log(uly)); ++ulx) 
        if (std.math.pow(cast(real)uly, cast(real)ulx) == ulz) 
            return true;

    return false;
}

// Recursively computing n!
ulong factorial(in ulong ulFactor)
{
    // Can handle only up to the 20th factorial"
    if (ulFactor > 20UL)
        throw new Exception( "factorial() limitation, input parameter used can't be greater-than 20."c );

    return cast(ulong)(ulFactor <= 1U) ? 1U : ulFactor * factorial(ulFactor - 1U);
}

// fib() returns the n'th Fibonacci number
ulong fib(in ulong ulNth)
{
    ulong ula = 1UL;
    ulong ulb = 1UL;
    ulong ulc = 0UL;
    
    if (ulNth == 0UL)
        return 0UL;
        
    // Can handle only up to the 94th Fibonacci number"
    if (ulNth > 94UL)
        throw new Exception( "fib() limitation, input parameter used can't be greater-than 94."c );
    
    for (ulong ulx = 3UL; ulx <= ulNth; ulx++)
    {
        ulc = ula + ulb;
        ula = ulb;
        ulb = ulc;
    }
    
    return ulb;
}

// finds the Greatest Common Factor between 2 arguments
ulong gcf(in ulong ulA, in ulong ulB)
{
    ulong ulHigh = 1UL;
  
    for (ulong ulX = 2UL; ulX <= ulA; ulX++)
        if ((ulA % ulX == 0UL) && (ulB % ulX == 0UL))
            ulHigh = ulX;
    
    return ulHigh;
} 

// The least common multiple (LCM) of two numbers is the 
// smallest number (not zero) that is a multiple of both.
ulong lcm(in ulong ulA, in ulong ulB)
{
    ulong ulHigh;
    ulong ulLow;
    ulong ulInc;
  
    if (ulA > ulB)
    {
        ulHigh = ulA;
        ulLow  = ulB;
    }
    else
    {
        ulHigh = ulB;
        ulLow  = ulA;
    }
  
    ulInc = ulHigh;
  
    while (ulHigh % ulLow != 0UL)
        ulHigh += ulInc;
      
    return ulHigh;
} 

unittest
{
    writefln( "mathext.isprimepower( in ulong ).unittest"c );
    assert( isprimepower(25 ^ 2) == 1 );

    writefln( "mathext.isprime( in ulong ).unittest"c );
    assert( isprime(571) == true );
    assert( isprime(25) == false );

    writefln( "mathext.gcd( in ulong, in ulong ).unittest"c );
    assert( gcd(256, 512) == 256 );

    writefln( "mathext.factorial( in ulong ).unittest"c );
    assert( factorial(20) == 2432902008176640000 );

    writefln( "mathext.fib( in ulong ).unittest"c );
    assert( fib(18) == 2584 );
    assert( fib(94) == 1293530146158671551 );

    writefln( "mathext.gcf( in ulong, in ulong ).unittest"c );
    assert( gcf(2940, 3150) == 210 );

    writefln( "mathext.lcm( in ulong, in ulong ).unittest"c );
    assert( lcm(2940, 3150) == 44100 );
}

debug( mathext )
{
int main()
{
    writefln( "mathext.unittest() done!"c );
    writefln();

    writefln( "fib(0)="c, fib(0), ", ans=0"c );
    assert( fib(0) == 0 );

    writefln( "fib(1)="c, fib(1), ", ans=1"c );
    assert( fib(1) == 1 );

    writefln( "fib(2)="c, fib(2), ", ans=1"c );
    assert( fib(2) == 1 );

    writefln( "fib(3)="c, fib(3), ", ans=2"c );
    assert( fib(3) == 2 );

    writefln( "fib(16)="c, fib(16), ", ans=987"c );
    assert( fib(16) == 987 );

    writefln( "fib(18)="c, fib(18), ", ans=2584"c );
    assert( fib(18) == 2584 );

    writefln( "fib(46)="c, fib(46), ", ans=1836311903"c );
    assert( fib(46) == 1836311903 );

    writefln( "fib(94)="c, fib(94), ", ans=1293530146158671551"c );
    assert( fib(94) == 1293530146158671551 );
    //writefln( "fib(95)=", fib(95) ); // beyond max...an error

    writefln( "isprimepower(25 ^ 2)=%b, ans=1"c, isprimepower(25 ^ 2) );
    assert( isprimepower(25 ^ 2) == 1 );

    writefln( "isprime(25)=%b, ans=0"c, isprime(25) );
    assert( isprime(25) == 0 );

    writefln( "isprime(2)=%b, ans=1"c, isprime(2) );
    assert( isprime(2) == 1 );

    writefln( "gcd(256, 512)=%d, ans=256"c, gcd(256, 512) );
    assert( gcd(256, 512) == 256 );

    writefln( "factorial(4)=%d, ans=24"c, factorial(4) );
    assert( factorial(4) == 24 );

    writefln( "factorial(12)=%d, ans=479001600"c, factorial(12) );
    assert( factorial(12) == 479001600 );

    writefln( "factorial(20)=%d, ans=2432902008176640000"c, factorial(20) );
    assert( factorial(20) == 2432902008176640000 );
    //writefln( "factorial( 21 )=%d, ans=51090942171709440000", factorial(21) );

    writefln( "lcm(3, 4)=%d, ans=12"c, lcm(3, 4) );
    assert( lcm(3, 4) == 12 );

    writefln( "lcm(2940, 3150)=%d, ans=44100"c, lcm(2940, 3150) );
    assert( lcm(2940, 3150) == 44100 );

    writefln( "gcf(2940, 3150)=%d, ans=210"c, gcf(2940, 3150) );
    assert( gcf(2940, 3150) == 210 );

    writefln();
    writefln("Compiled and tested with: \"%s v%d.%03d\"", scomp.name, scomp.version_major, scomp.version_minor);
    writefln("mathext.unittest finished!"c); 

    return 0;
}
}