/+
 ' File / Module   : stringw.d - uses UTF16 as the base character internally
 ' Version         : v1.0 
 ' Author(s)       : David L. 'SpottedTiger' Davis
 '                 : Walter Bright's std.string.d used as the foundation
 ' Source Platform : Intel x86 / WinXP SP2 
 ' Target Platform : Should work on any CPU / OS combo that D compiles on.
 '                 :
 ' Complied        : DigitalMars "D" aka Mars v0.101
 ' Created Date    : 03.Sep.04 Complied and Tested with dmd v0.101
 ' Modified Date   : 04.Sep.04 Added unittests for every function.
 '                 : 07.Sep.04 Added 12 functions.
 '                 : 11.Sep.04 Added 4 functions.
 '                 : 05.Oct.04 Complied and Tested with dmd v0.102, after
 '                 :           adding 21 toString() functions from std.string
 '                 :           and also added / created unittests.
 '                 : 14.Oct.04 Added the toSoundexW() function. 
 '                 : 31.Oct.04 For completeness with std.string I've added the
 '                 :           last 9 missing functions with little change, other
 '                 :           than to convert the char[]s to wchar[]s... and then
 '                 :           Complied and Tested with dmd v0.105.
 '                 : 24.Nov.04 Added the isDigitsW() and addOrdSuffixW() functions,
 '                 :           Recomplied and Tested with dmd v0.106.
 '                 : 30.Nov.04 Recomplied and Tested with dmd v0.108, correct a few
 '                 :           minor errors...the D compiler is checking the code 
 '                 :           even better than before. Thanks Walter! :))
 '                 : 22.Jan.05 Added / converted char[] to wchar[] in the two new toString() 
 '                 :           functions that Walter included into std.string for dmd v0.111.
 '                 :           Recomplied and Tested with dmd v0.111.
 '                 : 20.Mar.05 Recomplied and Tested with dmd v0.119.
 '                 : 10.Jun.05 Recomplied and Tested with dmd v0.126, and fixed a few minor issues.
 '                 : 10.Sep.05 Updated the doFormatW() function to use std.utf.decode(wchar[]),
 '                 :           and recomplied and tested with dmd v0.131.
 '                 : 26.Oct.05 Recompiled and tested with dmd v0.137.    
 '                 : 30.Dec.05 Recompiled and tested with dmd v0.142.  
 '                 : 19.Mar.06 Recompiled and tested with dmd v0.150. 
 '                 : 28.May.06 Corrected lowercase 'l' suffix which is now deprecated, with the 
 '                 :           uppercase 'L', then recompiled and tested with dmd v0.159.
 '                 : 19.Jul.06 All "Imports" are now private by default (which shouldn't 
 '                 :           be a problem, since I always use the private keyword) in 
 '                 :           dmd v0.163, recompiled and tested.
 '                 : 23.Oct.06 Recompiled and tested with dmd v0.172. 
 '                 : 23.Oct.06 Corrected one "shadowing declaration" error,
 '                 :           then compiled and tested with dmd v0.172.
 '                 : 05.Nov.06 Recompiled and tested with dmd v0.173.
 '                 : 09.Dec.06 Recompiled and tested with dmd v0.177.
 '                 : 05.Jan.07 Recompiled and tested with dmd v1.0.
 '                 : 09.Feb.07 Recompiled and tested with dmd v1.005.
 '                 : 03.Jan.08 Recompiled and tested with dmd v1.025. 
 '                 : 20.Mar.08 Recompiled and tested with dmd v1.028.  
 '                 : 19.Jun.08 Recompiled and tested with dmd v1.031. 
 '                 : 17.Jul.08 Recompiled and tested with dmd v1.033. 
 '                 : 13.May.09 Recompiled and tested with dmd v1.045. 
 '                 :
 '                 : Darn it...I just had to add a "W" suffix to the 
 '                 : function names so these functions could play in the
 '                 : same sandbox as Phobos' std.string functions.
 '                 :
 ' Requirements    : Must complie with dmd v0.92+ and above, because
 '                 : default parameters are being used, also std.format 
 '                 : needs to be imported for the doFormatW() function, and
 '                 : std.ctype for the tolowerW( dchar ) / toupperW( dchar )
 '                 : functions.
 '                 :
 ' Licence         : Public Domain
 ' ---------------------------------------------------------------------------
 ' Note: This module is meant to extend Phobos' std.string functions with support 
 '       for wchar[] character strings as the base. Also though Phobos' char[] functions  
 '       were used as the source for these functions, many of these functions have been
 '       rewritten, but all of the function descriptions and most of unittest were 
 '       copied as is. Inwhich case, these functions may not perform as well as code  
 '       written by Walter himself. Over time I hope to add more functions, and to better  
 '       fine tune the existing ones more.
 '
 ' Note2: In dmd v0.114 (27.Feb.05) Walter added the functions tr, succ, squeeze, removechars, 
 '        countchars, inPattern, chop, and chomp to std.string. Currently, I haven't decided 
 '        to add them to stringw.d, but maybe in some future date.
 '
 ' To complie for a unittest:
 ' dmd stringw.d -debug=stringW -unittest
 '
 ' To compile a D program that use stringw.d:
 ' dmd MySource.d stringw.d
 '
 ' Wide-Character (16-bit) "Non-Case Sensitive" functions:
 ' -------------------------------------------------------
 ' 01) int ifindW( in wchar[], in wchar )
 ' 02) int ifindW( in wchar[], in wchar[] )
 ' 03) int irfindW( in wchar[], in wchar )
 ' 04) int irfindW( in wchar[], in wchar[] )
 ' 05) int icmpW( in wchar[], in wchar[] ) 
 ' 06) wchar[] ireplaceW( in wchar[], in wchar[], in wchar[] )
 ' 07) int icountW( in wchar[], in wchar[] )
 ' 08) int ifindxW( in wchar[], in wchar[], in int = 0 )
 ' 09) int irfindxW( in wchar[], in wchar[], in int = 0 )
 ' 10) wchar[][] isplitW( in wchar[], in wchar[] )
 '
 ' Wide-Character (16-bit) "Case Sensitive" functions:
 ' ---------------------------------------------------
 ' 01) int findW( in wchar[], in wchar )
 ' 02) int findW( in wchar[], in wchar[] )
 ' 03) int rfindW( in wchar[], in wchar )
 ' 04) int rfindW( in wchar[], in wchar[] )
 ' 05) int cmpW( in wchar[], in wchar[] )
 ' 06) wchar[] replaceW( in wchar[], in wchar[], in wchar[] )
 ' 07) int countW( in wchar[], in wchar[] 
 ' 08) int findxW( in wchar[], in wchar[], in int = 0 )
 ' 09) int rfindxW( in wchar[], in wchar[], in int = 0 )
 ' 10) wchar[][] splitW( in wchar[], in wchar[] )
 '
 ' Wide-Character (16-bit) other functions:
 ' ----------------------------------------
 ' 01) bool isRomanNumeralW( in wchar )
 ' 02) bool isRomanNumeralW( in wchar[] )
 ' 03) wchar[] tolowerW( in wchar[] )
 ' 04) wchar[] toupperW( in wchar[] )
 ' 05) wchar tolowerW( in dchar )
 ' 06) wchar toupperW( in dchar )
 ' 07) wchar[] capitalizeW( wchar[] )
 ' 08) wchar[] striplW( in wchar[] )
 ' 09) wchar[] striprW( in wchar[] )
 ' 10) wchar[] stripW( in wchar[] )
 ' 11) bool isASCIIStringW( in wchar[], in bool = false )
 ' 12) wchar[] asciiProperCaseW( in wchar[] )
 ' 13) wchar[] rjustifyW( in wchar[], in int, in wchar = '\u0020' )
 ' 14) wchar[] zfillW( in wchar[], in int )
 ' 15) wchar[] ljustifyW( in wchar[], in int, in wchar = '\u0020' )
 ' 16) wchar[] centerW( in wchar[], in int, in wchar = '\u0020' )
 ' 17) wchar[] doFormatW( ... )
 ' 18) wchar[] toSoundexW( in wchar[] )
 ' 19) bool isDigitsW( in wchar[] )
 ' 20) wchar[] addOrdSuffixW( in ulong )
 '
 ' --------------------------------------------------------------------------
 ' Below functions are copies of Walter's std.string code and have only been  
 ' converted from char[] to wchar[], none thus far have been rewritten.
 ' --------------------------------------------------------------------------
 ' 21) wchar[] capwordsW( in wchar[] )
 ' 22) wchar[] joinW( in wchar[][], in wchar[] )
 ' 23) wchar[][] splitW( in wchar[] )
 ' 24) wchar[][] splitlinesW( in wchar[] )
 ' 25) wchar[] replaceSliceW( in wchar[], in wchar[], in wchar[] )
 ' 26) wchar[] insertW( in wchar[], in int, in wchar[] )
 ' 27) wchar[] expandtabsW( in wchar[], in int )
 ' 28) wchar[] maketransW( in wchar[], in wchar[] )
 ' 29) wchar[] translateW( in wchar[], in wchar[], in wchar[] )
 '
 ' Wide-Character version of std.string's toString() conversion functions:
 ' -----------------------------------------------------------------------
 ' 01) wchar[] toStringW( in bit )
 ' 02) wchar[] toStringW( in wchar )
 ' 03) wchar[] toStringW( in ubyte )
 ' 04) wchar[] toStringW( in ushort )
 ' 05) wchar[] toStringW( in uint )
 ' 06) wchar[] toStringW( in ulong )
 ' 07) wchar[] toStringW( in byte )
 ' 08) wchar[] toStringW( in short )
 ' 09) wchar[] toStringW( in int )
 ' 10) wchar[] toStringW( in long )
 ' 11) wchar[] toStringW( in float )
 ' 12) wchar[] toStringW( in double )
 ' 13) wchar[] toStringW( in real )   
 ' 14) wchar[] toStringW( in ifloat )
 ' 15) wchar[] toStringW( in idouble )
 ' 16) wchar[] toStringW( in ireal )   
 ' 17) wchar[] toStringW( in cfloat )
 ' 18) wchar[] toStringW( in cdouble )
 ' 19) wchar[] toStringW( in creal )  
 ' 20) wchar[] toStringW( in wchar * )
 ' 21) wchar[] toStringzW( in wchar[] )
 ' -------------------------------------
 ' 22) wchar[] toStringW( in long, in uint )
 ' 23) wchar[] toStringW( in ulong, in uint )
 '
 ' Private Functions:
 ' ------------------
 ' 01) uint strlenW( in wchar * )
 +/
module stringw;

debug( stringW ) private import scomp = std.compiler;
private import std.ctype;  // Needed for tolowerW( dchar ) and toupperW( dchar )
private import std.format; // Needed for the doFormatW() function.
private import std.utf;    // Needed for decode( wchar[] )
debug( stringW ) private import std.stdio;  // for writefln()
//private import std.stdio;

const wchar[16] HEXDIGITS_W  = "0123456789ABCDEF"w;
const wchar[10] DIGITS_W     = "0123456789"w;
const wchar[8]  OCTDIGITS_W  = "01234567"w;
const wchar[]   WHITESPACE_W = "\u0020\u0009\u000B\u000D\u000A\u000C"w; // "\x20\t\v\r\n\f";
const wchar[14] ROMAN_W      = "IVXLCDMivxlcdm"w;

/+
 ' Gathered the 1000 most commonly used English words 
 ' together from the web, and then got a character count 
 ' on each. Which was then used to reordered the characters 
 ' from the most used, to the least used.
 '
 ' e=638, t=385, a=370, r=369, o=354, n=300,
 ' s=290, i=280, l=270, c=197, h=182, d=175,
 ' u=149, p=143, m=126, g=118, w=099, f=088,
 ' b=087, y=081, v=052, k=051, x=014, q=011,
 ' j=008, z=001
 '
 +/
const wchar[ 26 ] LOWERCASE  = "etaronsilchdupmgwfbyvkxqjz"w;
const wchar[ 26 ] UPPERCASE  = "ETARONSILCHDUPMGWFBYVKXQJZ"w; 

//debug = stringW;    // uncomment to turn on debugging writefln()'s

debug( stringW )
{
int main()
{
	writefln();
    writefln("Compiled and tested with: \"%s v%d.%03d\"", scomp.name, scomp.version_major, scomp.version_minor);
    writefln("stringw.unittest finished!"c); 

    return 0;
    
} // end int main()
} // end debug( stringW )

wchar toupperW(in dchar dc)
{
    return std.ctype.toupper(dc);
    
} // end wchar toupperW( in dchar )

unittest
{
    debug( stringW ) writefln( "stringw.toupperW( in dchar ).unittest"w );

    wchar wc1 = 'w';
    wchar wc2;

    wc2 = toupperW(wc1);
    assert( wc1 != wc2 );
    
    assert( toupperW('y') == 'Y' );
    assert( toupperW('\n') == '\n' );
}

wchar tolowerW(in dchar dc)
{
    return std.ctype.tolower(dc);
    
} // end wchar tolowerW( in dchar )

unittest
{
    debug( stringW ) writefln( "stringw.tolowerW( in dchar ).unittest"w );

    wchar wc1 = 'F';
    wchar wc2;

    wc2 = tolowerW(wc1);
    assert( wc1 != wc2 );
    
    assert( tolowerW('M') == 'm' );
    assert( tolowerW('\n') == '\n' );
}

/+
 ' int icmpW( in wchar[], in wchar[] )
 '
 ' returns: 
 '   -2 if either string is an empty string
 '   -2 if the compare string's length is greater than the source string
 '   -1 if source string is less than compare string
 '    0 if both are equal to one another
 '    1 if the source string is greater then the compare string
 '
 ' Note: Is Non-Case Sensitive.
 +/
int icmpW(in wchar[] ws1, in wchar[] ws2)
{
    wchar wc1;
    wchar wc2;
    uint  uiWS1Len = ws1.length;
    uint  uiWS2Len = ws2.length;
    
    if ((uiWS1Len == 0  || uiWS2Len == 0) || uiWS2Len > uiWS1Len) 
        return -2;
        
    if (uiWS2Len < uiWS1Len) 
        uiWS1Len = uiWS1Len;

    for (int ix = 0; ix < uiWS1Len; ix++)
    {
        wc1 = ws1[ix];
        wc2 = ws2[ix];
         
        // If ASCII, convert to lowercase 
        wc1 = (wc1 >= '\u0041' && wc1 <= '\u005A' ? wc1 + '\u0020' : wc1);
        wc2 = (wc2 >= '\u0041' && wc2 <= '\u005A' ? wc2 + '\u0020' : wc2);
        
        if (wc1 > wc2) 
            return 1;

        if (wc1 < wc2) 
            return -1; 
    }

    return 0;
    
} // end int icmpW( in wchar[], in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.icmpW( in wchar[], in wchar[] ).unittest"w );
    
    int i;
    
    // "text" == "text"
    i = icmpW("TexT"w, "text"w);
    assert( i == 0 );
    
    // "text20" larger then "text"
    i = icmpW("TexT"w, "text20"w);
    assert( i == -2 );
    
    // One or both or empty strings
    i = icmpW("TexT"w, ""w);
    assert( i == -2 );
    
    // "text" > "test"
    i = icmpW("TexT"w, "tESt"w);
    assert( i == 1 );
    
    // "test" < "text"
    i = icmpW("TeST"w, "tExt"w);
    assert( i == -1 );
}

/+
 ' int iwildcmpW( in wchar[], in wchar[] )
 '
 ' Note: Work-In-Process, it needs more work.
 +/
int iwildcmpW(in wchar[] ws1, in wchar[] ws2)
{
    wchar wc1;
    wchar wc2;
    uint  uiWS1Len = ws1.length;
    uint  uiWS2Len = ws2.length;
    int   iy = 0;
    int   iz = 0;    

    if (uiWS1Len == 0  || uiWS2Len == 0) 
        return -2;

    if (uiWS2Len > uiWS1Len) 
        return -2;
    
    if (uiWS2Len < uiWS1Len)
        return -2;    

    if (ws1[0] == '*')
    {
        iz++;
        iy = ws2.length - (ws1.length - 1); 
    }

    for (int ix = iy; ix < uiWS1Len; ix++)
    {
        wc1 = ws1[iz];
        
        iz++;

        if (wc1 == '*')
            return 0; //break;
        
        if (wc1 == '?')
            continue;
            
        wc2 = ws2[ix];
         
        // If ASCII, convert to lowercase 
        wc1 = (wc1 >= '\u0041' && wc1 <= '\u005A' ? wc1 + '\u0020' : wc1);
        wc2 = (wc2 >= '\u0041' && wc2 <= '\u005A' ? wc2 + '\u0020' : wc2);
        
        if (wc1 > wc2) 
            return 1;
            
        if (wc1 < wc2) 
            return -1; 
    }

    return 0;
    
} // end int iwildcmpW( in wchar[], in wchar[] )

/+
 ' int cmpW( in wchar[], in wchar[] )
 '
 ' returns: 
 '   -2 if either string is an empty string
 '   -2 if the compare string's length is greater then the source string
 '   -1 if source string is less than compare string
 '    0 if both are equal to one another
 '    1 if the source string is greater then the compare string
 '
 ' Note: Is Case Sensitive.
 +/
int cmpW(in wchar[] ws1, in wchar[] ws2)
{
    wchar wc1;
    wchar wc2;
    uint  uiWS1Len = ws1.length;
    uint  uiWS2Len = ws2.length;
    
    if ((uiWS1Len == 0  || uiWS2Len == 0) || uiWS2Len > uiWS1Len) 
        return -2;
        
    if (uiWS2Len < uiWS1Len) 
        uiWS1Len = uiWS1Len;

    // Convert to lowercase if character is in ASCII
    for (int ix = 0; ix < uiWS1Len; ix++)
    {
        wc1 = ws1[ix];
        wc2 = ws2[ix];
        
        if (wc1 > wc2) 
           return 1;
           
        if (wc1 < wc2) 
            return -1; 
    }

    return 0;
    
} // end int cmpW( in wchar[], in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.cmpW( in wchar[], in wchar[] ).unittest"w );
    
    int i;
    
    // "text" == "text"
    i = cmpW("Text"w, "Text"w);
    assert( i == 0 );
    
    // "text20" larger then "text"
    i = cmpW("texT"w, "texT20"w);
    assert( i == -2 );
    
    // One or both or empty strings
    i = cmpW("text"w, ""w);
    assert( i == -2 );
    
    // "text" > "test"
    i = cmpW("text"w, "test"w);
    assert( i == 1 );
    
    // "test" < "text"
    i = cmpW("tesT"w, "texT"w);
    assert( i == -1 );
}

/+
 ' int ifindW( in wchar[], in wchar )
 '
 ' Find first occurrance of wc in string ws.
 ' Return index in ws where it is found.
 ' Return -1 if not found.
 '
 ' Note: Is Non-Case Sensitive.
 +/
int ifindW(in wchar[] ws, in wchar wc)
{
    wchar wc1 = wc;
    wchar wc2;

    // If ASCII, convert to lowercase
    wc1 = (wc1 >= '\u0041' && wc1 <= '\u005A' ? wc1 + '\u0020' : wc1);

    for (int i = 0; i  < ws.length; i++)
    {   
        wc2 = ws[i]; 
        
        // If ASCII, convert to lowercase
        wc2 = (wc2 >= '\u0041' && wc2 <= '\u005A' ? wc2 + '\u0020' : wc2);
        if (wc1 == wc2) 
            return i;   
    }

    return -1;
    
} // end ifindW( in wchar[], in wchar )

unittest
{
    debug( stringW ) writefln( "stringw.ifindW( in wchar[], in wchar ).unittest"w );

    wchar[] wsPlts = "Mars: the fourth Rock (Planet) from the Sun."w;
    int     i;

    i = ifindW(null, 'a');
    assert( i == -1 );
    i = ifindW("def"w, 'a');
    assert( i == -1 );
    i = ifindW("abba"w, 'a');
    assert( i == 0 );
    i = ifindW("def"w, 'f');
    assert( i == 2 );

    i = ifindW(wsPlts, 'P');
    assert( i == 23 );
    i = ifindW(wsPlts, 'R');
    assert( i == 2 );
}

/+
 ' int ifindW( in wchar[], in wchar[] )
 '
 ' Find first occurrance of ws2[] in string ws1[].
 ' Return index in ws1[] where it is found.
 ' Return -1 if not found.
 '
 ' Note: Is Non-Case Sensitive.
 +/
int ifindW(in wchar[] ws1, in wchar[] ws2)
{    
    wchar wc;
    uint  uiWS1Len = ws1.length;
    uint  uiWS2Len = ws2.length;
    int   ip = -1; 
        
    if (uiWS2Len == 0 || uiWS1Len == 0 || uiWS2Len > uiWS1Len)  
            return -1; 

    if (uiWS2Len == 1)
    {
        wc = ws2[0];
        return ifindW(ws1, wc);
    }
    else
    {
        wc = ws2[0];
        ip = ifindW(ws1, wc);

        if (ip == -1 || (ip + uiWS2Len > uiWS1Len))  
            return -1;
                
        else
        {
            for (int x = ip; x < (uiWS1Len - uiWS2Len) + 1; x++)
                if (icmpW(ws1[ x .. (x + uiWS2Len)], ws2) == 0) 
                    return x;
        }
    }

    return -1;

} // end ifindW( in wchar[], in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.ifindW( in wchar[], in wchar[] ).unittest"w );
    
    wchar[] wsPlts = "Mars: the fourth Rock (Planet) from the Sun."w;
    wchar[] wsMars = "Who\'s \'My Favorite Maritian?\'"w;
    int     i;

    i = ifindW(null, "a"w);
    assert( i == -1 );
    i = ifindW("def"w, "a"w);
    assert( i == -1 );
    i = ifindW("abba"w, "a"w);
    assert( i == 0 );
    i = ifindW("def"w, "f"w);
    assert( i == 2 );
    i = ifindW("dfefffg"w, "fff"w);
    assert( i == 3 );
    i = ifindW("dfeffgfff"w, "fff"w);
    assert( i == 6 );
    
    i = ifindW(wsMars, "MY fAVe"w);
    assert( i == -1 );
    i = ifindW(wsMars, "mY fAVOriTe"w);
    assert( i == 7 );
    i = ifindW(wsPlts, "mArS:"w);
    assert( i == 0 );
    i = ifindW(wsPlts, "rOcK"w);
    assert( i == 17 );
    i = ifindW(wsPlts, "Un."w);
    assert( i == 41 );
    i = ifindW(wsPlts, wsPlts);
    assert( i == 0 );
}

/+
 ' int irfindW( in wchar[], in wchar )
 '
 ' Find last occurrance of wc in string ws.
 ' Return index in ws where it is found.
 ' Return -1 if not found.
 '
 ' Note: Is Non-Case Sensitive.
 +/
int irfindW(in wchar[] ws, in wchar wc)
{
    wchar wc1 = wc;
    wchar wc2;

    // If ASCII, convert to lowercase
    wc1 = (wc1 >= '\u0041' && wc1 <= '\u005A' ? wc1 + '\u0020' : wc1);

    for (int i = ws.length - 1; i >= 0; i--)
    {   
        wc2 = ws[i]; 
        
        // If ASCII, convert to lowercase
        wc2 = (wc2 >= '\u0041' && wc2 <= '\u005A' ? wc2 + '\u0020' : wc2);
        
        if (wc1 == wc2) 
            return i;   
    }

    return -1;
    
} // end irfindW( in wchar[], in wchar )

unittest
{
    debug( stringW ) writefln( "stringw.irfindW( in wchar[], in wchar ).unittest"w );

    wchar[] wsPlts = "Mars: the fourth Rock (Planet) from the Sun."w;
    int     i;

    i = irfindW(null, cast(wchar)'a');
    assert( i == -1 );
    i = irfindW("def"w, cast(wchar)'a');
    assert( i == -1 );
    i = irfindW("abba"w, cast(wchar)'a');
    assert( i == 3 );
    i = irfindW("def"w, cast(wchar)'f');
    assert( i == 2 );

    i = irfindW(wsPlts, cast(wchar)'M');
    assert( i == 34 );
    i = irfindW(wsPlts, cast(wchar)'S');
    assert( i == 40 );
}

/+
 ' int irfindW( in wchar[], in wchar[] )
 '
 ' Find last occurrance of ws2[] in string ws1[].
 ' Return index in ws1[] where it is found.
 ' Return -1 if not found.
 '
 ' Note: Is Non-Case Sensitive.
 +/
int irfindW(in wchar[] ws1, in wchar[] ws2)
{    
    wchar wc;
    uint  uiWS1Len = ws1.length;
    uint  uiWS2Len = ws2.length;
    int   ip = -1; 

    if (uiWS2Len == 0 || uiWS1Len == 0 || uiWS2Len > uiWS1Len)  
            return -1; 

    if (uiWS2Len == 1)
    {
        wc = ws2[0];
        return irfindW(ws1, wc);
    }
    else
    {
        wc = ws2[0];
        ip = irfindW(ws1, wc);

        if (ip == -1) 
            return -1;
            
        else
        {
            for (int ix = ip; ix >= 0; ix--)
            {
                if (ix + uiWS2Len > uiWS1Len) 
                    continue;

                if (icmpW(ws1[ix .. (ix + uiWS2Len)], ws2) == 0) 
                    return ix;
            }        
        }
    }

    return -1;

} // end irfindW( in wchar[], in wchar[] )

unittest
{
    debug( stringW ) writefln("stringw.irfindW( in wchar[], in wchar[] ).unittest"w );

    wchar[] wsPlts = "Mars: the fourth Rock (Planet) from the Sun."w;
    wchar[] wsMars = "Who\'s \'My Favorite Maritian?\'"w;
    int     i;
    
    i = irfindW("abcdefcdef"w, "c"w);
    assert( i == 6 );
    i = irfindW("abcdefcdef"w, "cd"w);
    assert( i == 6 );
    i = irfindW("abcdefcdef"w, "x"w);
    assert( i == -1 );
    i = irfindW("abcdefcdef"w, "xy"w);
    assert( i == -1 );
    i = irfindW("abcdefcdef"w, ""w);
    assert( i == -1 );
    i = irfindW("abcdefcdef"w, "def"w);
    assert( i == 7 );
    
    i = irfindW(wsMars, "RiTE maR"w);
    assert( i == 14 );
    i = irfindW(wsPlts, "FOuRTh"w);
    assert( i == 10 );
    i = irfindW(wsMars, "whO\'s \'MY"w);
    assert( i == 0 );
    i = irfindW(wsMars, wsMars);
    assert( i == 0 );
}    

/+
 ' int findW( in wchar[], in wchar )
 '
 ' Find first occurrance of wc in string ws.
 ' Return index in ws where it is found.
 ' Return -1 if not found.
 '
 ' Note: Is Case Sensitive.
 +/
int findW(in wchar[] ws, in wchar wc)
{
    wchar wc1 = wc;
    wchar wc2;

    for (int i = 0; i  < ws.length; i++)
    {   
        wc2 = ws[i]; 
        
        if (wc1 == wc2) 
            return i;   
    }

    return -1;
    
} // end findW( in wchar[], in wchar )

unittest
{
    debug( stringW ) writefln( "stringw.findW( in wchar[], in wchar ).unittest"w );

    wchar[] wsPlts = "Mars: the fourth Rock (Planet) from the Sun."w;
    int     i;

    i = findW(null, cast(wchar)'a');
    assert( i == -1 );
    i = findW("def"w, cast(wchar)'a');
    assert( i == -1 );
    i = findW("abba"w, cast(wchar)'a');
    assert( i == 0 );
    i = findW("def"w, cast(wchar)'f');
    assert( i == 2 );

    i = findW(wsPlts, cast(wchar)'P');
    assert( i == 23 );
    i = findW(wsPlts, cast(wchar)'R');
    assert( i == 17 );
}

/+
 ' int findW( in wchar[], in wchar[] )
 '
 ' Find first occurrance of ws2[] in string ws1[].
 ' Return index in ws1[] where it is found.
 ' Return -1 if not found.
 '
 ' Note: Is Case Sensitive.
 +/
int findW(in wchar[] ws1, in wchar[] ws2)
{    
    wchar wc;
    uint  uiWS1Len = ws1.length;
    uint  uiWS2Len = ws2.length;
    int   ip = -1; 
        
    if (uiWS2Len == 0 || uiWS1Len == 0 || uiWS2Len > uiWS1Len)  
            return -1; 

    if (uiWS2Len == 1)
    {
        wc = ws2[0];
        return findW(ws1, wc);
    }
    else
    {
        wc = ws2[0];
        ip = findW(ws1, wc);

        if (ip == -1 || (ip + uiWS2Len > uiWS1Len))  
            return -1;
                
        else
        {
            for (int x = ip; x < (uiWS1Len - uiWS2Len) + 1; x++)
                if (cmpW(ws1[x .. (x + uiWS2Len)], ws2) == 0) 
                    return x;
        }
    }

    return -1;

} // end findW( in wchar[], in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.findW( in wchar[], in wchar[] ).unittest"w );
    
    wchar[] wsPlts = "Mars: the fourth Rock (Planet) from the Sun."w;
    wchar[] wsMars = "Who\'s \'My Favorite Maritian?\'"w;
    int     i;

    i = findW(null, "a"w);
    assert( i == -1 );
    i = findW("def"w, "a"w);
    assert( i == -1 );
    i = findW("Abba"w, "a"w);
    assert( i == 3 );
    i = findW("def"w, "f"w);
    assert( i == 2 );
    i = findW("dfefFfg"w, "fFf"w);
    assert( i == 3 );
    i = findW("dfeffgffF"w, "ffF"w);
    assert( i == 6 );
    
    i = findW(wsMars, "MY fAVe"w);
    assert( i == -1 );
    i = findW(wsMars, "mY fAVOriTe"w);
    assert( i == -1 );
    i = findW(wsPlts, "Mars:"w);
    assert( i == 0 );
    i = findW(wsPlts, "Rock"w);
    assert( i == 17 );
    i = findW(wsPlts, "un."w);
    assert( i == 41 );
    i = findW(wsPlts, wsPlts);
    assert( i == 0 );
}

/+
 ' int rfindW( in wchar[], in wchar )
 '
 ' Find last occurrance of wc in string ws.
 ' Return index in ws where it is found.
 ' Return -1 if not found.
 '
 ' Note: Is Case Sensitive.
 +/
int rfindW(in wchar[] ws, in wchar wc)
{
    wchar wc1 = wc;
    wchar wc2;

    for (int i = ws.length - 1; i >= 0; i--)
    {   
        wc2 = ws[i]; 
        
        if (wc1 == wc2) 
            return i;   
    }

    return -1;
    
} // end rfindW( in wchar[], in wchar )

unittest
{
    debug( stringW ) writefln( "stringw.rfindW( in wchar[], in wchar ).unittest"w );

    int i;

    i = rfindW(null, cast(wchar)'a');
    assert( i == -1 );
    i = rfindW("def"w, cast(wchar)'a');
    assert( i == -1 );
    i = rfindW("Abba"w, cast(wchar)'A');
    assert( i == 0 );
    i = rfindW("def"w, cast(wchar)'f');
    assert( i == 2 );
}

/+
 ' int rfindW( in wchar[], in wchar[] )
 '
 ' Find last occurrance of ws2[] in string ws1[].
 ' Return index in ws1[] where it is found.
 ' Return -1 if not found.
 '
 ' Note: Is Case Sensitive.
 +/
int rfindW(in wchar[] ws1, in wchar[] ws2)
{    
    wchar wc;
    uint  uiWS1Len = ws1.length;
    uint  uiWS2Len = ws2.length;
    int   ip = -1; 
        
    if (uiWS2Len == 0 || uiWS1Len == 0 || uiWS2Len > uiWS1Len)  
            return -1; 

    if (uiWS2Len == 1)
    {
        wc = ws2[0];
        return rfindW(ws1, wc);
    }
    else
    {
        wc = ws2[0];
        ip = rfindW(ws1, wc);

        if (ip == -1) 
            return -1;
        else
        {
            for (int ix = ip; ix >= 0; ix--)
            {
                if (ix + uiWS2Len > uiWS1Len) 
                    continue;

                if (cmpW(ws1[ix .. (ix + uiWS2Len)], ws2) == 0) 
                    return ix;
            }        
        }
    }

    return -1;

} // end rfindW( in wchar[], in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.rfindW( in wchar[], in wchar[] ).unittest"w );

    wchar[] wsPlts = "Mars: the fourth Rock (Planet) from the Sun."w;
    int     i;

    i = rfindW(null, "a"w);
    assert( i == -1 );
    i = rfindW("def"w, "a"w);
    assert( i == -1 );
    i = rfindW("Abba"w, "a"w);
    assert( i == 3 );
    i = rfindW("def"w, "f"w);
    assert( i == 2 );
    i = rfindW("dfefFfg"w, "fFf"w);
    assert( i == 3 );
    i = rfindW("dfeffgffF"w, "ffF"w);
    assert( i == 6 );

    i = rfindW(wsPlts, "from"w);
    assert( i == 31 );
    i = rfindW(wsPlts, "M"w);
    assert( i == 0 );
    i = rfindW(wsPlts, "Su"w);
    assert( i == 40 );
}

/+
 ' wchar[] ireplaceW( in wchar[], in wchar[], in wchar[] )
 '
 ' Replace occurrences of wsFrom[] with wsTo[] in ws[].
 '
 ' Note: Is Non-Case Sensitive.
 +/
wchar[] ireplaceW(in wchar[] ws, in wchar[] wsFrom, in wchar[] wsTo)
{
    wchar   wc;
    wchar[] wsReturn;
    uint    uiWSLen   = ws.length;
    uint    uiFromLen = wsFrom.length;
    uint    uiToLen   = wsTo.length;
    int     ip;
    int     ix;
    
    // if the length of any of the strings equals zero return "ws",
    // except the "to" can be an "" (Empty String).
    if (uiFromLen == 0 || uiWSLen == 0 || 
        uiFromLen > uiWSLen - 1 || uiToLen > uiWSLen - 1) 
        return ws; 
       
    // find the position of the string from's 1st wchar
    wc = wsFrom[0];    
    ip = ifindW(ws, wc);
        
    // if the string from's 1st char was not found, return "ws" 
    if (ip == -1 || ip + uiFromLen > uiWSLen - 1) 
        return ws;
    
    if (ip > 0) 
        wsReturn = ws[0 .. ip]; 
    
    for (ix = ip; ix < (uiWSLen - uiFromLen) + 1; ix++)
    {
        if (icmpW(ws[ix .. (ix + uiFromLen)], wsFrom) == 0) 
        {
            wsReturn ~= wsTo;
            ix += uiFromLen - 1;
        }
        else
            wsReturn ~= ws[ix .. (ix + 1)];
    }
        
    if (ix < uiWSLen - 1) 
    {
        ip = uiWSLen - ix;
        wsReturn ~= ws[ix .. ix + ip];     
    }

    return wsReturn; 
    
} // end wchar[] ireplaceW( in wchar[], in wchar[], in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.ireplaceW( in wchar[], in wchar[], in wchar[] ).unittest"w );

    wchar[] ws     = "This is a FOO foO list"w;
    wchar[] wsFrom = "fOo"w;
    wchar[] wsTo   = "silly"w;
    wchar[] wsR;
    int     i;

    wsR = ireplaceW(ws, wsFrom, wsTo);
    i = icmpW(wsR, "This is a silly silly list"w);
    assert( i == 0 );
    
    ws = "Love is a two way Street!"w;
    wsR = ireplaceW( ws, "TwO"w, "one"w);
    assert( wsR == "Love is a one way Street!"w );
    
    wsR = ireplaceW("ApPLe-Bee"w, "pple"w, "te-A"w);
    assert( wsR == "Ate-A-Bee"w );
}

/+
 ' wchar[] replaceW( in wchar[], in wchar[], in wchar[] )
 '
 ' Replace occurrences of wsFrom[] with wsTo[] in ws[].
 '
 ' Note: Is case sensitive.
 +/
wchar[] replaceW(in wchar[] ws, in wchar[] wsFrom, in wchar[] wsTo)
{
    wchar   wc;
    wchar[] wsReturn;
    uint    uiWSLen   = ws.length;
    uint    uiFromLen = wsFrom.length;
    uint    uiToLen   = wsTo.length;
    int     ip;
    int     ix;
    
    // if the length of any of the strings equals zero return "ws",
    // except the "wsTo" can be an "" (Empty String).
    if (uiFromLen == 0 || uiWSLen == 0 || 
        uiFromLen > uiWSLen - 1 || uiToLen > uiWSLen - 1) 
        return ws; 
       
    // find the position of the string from's 1st wchar
    wc = wsFrom[0];    
    ip = findW(ws, wc);
        
    // if the string from's 1st char was not found, return "ws" 
    if (ip == -1 || ip + uiFromLen > uiWSLen - 1) 
        return ws;
    
    if (ip > 0) 
        wsReturn = ws[0 .. ip]; 
    
    for (ix = ip; ix < (uiWSLen - uiFromLen) + 1; ix++)
    {
        if ( cmpW(ws[ix .. (ix + uiFromLen)], wsFrom) == 0) 
        {
            wsReturn ~= wsTo;
            ix += uiFromLen - 1;
        }
        else
            wsReturn ~= ws[ix .. (ix + 1)];
    }
        
    if (ix < uiWSLen - 1) 
    {
        ip = uiWSLen - ix;
        wsReturn ~= ws[ix .. ix + ip];     
    }

    return wsReturn; 
    
} // end replaceW( in wchar[], in wchar[], in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.replaceW( in wchar[], in wchar[], in wchar[] ).unittest"w );

    wchar[] ws     = "This is a FOO foO list"w;
    wchar[] wsFrom = "foO"w;
    wchar[] wsTo   = "silly"w;
    wchar[] wsR;
    int     i;

    wsR = replaceW(ws, wsFrom, wsTo);
    i = cmpW(wsR, "This is a FOO silly list"w);
    assert( i == 0 );
    
    ws = "Love is a two way Street!"w;
    wsR = replaceW(ws, "TwO"w, "one"w);
    assert( wsR != "Love is a one way Street!"w );
}

/+
 ' int icountW( in wchar[], in wchar[] )
 '
 ' Count up all instances of wsSub[] in ws[].
 '
 ' Note: Is Non-Case Sensitive.
 +/
int icountW(in wchar[] ws, in wchar[] wsSub)
{
    uint uiWSLen  = ws.length;
    uint uiSubLen = wsSub.length;
    int  ip       = 0;
    int  iCounter = 0;

    if (uiWSLen == 0 || uiSubLen == 0 || uiSubLen > uiWSLen) 
        return 0;
    
    for (int ix = 0; ix < uiWSLen; ix += ip + uiSubLen)
    {
        ip = ifindW(ws[ix .. uiWSLen], wsSub);

        if (ip == -1) 
            break;
    
        iCounter++;
    }
    
    return iCounter;
    
} // end int icountW( in wchar[], in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.icountW( in wchar[], in wchar[] ).unittest"w );

    wchar[] ws    = "This iS a foFofof liSt"w;
    wchar[] wsSub = "FOf"w;
    int     i;

    i = icountW(ws, wsSub);
    assert( i == 2 );

    i = icountW(ws, "S"w);
    assert( i == 3 );
}

/+
 ' int countW( in wchar[], in wchar[] )
 '
 ' Count up all instances of wsSub[] in ws[].
 '
 ' Note: Is Case Sensitive.
 +/
int countW(in wchar[] ws, in wchar[] wsSub)
{
    uint uiWSLen  = ws.length;
    uint uiSubLen = wsSub.length;
    int  ip       = 0;
    int  iCounter = 0;

    if (uiWSLen == 0 || uiSubLen == 0 || uiSubLen > uiWSLen) 
        return 0;
    
    for (int ix = 0; ix < uiWSLen; ix += ip + uiSubLen)
    {
        ip = findW(ws[ix .. uiWSLen], wsSub);

        if (ip == -1) break;
            iCounter++;
    }
    
    return iCounter;
    
} // end int countW( in wchar[], in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.countW( in wchar[], in wchar[] ).unittest"w );

    wchar[] ws    = "This iS a foFOfof liSt"w;
    wchar[] wsSub = "FOf"w;
    int     i;

    i = countW(ws, wsSub);
    assert( i == 1 );
    
    i = countW(ws, "S"w);
    assert( i == 2 );
}

/+
 ' Function      : int ifindxW( in wchar[], in wchar[], in int = 0 )
 ' Author        : David L. 'SpottedTiger' Davis
 ' Complied      : DigitalMars "D" aka Mars v0.92
 ' Created Date  : 03.Jun.04 
 ' Modified Date : 08.Jun.04 Removed the wrapper function and set 
 '                           the third parameter as a default of 0
 '               : 24.Aug.04 Now that ifind() is a part of Phobos,
 '                           and the fact that I wrote this original
 '                           function to use a default...I've renamed
 '                           this function from ifind() to ifindx() 
 '                           [ meaning ifind() eXtended :)) ]
 '               : 24.Aug.04 Compiled and tested with dmd v0.99
 '               : 01.Sep.04 Compiled and tested with dmd v0.101
 '               : 03.Sep.04 Duped from the original, and recreated for wchar[]
 ' Requirements  : N/A
 ' Licence       : Public Domain
 ' ---------------------------------------------------------------------------
 ' Note: Meant to extend the Non-Case Sensitive version of ifind() with 
 '       an optional start searching from here string position parameter.
 +/
int ifindxW(in wchar[] wsStr, in wchar[] wsSubStr, in int iStartPos = 0)
{
    wchar[] wsTmpStr;
    int     iRtnVal;
    int     iStrLen = wsStr.length;
    
    // If either of the string parameters are empty, return not found
    if (iStrLen < 1 || wsSubStr.length < 1) 
        return -1;
    
    // If greater than to upper boundary return not found    
    if (iStartPos > iStrLen - 1) 
        return -1;  
        
    // If less than to lower boundary return not found   
    else if (iStartPos < 0) 
        return - 1; 
           
    wsTmpStr = wsStr[iStartPos .. iStrLen];
  
    if (iStartPos == 0) 
        return ifindW(wsTmpStr, wsSubStr);   
    else
    {
        iRtnVal = ifindW(wsTmpStr, wsSubStr);

        return (iRtnVal != -1) ? iStartPos + iRtnVal : iRtnVal;
    } 
    
} // end int ifindxW( in wchar[], in wchar[], in int = 0 ) 

unittest
{
    debug( stringW ) writefln( "stringw.ifindxW( in wchar[], in wchar[], in int = 0 ).unittest"w );
    
    wchar[] wsPlts = "Mars: the fourth Rock (Planet) from the Sun."w;
    wchar[] wsMars = "Who\'s \'My Favorite Maritian?\'"w;
    int     i;

    i = ifindxW(null, "a"w);
    assert( i == -1 );
    i = ifindxW("def"w, "a"w);
    assert( i == -1 );
    i = ifindxW("abbA"w, "a"w, 1);
    assert( i == 3 );
    i = ifindxW("deF"w, "f"w);
    assert( i == 2 );
    i = ifindxW("dfefffg"w, "fFf"w);
    assert( i == 3 );
    i = ifindxW("dfeffgfff"w, "FFF"w, 4);
    assert( i == 6 );
    
    i = ifindxW(wsMars, "MY fAVe"w);
    assert( i == -1 );
    i = ifindxW(wsMars, "mY fAVOriTe"w, 3);
    assert( i == 7 );
    i = ifindxW(wsPlts, "mArS:"w);
    assert( i == 0 );
    i = ifindxW(wsPlts, "rOcK"w);
    assert( i == 17 );
    i = ifindxW(wsPlts, "rOcK"w, 19);
    assert( i == -1 );
    i = ifindxW(wsPlts, "Un."w, 35);
    assert( i == 41 );
    i = ifindxW(wsPlts, wsPlts);
    assert( i == 0 );
}

/+
 ' Function      : int irfindxW( in wchar[], in wchar[], in int = -1 )
 ' Author        : David L. 'SpottedTiger' Davis
 ' Compiled      : DigitalMars "D" aka Mars v0.92
 ' Created Date  : 03.Jun.04 
 ' Modified Date : 08.Jun.04 Removed the wrapper function and set
 '                           the third parameter as a default of -1
 '               : 24.Aug.04 Now that irfind() is a part of Phobos,
 '                           and the fact that I wrote this original
 '                           function to use a default...I've renamed
 '                           this function from irfind() to irfindx() 
 '                           [ meaning irfind() eXtended :)) ]
 '               : 24.Aug.04 Compiled and tested with dmd v0.99
 '               : 01.Sep.04 Compiled and tested with dmd v0.101
 '               : 03.Sep.04 Duped from the original, and recreated for wchar[]
 ' Requirements  : N/A
 ' Licence       : Public Domain
 ' --------------------------------------------------------------------------
 ' Note: Meant to extend the Non-Case Sensitive version of irfind() with 
 '       an optional start searching from here string position parameter.
 +/
int irfindxW(in wchar[] wsStr, in wchar[] wsSubStr, in int iEndPos = -1)
{
    wchar[] wsTmpStr;
    int     iStrLen = wsStr.length;
    
    // If either of the string parameters are empty, return not found
    if (iStrLen < 1 || wsSubStr.length < 1) 
        return -1;

    // If iEndPos == -1 get the full length of the string
    if (iEndPos == -1) 
        iEndPos = iStrLen - 1;
        
    // If greater than to upper boundary return not found    
    else if (iEndPos > iStrLen - 1) 
        return -1;  
        
    // If less than to lower boundary return not found   
    else if (iEndPos < 0) 
        return -1;      
    
    wsTmpStr = wsStr[0 .. iEndPos + 1];

    if (wsTmpStr.length < wsSubStr.length)
        return -1;
        
    return irfindW(wsTmpStr, wsSubStr);

} // end int irfindxW( in wchar[], in wchar[], in int = -1 ) 

unittest
{
    debug( stringW ) writefln("stringw.irfindxW( in wchar[], in wchar[], in int = -1 ).unittest"w );

    wchar[] wsPlts = "Mars: the fourth Rock (Planet) from the Sun."w;
    wchar[] wsMars = "Who\'s \'My Favorite Maritian?\'"w;
    int     i;

    i = irfindxW("abcdefcdef"w, "DEF"w, 8);
    assert( i == 3 );
    i = irfindxW("abcdefcdef"w, "Cd"w, 6);
    assert( i == 2 );
    
    i = irfindxW("abcdefcdef"w, "c"w, 5);
    assert( i == 2 );
    i = irfindxW("abcdefcdef"w, "x"w);
    assert( i == -1 );
    i = irfindxW("abcdefcdef"w, "xY"w);
    assert( i == -1 );
    i = irfindxW("abcdefcdef"w, ""w);
    assert( i == -1 );
    
    i = irfindxW(wsMars, "RiTE maR"w);
    assert( i == 14 );
    i = irfindxW(wsPlts, "FOuRTh"w);
    assert( i == 10 );
    i = irfindxW(wsMars, "whO\'s \'MY"w);
    assert( i == 0 );
    i = irfindxW(wsMars, wsMars);
    assert( i == 0 );
}    

/+
 ' Function      : int findxW( in wchar[], in wchar[], in int = 0 )
 ' Author        : David L. 'SpottedTiger' Davis
 ' Complied      : DigitalMars "D" aka Mars v0.92
 ' Created Date  : 03.Jun.04 
 ' Modified Date : 08.Jun.04 Removed the wrapper function and set 
 '                           the third parameter as a default of 0
 '               : 24.Aug.04 Now that ifind() is a part of Phobos,
 '                           and the fact that I wrote this original
 '                           function to use a default...I've renamed
 '                           this function from ifind() to ifindx() 
 '                           [ meaning ifind() eXtended :)) ]
 '               : 24.Aug.04 Compiled and tested with dmd v0.99
 '               : 01.Sep.04 Compiled and tested with dmd v0.101
 '               : 03.Sep.04 Duped from the original, and recreated for wchar[]
 ' Requirements  : N/A
 ' Licence       : Public Domain
 ' ---------------------------------------------------------------------------
 ' Note: Meant to extend the Case Sensitive version of ifind() with 
 '       an optional start searching from here string position parameter.
 +/
int findxW(in wchar[] wsStr, in wchar[] wsSubStr, in int iStartPos = 0)
{
    wchar[] wsTmpStr;
    int     iRtnVal;
    int     iStrLen = wsStr.length;
    
    // If either of the string parameters are empty, return not found
    if (iStrLen < 1 || wsSubStr.length < 1) 
        return -1;
    
    // If greater than to upper boundary return not found    
    if (iStartPos > iStrLen - 1) 
        return -1;  
        
    // If less than to lower boundary return not found   
    else if (iStartPos < 0) 
        return - 1; 
           
    wsTmpStr = wsStr[iStartPos .. iStrLen];
  
    if (iStartPos == 0) 
        return findW(wsTmpStr, wsSubStr);   
    else
    {
        iRtnVal = findW(wsTmpStr, wsSubStr);

        return (iRtnVal != -1) ? iStartPos + iRtnVal : iRtnVal;
    } 
    
} // end int findxW( in wchar[], in wchar[], in int = 0 ) 

unittest
{
    debug( stringW ) writefln( "stringw.findxW( in wchar[], in wchar[], in int = 0 ).unittest"w );
    
    wchar[] wsPlts = "Mars: the fourth Rock (Planet) from the Sun."w;
    wchar[] wsMars = "Who\'s \'My Favorite Maritian?\'"w;
    int     i;

    i = findxW(null, "a"w);
    assert( i == -1 );
    i = findxW("def"w, "a"w);
    assert( i == -1 );
    i = findxW("abbA"w, "a"w, 0);
    assert( i == 0 );
    i = findxW("deF"w, "f"w);
    assert( i == -1 );
    i = findxW("dfefFfg"w, "fFf"w);
    assert( i == 3 );
    i = findxW("dfeffgfff"w, "FFF"w, 4);
    assert( i == -1 );
    
    i = findxW(wsMars, "MY fAVe"w);
    assert( i == -1 );
    i = findxW(wsMars, "mY fAVOriTe"w, 3);
    assert( i == -1 );
    i = findxW(wsPlts, "Mars:"w);
    assert( i == 0 );
    i = findxW(wsPlts, "Rock"w);
    assert( i == 17 );
    i = findxW(wsPlts, "Rock"w, 19);
    assert( i == -1 );
    i = findxW(wsPlts, "un."w, 35);
    assert( i == 41 );
    i = findxW(wsPlts, wsPlts);
    assert( i == 0 );
}

/+
 ' Function      : int rfindxW( in wchar[], in wchar[], in int = -1 )
 ' Author        : David L. 'SpottedTiger' Davis
 ' Compiled      : DigitalMars "D" aka Mars v0.92
 ' Created Date  : 03.Jun.04 
 ' Modified Date : 08.Jun.04 Removed the wrapper function and set
 '                           the third parameter as a default of -1
 '               : 24.Aug.04 Now that irfind() is a part of Phobos,
 '                           and the fact that I wrote this original
 '                           function to use a default...I've renamed
 '                           this function from irfind() to irfindx() 
 '                           [ meaning irfind() eXtended :)) ]
 '               : 24.Aug.04 Compiled and tested with dmd v0.99
 '               : 01.Sep.04 Compiled and tested with dmd v0.101
 '               : 03.Sep.04 Duped from the original, and recreated for wchar[]
 ' Requirements  : N/A
 ' Licence       : Public Domain
 ' --------------------------------------------------------------------------
 ' Note: Meant to extend the Case Sensitive version of rfind() with 
 '       an optional start searching from here string position parameter.
 +/
int rfindxW(in wchar[] wsStr, in wchar[] wsSubStr, in int iEndPos = -1)
{
    wchar[] wsTmpStr;
    int     iStrLen = wsStr.length;
    
    // If either of the string parameters are empty, return not found
    if (iStrLen < 1 || wsSubStr.length < 1) 
        return -1;

    // If iEndPos == -1 get the full length of the string
    if (iEndPos == -1) 
        iEndPos = iStrLen - 1;
        
    // If greater than to upper boundary return not found    
    else if (iEndPos > iStrLen - 1) 
        return -1;  
        
    // If less than to lower boundary return not found   
    else if (iEndPos < 0) 
        return - 1;      
 
    wsTmpStr = wsStr[0 .. iEndPos + 1];

    return rfindW( wsTmpStr, wsSubStr );  

} // end int rfindxW( in wchar[], in wchar[], in int = -1 ) 

unittest
{
    debug( stringW ) writefln( "stringw.rfindxW( in wchar[], in wchar[], in int = -1 ).unittest"w );

    wchar[] wsPlts = "Mars: the fourth Rock (Planet) from the Sun."w;
    int     i;

    i = rfindxW("Abba"w, "a"w, 5);
    assert( i == -1 );
    i = rfindxW(null, "a"w);
    assert( i == -1 );
    i = rfindxW("def"w, "a"w);
    assert( i == -1 );
    i = rfindxW("def"w, "f"w);
    assert( i == 2 );
    i = rfindxW("dfefFfg"w, "fFf"w, 6);
    assert( i == 3 );
    i = rfindxW("dfeffgffF"w, "ffF"w);
    assert( i == 6 );

    i = rfindxW( wsPlts, "from"w, 39);
    assert( i == 31 );
    i = rfindxW( wsPlts, "M"w, 15);
    assert( i == 0 );
    i = rfindxW( wsPlts, "Su"w);
    assert( i == 40 );
 }

/+
 ' wchar[][] isplitW( in wchar[], in wchar[] )
 '
 ' Split ws[] into an array of words,
 ' using wsDelim[] as the delimiter.
 '
 ' Note: Is Non-Case Sensitive.
 +/
wchar[][] isplitW(in wchar[] ws, in wchar[] wsDelim)
{
    int       iPos1;
    int       iPos2;
    wchar[][] wsWords;
    uint      uiWSLen    = ws.length;
    uint      uiDelimLen = wsDelim.length;
    uint      uiCounter  = 0;
    
    if (uiWSLen == 0 || uiDelimLen == 0 || uiDelimLen > uiWSLen) 
    {
        wsWords.length = 1;
        wsWords[0]   = null; 
        return wsWords;
    }
    
    uiCounter = icountW(ws, wsDelim);
    wsWords.length = (uiCounter == 0 ? 1 : uiCounter + 1);
    
    // No delimiter found, return ws.
    if (uiCounter == 0) 
    {
        wsWords[0] = ws;
        
        return wsWords;
    }
    
    iPos1 = 0;
    iPos2 = -1;
    
    for (int ix = 0; ix < uiCounter + 1; ix++)
    {
        iPos2 = ifindxW(ws, wsDelim, iPos1);
        
        if (iPos2 == -1)
        {
            wsWords[ix] = ws[iPos1 .. uiWSLen];
            break;
        }    
        else
        {
            wsWords[ix] = ws[iPos1 .. iPos2];
            iPos1 = iPos2 + uiDelimLen; 
        }    
    }
    
    return wsWords;
    
} // end wchar[][] isplitW( in wchar[], in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.isplitW( in wchar[], in wchar[] ).unittest"w );

    wchar[][] wsTest;
    
    wsTest = isplitW("usa_name/last_names/name"w, "/"w);
    for (int ix = 0; ix < wsTest.length; ix++)
    {
        switch ( ix )
        {
          case 0: assert( wsTest[0] == "usa_name"w );
          case 1: assert( wsTest[1] == "last_names"w );
          case 2: assert( wsTest[2] == "name"w );
          default:
        }     
    }

    wsTest = isplitW("usa_name@a@last_names@A@name"w, "@A@"w);
    for (int ix = 0; ix < wsTest.length; ix++)
    {
        switch (ix)
        {
          case 0: assert( wsTest[0] == "usa_name"w );
          case 1: assert( wsTest[1] == "last_names"w );
          case 2: assert( wsTest[2] == "name"w );
          default:
        }     
    }
        
    wsTest = isplitW("usa_name", "/"w);
    assert( wsTest[0] == "usa_name"w );
    
    wsTest = isplitW(""w, "/"w);
    assert( wsTest[0] == ""w );
}

/+
 ' wchar[][] splitW( in wchar[], in wchar[] )
 '
 ' Split ws[] into an array of words,
 ' using wsDelim[] as the delimiter.
 '
 ' Note: Is Case Sensitive.
 +/
wchar[][] splitW(in wchar[] ws, in wchar[] wsDelim)
{
    int       iPos1;
    int       iPos2;
    wchar[][] wsWords;
    uint      uiWSLen    = ws.length;
    uint      uiDelimLen = wsDelim.length;
    uint      uiCounter  = 0;
    
    if (uiWSLen == 0 || uiDelimLen == 0 || uiDelimLen > uiWSLen) 
    {
        wsWords.length = 1;
        wsWords[0] = null; 
        return wsWords;
    }
    
    uiCounter = countW(ws, wsDelim);
    wsWords.length = (uiCounter == 0 ? 1 : uiCounter + 1);
    
    // No delimiter found, return ws.
    if (uiCounter == 0) 
    {
        wsWords[0] = ws;
        return wsWords;
    }
    
    iPos1 = 0;
    iPos2 = -1;
    
    for (int ix = 0; ix < uiCounter + 1; ix++)
    {
        iPos2 = findxW(ws, wsDelim, iPos1);
        
        if (iPos2 == -1)
        {
            wsWords[ix] = ws[iPos1 .. uiWSLen];
            break;
        }    
        else
        {
            wsWords[ix] = ws[iPos1 .. iPos2];
            iPos1 = iPos2 + uiDelimLen;
        }    
    }
    
    return wsWords;
    
} // end wchar[][] splitW( in wchar[], in wchar[] )  

unittest
{
    debug( stringW ) writefln( "stringw.splitW( in wchar[], in wchar[] ).unittest"w );

    wchar[][] wsTest;
    
    wsTest = splitW("usa_name/Last_Names/name"w, "/"w);
    for (int ix = 0; ix < wsTest.length; ix++)
    {
        switch (ix)
        {
          case 0: assert( wsTest[0] == "usa_name"w );
          case 1: assert( wsTest[1] == "Last_Names"w );
          case 2: assert( wsTest[2] == "name"w );
          default:
        }     
    }

    wsTest = splitW("usa_name@a@last_names@A@name"w, "@A@"w);
    assert( wsTest[0] == "usa_name@a@last_names"w );
    assert( wsTest[1] == "name"w );
          
    wsTest = splitW("USA_name"w, "/"w);
    assert( wsTest[0] == "USA_name"w );

    wsTest = splitW(""w, "/"w);
    assert( wsTest[0] == ""w );
}  

/+
 ' wchar[] tolowerW( in wchar[] )
 '
 ' Convert an ASCII string to lowercase.
 +/
wchar[] tolowerW(in wchar[] ws)
{
    wchar[] wsReturn;
    wchar   wc;

    // Are there any characters
    if (isASCIIStringW(ws, true))
        wsReturn = ws.dup;
    else    
        return ws;
        
    for (int ix = 0; ix < ws.length; ix++)
    {
        wc = wsReturn[ix];

        // If within ASCII 'A' thru 'Z' range, 
        // convert the character to lowercase.
        if (wc >= '\u0041' && wc <= '\u005A')
            wsReturn[ix] = wc + '\u0020';
    }
    
    return wsReturn;
    
} // end wchar[] tolowerW( in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.tolowerW( in wchar[] ).unittest"w );

    wchar[] ws1 = "FoL"w;
    wchar[] ws2;

    ws2 = tolowerW(ws1);
    assert( cmpW(ws2, "fol"w) == 0 );
    assert( ws2 != ws1 );
}

/+
 ' wchar[] tolowerW( in wchar[] )
 '
 ' Convert an ASCII string to uppercase.
 +/
wchar[] toupperW(in wchar[] ws)
{
    wchar[] wsReturn;
    wchar   wc;

    if (isASCIIStringW(ws, true))
        wsReturn = ws.dup;
    else    
        return ws;
        
    for (int ix = 0; ix < ws.length; ix++)
    {
        wc = wsReturn[ix];

        // If within ASCII 'a' thru 'z' range, 
        // convert the character to uppercase.
        if (wc >= '\u0061' && wc <= '\u007A')
            wsReturn[ix] = wc - '\u0020';
    }
    
    return wsReturn;
    
} // end wchar[] toupperW( in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.toupperW( in wchar[] ).unittest"w );

    wchar[] s1 = "FoL"w;
    wchar[] s2;

    s2 = toupperW(s1);
    assert( cmpW(s2, "FOL"w) == 0 );
    assert( s2 !is s1 );
}

/+
 ' bool isASCIIStringW( in wchar[], in bool = false )
 '
 ' returns true if an ASCII character is found.
 +/
bool isASCIIStringW(in wchar[] ws, in bool bIgnoreWhiteSpace = false)
{
    /+
     ' Returns true if any character below '\u007F' is found...    
     ' these are the characters that fall into the ASCII 
     ' 7-bit range of characters. Unless the Ignore White 
     ' Space is set to true, then bypass checking those characters.
     +/
    for (int ix = 0; ix < ws.length; ix++)
    {
        if (bIgnoreWhiteSpace)
            if (findW(WHITESPACE_W, ws[ix]) != -1)
                continue;
                
        if (ws[ix] < '\u0080') 
            return true;
    }
    
    return false;
         
} // end bool isASCIIStringW( in wchar[], in bool = false )

unittest
{
    debug( stringW ) writefln( "stringw.isASCIIStringW( in wchar[], in bool = false ).unittest"w );

    wchar[] ws1 = "\u0392\u1ECE\u0154\u0419 "w;
    wchar[] ws2;
    bool    bFlag = false;
    
    bFlag = isASCIIStringW(ws1, true);
    assert( bFlag == false );
    bFlag = isASCIIStringW(ws1);
    assert( bFlag == true );
    bFlag = isASCIIStringW("Apple-\u03B2"w ~ "ee"w);
    assert( bFlag == true );
}

/+
 ' Function      : bool isRomanNumeralW( in wchar )
 ' Date Created  : 03.Jun.04 
 ' Date Modified : 06.Sep.04 - Tested and complied with dmd v0.101
 ' Requirements  : none
 +/
bool isRomanNumeralW(in wchar wcChar)
{
    return (findW(ROMAN_W, wcChar) != -1) ? true : false;

} // end bool isRomanNumeralW( in wchar )

unittest
{
    debug( stringW ) writefln( "stringw.isRomanNumeralW( in wchar ).unittest"w );

    bool    bFlag = false;
    
    bFlag = isRomanNumeralW('x');
    assert( bFlag == true );
    bFlag = isRomanNumeralW('M');
    assert( bFlag == true );
    bFlag = isRomanNumeralW('a');
    assert( bFlag == false );
}

/+
 ' Function      : bool isRomanNumeralW( in wchar[] )
 ' Date Created  : 03.Jun.04 
 ' Date Modified : 06.Sep.04 - Tested and complied with dmd v0.101
 ' Requirements  : none
 +/
bool isRomanNumeralW(in wchar[] wsStr)
{
    foreach(int iStrPos, wchar wcChar; wsStr)
    {
        if (findW(ROMAN_W, wcChar) == -1) 
            return false;
    } 
      
    return true;
        
} // end bool isRomanNumeralW( in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.isRomanNumeralW( in wchar[] ).unittest"w );

    wchar[] ws1 = "MMVII"w;
    bool    bFlag = false;
    
    bFlag = isRomanNumeralW(ws1);
    assert( bFlag == true );
    bFlag = isRomanNumeralW("XiIi"w);
    assert( bFlag == true );
    bFlag = isRomanNumeralW("XiXa"w);
    assert( bFlag == false );
}

/+
 ' wchar[] capitalizeW( in wchar[] ws )
 '
 ' Capitalize first character of string.
 +/
wchar[] capitalizeW(in wchar[] ws)
{
    wchar[] wsReturn;
    wchar   wc;
    
    if (ws.length > 0)
    {
        wc = ws[0];
        
        // If within ASCII 'a' thru 'z' range, 
        // convert the character to uppercase.
        if (wc >= '\u0061' && wc <= '\u007A')
        {
            wsReturn = ws.dup;
            wsReturn[0] = wc - '\u0020';
            return wsReturn;
        }
    }
    
    return ws;
    
} // end wchar[] capitalizeW( in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.capitalizeW( in wchar[] ).unittest"w );

    wchar[] ws1 = "foL"w;
    wchar[] ws2;
    ws2 = capitalizeW(ws1);
    assert( cmpW( ws2, "FoL"w ) == 0 );
    assert( ws2 !is ws1 );
}

/+
 ' wchar[] striplW( in wchar ws )
 '
 ' Strips leading whitespace.
 +/
wchar[] striplW(in wchar[] ws)
{
    int ix;
    
    for (ix = 0; ix < ws.length; ix++)
        if ((findW(WHITESPACE_W, ws[ix]) != -1) ? false : true)
            break;
        
    return ws[ix .. ws.length];
    
} // end wchar[] striplW( in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.striplW( in wchar[] ).unittest"w );
    
    wchar[] ws;
    int     i;

    ws = striplW("\n\t \f \v\r  foo\r\t\f\v\n   "w);
    i = cmpW(ws, "foo\r\t\f\v\n   "w);
    assert( i == 0 );
}

/+
 ' wchar[] striprW( in wchar[] ws )
 '
 ' Strips off trailing whitespace.
 +/
wchar[] striprW(in wchar[] ws)
{
    int ix;
    
    for (ix = ws.length - 1; 0 < ix; ix--)
        if ((findW(WHITESPACE_W, ws[ix] ) != -1) ? false : true)
            break;
    
    return ws[0 .. ix + 1];
    
} // end wchar[] striprW( in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.striprW( in wchar[] ).unittest"w );
    
    wchar[] ws;
    int     i;

    ws = striprW("  foo\r\t\f\v\n   "w);
    i = cmpW(ws, "  foo"w);
    assert( i == 0 );
}

/+
 ' wchar[] stripW( wchar[] ws )
 '
 ' Strips off all leading and trailing whitespace.
 +/
wchar[] stripW(in wchar[] ws)
{
    return striprW(striplW(ws));
    
} // wchar[] stripW( in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.stripW( in wchar[] ).unittest"w );
    
    wchar[] ws;
    int     i;

    ws = stripW("  foo\t\n\r\v\f  "w);
    i = cmpW(ws, "foo"w);
    assert( i == 0 );
}

/+
 ' Function        : wchar[] asciiProperCaseW( in wchar[] )
 ' Author          : David L. 'SpottedTiger' Davis
 ' Compiler        : dmd v0.92
 ' Date Created    : 03.Jun.04 
 ' Date Modified   : 24.Aug.04 Compiled and tested with dmd v0.99
 '                 : 07.Sep.04 Converted the char[] version to wchar[]s
 '                 :           Compiled and tested with dmd v0.101
  '
 ' Licence         : Pubic Domain
 ' Requirements    : std.string
 ' ------------------------------------------------------------------
 ' Purpose : This function was created to handle Names and Addresses.
 '
 ' 01) Caps the first character of each word.
 ' 02) Converts o'bein to O'Bein
 ' 03) Converts apple-bees to Apple-Bees
 ' 04) Ignores commas, ex: doe,jane to Doe,Jane
 ' 05) Coverts mcdonald to McDonald
 ' 06) Converts all roman numerials to caps
 ' 07) Converts 1ST, 2ND, 3RD, 4TH to 1st, 2nd, 3rd, 4th
 ' 08) Trims off leading and trailing spaces
 ' 09) Converts po and pob to PO and POB
 ' 10) Converts p.o. to P.O. or caps the next character that follow a period
 '
 +/
wchar[] asciiProperCaseW(in wchar[] wsStrPassed)
{
    // "\x20\t\v\r\n\f'-.,";
    const wchar[] WHITESPACE_WX = "\u0020\u0009\u000B\u000D\u000A\u000C\u002D\u0027\u002E\u002C"w; 
        
    bool    bWhitespaceFound = false;
    bool    bRomanFound      = false;
    int     iStrPosWK        = 0;
    int     iStrPos          = 0;
    wchar[] wsStr            = new wchar[ wsStrPassed.length ];
    int     iWsStrLen        = wsStr.length; 
    
    wsStr = wsStrPassed; 
    
    if (iWsStrLen > 0) 
    {
        /+
         ' Initial pass:
         ' ---------------
         ' 1) strip all leading and trailing spaces
         ' 2) lowercase all characters
         ' 3) very 1st character set to uppercase
         ' 4) capitalize each 1st character in every word
         +/

        wsStr = capitalizeW(tolowerW(stripW(wsStr)));

        //sStr = capwordsW( wsStr ); // not used because it does too much for my needs
        
        // Reset iWsStrLen after stripW() has been applied.
        iWsStrLen = wsStr.length;
        
        foreach(int iStrPos2, wchar wc; wsStr)
        { 
            if (findW(WHITESPACE_WX, wc) != -1)     
            { 
                bWhitespaceFound = true;
            } 
            else if (bWhitespaceFound) 
            { 
                bWhitespaceFound = false;
                wsStr[iStrPos2] = toupperW(wsStr[iStrPos2]); 
            } 

            /+*********************
             *  Roman Numeral(s)  *
             *********************+/   
            bRomanFound = isRomanNumeralW(wsStr[iStrPos2]);
            iStrPosWK = iStrPos2;
    
            while (iStrPosWK < iWsStrLen && bRomanFound)
            {
                //writefln("Roman Number found, sStr.length=%d, iStrPosWK=%d", sStr.length, iStrPosWK );
                iStrPosWK++;
                if (iStrPosWK < iWsStrLen) 
                    bRomanFound = isRomanNumeralW(wsStr[iStrPosWK]);
            } 
    
            if (iStrPosWK >= iWsStrLen) 
                iStrPosWK = iWsStrLen - 1;

            if ((findW(WHITESPACE_WX, wsStr[iStrPosWK]) != -1 || 
                   iStrPosWK == iWsStrLen - 1)) 
            {
                // Handles before WHITESPACE
                bRomanFound = isRomanNumeralW(wsStr[(iStrPosWK - 2 >= 0) ? iStrPosWK - 2 : iStrPosWK]);
        
                if (bRomanFound)
                {      
                    wsStr[iStrPos2 .. iStrPosWK] = toupperW(wsStr[iStrPos2 .. iStrPosWK]);
            
                    // Needed for the very last character if it's a Roman Numeral
                    if (iStrPosWK == iWsStrLen - 1) 
                        wsStr[iStrPosWK] = toupperW(wsStr[iStrPosWK]);  
            
                    iStrPos2 = iStrPosWK;
                    bWhitespaceFound = true;
                }     
            } 
    
            //writefln("Char=%s", sStr[ iStrPos2 ]);

        } // end-foreach

        /+
         ' * Work on all the special cases *
         '
         ' Need a find with replace for the following:
         ' 1) po     -> PO
         ' 2) pob    -> POB
         ' 3) mccain -> McCain
         '
         +/

        // Handles " po "
        iStrPos = 0;

        if (icmpW(wsStr[0 .. 3], "PO\u0020"w) == 0) 
            wsStr[0 .. 3] = "PO\u0020"w;

        while (iStrPos < iWsStrLen - 1 && iStrPos != -1)
        {
            iStrPos = ifindxW(wsStr, "\u0020PO\u0020"w, iStrPos);

            if (iStrPos != -1) 
            {
                wsStr[iStrPos .. iStrPos + 4] = "\u0020PO\u0020"w;
                iStrPos = iStrPos + 5;
            } 
        } 

        // Handles " pob " 
        iStrPos = 0;

        if (icmpW(wsStr[0 .. 4], "POB\u0020"w) == 0) 
            wsStr[0 .. 4] = "POB\u0020"w;

        while (iStrPos < iWsStrLen - 1 && iStrPos != -1)
        {
            iStrPos = ifindxW(wsStr, "\u0020POB\u0020"w, iStrPos);

            if (iStrPos != -1) 
            {
                wsStr[iStrPos .. iStrPos + 5] = "\u0020POB\u0020"w;
                iStrPos = iStrPos + 6;
            } 
        } 

        // Handles " Mc"
        iStrPos = 0;

        if ( icmpW(wsStr[0 .. 2], "MC"w) == 0 && findW(WHITESPACE_W, wsStr[2 .. 3] ) == -1) 
        {
            wsStr[0 .. 2] = "Mc"w;
            wsStr[2] = toupperW(wsStr[2]);  
        } 

        while (iStrPos < iWsStrLen - 1 && iStrPos != -1)
        {
            iStrPos = ifindxW(wsStr, "\u0020MC"w, iStrPos);

            if (iStrPos != -1) 
            {
                if (iStrPos + 4 < iWsStrLen)
                { 
                    wsStr[iStrPos .. iStrPos + 3] = "\u0020Mc"w;
                    wsStr[iStrPos + 3] = toupperW(wsStr[iStrPos + 3]);  
                    iStrPos = iStrPos + 5;
                } 
            } 
        } 
    } 

    return wsStr;

} // end wchar[] asciiProperCaseW( in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.asciiProperCaseW( in wchar[] ).unittest"w );
    
    wchar[] ws1 = "mccain c apPlebEe 123RD p.o. po pob bOx mcgally 32 aNy-PLAce ivx souTH sT aNy'citY,pa usa mmcxv "w;
    wchar[] ws2;

    ws2 = asciiProperCaseW(ws1);
    assert( ws2 != ws1 );
    assert( ws2 == "McCain C Applebee 123rd P.O. PO POB Box McGally 32 Any-Place IVX South St Any'City,Pa Usa MMCXV"w );
}   

/+
 ' wchar[] rjustifyW( in wchar[], in int, in wchar = '\u0020' )
 ' 
 ' Right justify a wide-string with spaces by default
 +/
wchar[] rjustifyW
(
    in wchar[] ws, 
    in int     iWidth,
    in wchar   wsFillChar = '\u0020'
)
{
    wchar[] wsCopy;
    
    if (ws.length >= iWidth)
        return ws;

    wsCopy = new wchar[iWidth];
    
    wsCopy[0 .. iWidth - ws.length] = cast(wchar)wsFillChar;
    wsCopy[iWidth - ws.length .. iWidth] = ws;
    
    return wsCopy;
    
} // wchar[] rjustifyW( in wchar[], in int, in wchar = '\u0020' )

unittest
{
    debug( stringW ) writefln( "stringw.rjustifyW( in wchar[], in int, in wchar = \'\\u0020\' ).unittest"w );

    wchar[] ws1 = "This a test."w;
    wchar[] ws2;
    int     i;

    ws2 = rjustifyW(ws1, 15);
    assert( ws2 == "   This a test."w );
}

/+
 ' wchar[] zfillW( in wchar[], in int )
 ' 
 ' Right justify a wide-string with '0's
 +/
wchar[] zfillW(in wchar[] ws, in int iWidth)
{
    return rjustifyW(ws, iWidth, '0');
    
} // end wchar[] zfillW( in wchar[], in int )

unittest
{
    debug( stringW ) writefln( "stringw.zfillW( in wchar[], in int ).unittest"w );

    wchar[] ws1 = "This a test."w;
    wchar[] ws2;
    int     i;
    
    ws2 = zfillW(ws1, 15);
    assert( ws2 == "000This a test."w );
}

/+
 ' wchar[] ljustifyW( in wchar[], in int, in wchar = '\u0020' )
 ' 
 ' Left justify a wide-string with spaces by default
 +/
wchar[] ljustifyW
(
    in wchar[] ws, 
    in int     iWidth,
    in wchar   wsFillChar = '\u0020'
)
{
    wchar[] wsCopy;
    
    if (ws.length >= iWidth)
        return ws;
        
    wsCopy = new wchar[iWidth];
    wsCopy[0 .. ws.length] = ws;
    wsCopy[ws.length .. iWidth] = cast(wchar)wsFillChar;
    
    return wsCopy;
    
} // end wchar[] ljustifyW( in wchar[], in int, in wchar = '\u0020' )

unittest
{
    debug( stringW ) writefln( "stringw.ljustifyW( in wchar[], in int, in wchar = \'\\u0020\' ).unittest"w );

    wchar[] ws1 = "This a test."w;
    wchar[] ws2;
    int     i;

    ws2 = ljustifyW(ws1, 15);
    assert( ws2 == "This a test.   "w );
}

/+
 ' wchar[] centerW( in wchar[], in int, in wchar = '\u0020' )
 ' 
 ' Center a wide-string with spaces by default
 +/
wchar[] centerW
(
    in wchar[] ws, 
    in int     iWidth,
    in wchar   wsFillChar = '\u0020'
)
{
    wchar[] wsCopy;
    int     iLeft;

    if (ws.length >= iWidth)
        return ws;
        
    wsCopy = new wchar[iWidth];
    
    iLeft = (iWidth - ws.length) / 2;
    wsCopy[0 .. iLeft] = cast(wchar)wsFillChar;
    wsCopy[iLeft .. iLeft + ws.length] = ws;
    wsCopy[iLeft + ws.length .. iWidth] = cast(wchar)wsFillChar;
    
    return wsCopy;
    
} // end wchar[] centerW( in wchar[], in int, in wchar = '\u0020' )

unittest
{
    debug( stringW ) writefln( "stringw.centerW( in wchar[], in int, in wchar = \'\\u0020\' ).unittest"w );

    wchar[] ws1 = "This a test."w;
    wchar[] ws2;
    int     i;

    ws2 = centerW(ws1, 15);
    assert( ws2 == " This a test.  "w);
}

/+
 ' wchar[] toStringW( in bit )
 ' 
 ' Converts a bit to a Wide-Character String
 +/
wchar[] toStringW(in bit b)
{
    return (b ? "true"w : "false"w); 
    
} // end wchar[] toStringW( in bit )

/+
 ' wchar[] toStringW( in wchar )
 ' 
 ' Converts a wchar to a Wide-Character String
 +/
wchar[] toStringW(in wchar wc)
{
    wchar[] wsResult = new wchar[2];
    wsResult[0] = wc;
    wsResult[1] = 0;
    return wsResult[0 .. 1];
    
} // end wchar[] toStringW( in wchar )

unittest
{
    debug( stringW ) writefln( "stringw.toStringW( in wchar ).unittest"w );

    wchar[] ws = "foo"w;
    wchar[] ws2;
    
    foreach (wchar wc; ws)
        ws2 ~= toStringW(wc);
    
    assert( ws2 == "foo"w );
}

/+
 ' wchar[] toStringW( in ubyte )
 ' 
 ' Converts a ubyte to a Wide-Character String
 +/
wchar[] toStringW(in ubyte  ub) { return toStringW(cast(uint) ub); }

/+
 ' wchar[] toStringW( in ushort )
 ' 
 ' Converts a ushort to a Wide-Character String
 +/
wchar[] toStringW(in ushort us) { return toStringW(cast(uint) us); }

/+
 ' wchar[] toStringW( in uint )
 ' 
 ' Converts a uint to a Wide-Character String
 +/
wchar[] toStringW(in uint ui)
{   
    wchar[uint.sizeof * 3] wsBuffer;
    int     idigits;
    wchar   wc;
    wchar[] wsResult;

    idigits = 0;
    
    if (ui < 10)
    {
        // Avoid storage allocation for simple stuff
        wsResult = DIGITS_W[ui .. ui + 1];
    }    
    else
    {
        while(ui)
        {
            wc = (ui % 10) + cast(wchar)'0';
            ui /= 10;
            idigits++;
            wsBuffer[wsBuffer.length - idigits] = wc;
        }
    
        wsResult = new wchar[idigits];
        wsResult[] = wsBuffer[wsBuffer.length - idigits .. wsBuffer.length];
    }
    
    return wsResult;
    
} // end wchar[] toStringW( in uint )

unittest
{
    debug( stringW ) writefln( "stringw.toStringW( in uint ).unittest"w );

    wchar[] ws;

    ws = toStringW(0u);
    assert( ws == "0"w );

    ws = toStringW(9u);
    assert( ws == "9"w );

    ws = toStringW(123u);
    assert( ws == "123"w );
}

/+
 ' wchar[] toStringW( in ulong )
 ' 
 ' Converts a ulong to a Wide-Character String
 +/
wchar[] toStringW( in ulong ul )
{   
    wchar[ulong.sizeof * 3] wsBuffer;
    int     idigits;
    wchar   wc;
    wchar[] wsResult;

    if ( ul < 0x1_0000_0000 )
        return toStringW(cast(uint)ul);
    
    idigits = 0;
    
    while ( ul )
    {
        wc = ( ul % 10 ) + '0';
        ul /= 10;
        idigits++;
        wsBuffer[wsBuffer.length - idigits] = wc;
    }
    
    wsResult = new wchar[idigits];
    wsResult[] = wsBuffer[wsBuffer.length - idigits .. wsBuffer.length];
    
    return wsResult;
    
} // end wchar[] toStringW( in ulong )

unittest
{
    debug( stringW ) writefln( "stringw.toStringW( ulong ).unittest"w );

    wchar[] ws;

    ws = toStringW(0uL);
    assert( ws == "0"w );

    ws = toStringW(9uL);
    assert( ws == "9"w );

    ws = toStringW(123uL);
    assert( ws ==  "123"w );
}

/+
 ' wchar[] toStringW( in byte )
 ' 
 ' Converts a byte to a Wide-Character String
 +/
wchar[] toStringW(in byte  byt) { return toStringW(cast(int)byt); }

/+
 ' wchar[] toStringW( in short )
 ' 
 ' Converts a short to a Wide-Character String
 '
 +/
wchar[] toStringW(in short sht) { return toStringW(cast(int)sht); }

/+
 ' wchar[] toStringW( in int )
 ' 
 ' Converts a int to a Wide-Character String
 +/
wchar[] toStringW(in int i)
{   
    wchar[1 + int.sizeof * 3] wsBuffer;
    wchar   wc;
    wchar[] wsResult;

    if (i >= 0)
        return toStringW(cast(uint)i);

    uint ui      = -i;
    int  idigits = 1;
    
    while (ui)
    {
        wc = (ui % 10) + '0';
        ui /= 10;
        wsBuffer[wsBuffer.length - idigits] = wc;
        idigits++;
    }
    
    wsBuffer[ wsBuffer.length - idigits] = '-';
    wsResult = new wchar[idigits];
    wsResult[] = wsBuffer[wsBuffer.length - idigits .. wsBuffer.length];
    
    return wsResult;
    
} // end wchar[] toStringW( in int )

unittest
{
    debug( stringW ) writefln( "stringw.toStringW( int ).unittest"w );

    wchar[] ws;

    ws = toStringW(0);
    assert( ws == "0"w );

    ws = toStringW(9);
    assert( ws == "9"w );

    ws = toStringW(123);
    assert( ws == "123"w );

    ws = toStringW(-0);
    assert( ws == "0"w );

    ws = toStringW(-9);
    assert( ws == "-9"w );

    ws = toStringW(-123);
    assert( ws == "-123"w );
}

/+
 ' wchar[] toStringW( in long )
 ' 
 ' Converts a long to a Wide-Character String
 +/
wchar[] toStringW(in long l)
{   
    wchar[1 + long.sizeof * 3] wsBuffer;
    wchar   wc;
    wchar[] wsResult;

    if (l >= 0)
        return toStringW(cast(ulong)l);
    
    if (cast(int)l == l)
        return toStringW(cast(int)l);

    ulong ul     = -l;
    int  idigits = 1;
    
    while ( ul )
    {
         wc = (ul % 10 ) + cast(wchar)'0';
         ul /= 10;
         wsBuffer[wsBuffer.length - idigits] = wc;
         idigits++;
    }
    
    wsBuffer[wsBuffer.length - idigits] = '-';
    wsResult = new wchar[idigits];
    wsResult[] = wsBuffer[wsBuffer.length - idigits .. wsBuffer.length];
    
    return wsResult;
    
} // end wchar[] toStringW( in long )

unittest
{
    debug( stringW ) writefln( "stringw.toStringW( long ).unittest"w );

    wchar[] ws;

    ws = toStringW(0L);
    assert( ws == "0"w );

    ws = toStringW(9L);
    assert( ws == "9"w );

    ws = toStringW(123L);
    assert( ws == "123"w );

    ws = toStringW(-0L);
    assert( ws == "0"w );

    ws = toStringW(-9L);
    assert( ws == "-9"w );

    ws = toStringW(-123L);
    assert( ws == "-123"w );
}

/+
 ' wchar[] toStringW( in float )
 ' 
 ' Converts a float to a Wide-Character String
 +/
wchar[] toStringW(in float f) { return toStringW(cast(double)f).dup; }

/+
 ' wchar[] toStringW( in double )
 ' 
 ' Converts a double to a Wide-Character String
 +/
wchar[] toStringW(in double d) { return doFormatW("%g"w, d).dup; }

/+
 ' wchar[] toStringW( in real )
 ' 
 ' Converts a real to a Wide-Character String
 +/
wchar[] toStringW(in real r) { return doFormatW("%g"w, cast(real)r).dup; }

/+
 ' wchar[] toStringW( in ifloat )
 ' 
 ' Converts a ifloat to a Wide-Character String
 +/
wchar[] toStringW(in ifloat f) { return toStringW(cast(idouble)f).dup; }

/+
 ' wchar[] toStringW( in idouble )
 ' 
 ' Converts a idouble to a Wide-Character String
 +/
wchar[] toStringW(in idouble d) { return doFormatW("%gi"w, cast(idouble)d).dup; }

/+
 ' wchar[] toStringW( in ireal )
 ' 
 ' Converts a ireal to a Wide-Character String
 +/
wchar[] toStringW(in ireal r) { return doFormatW("%gi"w, cast(ireal)r).dup; }

/+
 ' wchar[] toStringW( in cfloat )
 ' 
 ' Converts a cfloat to a Wide-Character String
 +/
wchar[] toStringW(in cfloat f) { return toStringW(cast(cdouble) f).dup; }

/+
 ' wchar[] toStringW( in cdouble )
 ' 
 ' Converts a cdouble to a Wide-Character String
 +/
wchar[] toStringW(in cdouble d) { return doFormatW("%g+%gi"w, d.re, d.im).dup; }

/+
 ' wchar[] toStringW( in creal )
 ' 
 ' Converts a creal to a Wide-Character String
 +/
wchar[] toStringW(in creal r)   { return doFormatW("%g+%gi"w, r.re, r.im).dup; }

unittest
{
    wchar[] ws;
    
    debug( stringW ) writefln( "stringw.toStringW( in float ).unittest"w ); 
    ws = toStringW(1234.67694f);
    assert( ws == "1234.68"w );
    ws = toStringW(cast(float)1234E+300);    
    assert( ws == "inf" );
    ws = toStringW(float.min); 
    assert( ws == "1.17549e-38"w );
    ws = toStringW(float.max);    
    assert( ws == "3.40282e+38"w );

    debug( stringW ) writefln( "stringw.toStringW( in ifloat ).unittest"w );
    ws = toStringW(1234.67694fi);
    assert( ws == "1234.68i"w );
    ws = toStringW(cast(ifloat)1234e+300);  
    assert( ws == "0i"w );
    ws = toStringW(ifloat.min); 
    assert( ws == "1.17549e-38i"w );
    ws = toStringW(ifloat.max); 
    assert( ws == "3.40282e+38i"w );
    
    debug( stringW ) writefln( "stringw.toStringW( in cfloat ).unittest"w );
    ws = toStringW(cast(cfloat)1234.67694fi);
    assert( ws == "0+1234.68i"w );
    ws = toStringW(cast(cfloat)1234e+300);
    assert( ws == "inf+0i"w );

    ws = toStringW(cast(cfloat)cfloat.min); 
    assert( ws == "1.17549e-38+1.17549e-38i"w );
    ws = toStringW(cast(cfloat)cfloat.max);   
    assert( ws == "3.40282e+38+3.40282e+38i"w ); 
    
    debug( stringW ) writefln( "stringw.toStringW( in double ).unittest"w );
    ws = toStringW(1234.67694);
    assert( ws == "1234.68"w );
    ws = toStringW(1234e+300);
    assert( ws == "1.234e+303"w );
    ws = toStringW(double.min); 
    assert( ws == "2.22507e-308"w );
    ws = toStringW(double.max); 
    assert( ws == "1.79769e+308"w );
    
    debug( stringW ) writefln( "stringw.toStringW( in idouble ).unittest"w );
    ws = toStringW(1234.67694i);
    assert( ws == "1234.68i"w );
    ws = toStringW(1234e+300i);
    assert( ws == "1.234e+303i"w );
    ws = toStringW(idouble.min); 
    assert( ws == "2.22507e-308i"w );
    ws = toStringW(idouble.max); 
    assert( ws == "1.79769e+308i"w );
   
    debug( stringW ) writefln( "stringw.toStringW( in cdouble ).unittest"w );
    ws = toStringW(cast(cdouble)1234.67694);
    assert( ws == "1234.68+0i"w );
    ws = toStringW(cast(cdouble)1234e+300);
    assert( ws == "1.234e+303+0i"w );
    ws = toStringW(cdouble.min); 
    assert( ws == "2.22507e-308+2.22507e-308i"w );
    ws = toStringW(cdouble.max); 
    assert( ws == "1.79769e+308+1.79769e+308i"w );
    
    debug( stringW ) writefln( "stringw.toStringW( in real ).unittest"w );
    ws = toStringW(1234.95L);
    assert( ws == "1234.95"w );
    ws = toStringW(real.min);
    assert( ws == "3.3621e-4932"w );
    ws = toStringW(real.max);
    assert( ws == "1.18973e+4932"w );
    
    debug( stringW ) writefln( "stringw.toStringW( in ireal ).unittest"w );
    ws = toStringW(1234.67694Li);
    assert( ws == "1234.68i"w );
    ws = toStringW(ireal.min);
    assert( ws == "3.3621e-4932i"w );
    ws = toStringW(ireal.max);
    assert( ws == "1.18973e+4932i"w );
    
    debug( stringW ) writefln( "stringw.toStringW( in creal ).unittest"w );
    ws = toStringW(cast(creal)1234e+12L);
    assert( ws == "1.234e+15+0i"w );
    ws = toStringW(creal.min);
    assert( ws == "3.3621e-4932+3.3621e-4932i"w );
    ws = toStringW(creal.max);
    assert( ws == "1.18973e+4932+1.18973e+4932i"w );     
}

/+
 ' wchar[] toStringW( in wchar * )
 ' 
 ' Converts a wchar * to a Wide-Character String
 +/
wchar[] toStringW(in wchar *ws)
{
    return ws ? ws[0 .. strlenW(ws)] : cast(wchar[])null;
    
} // end wchar[] toStringW( in wchar * )

/+
 ' uint strlenW( in wchar * )
 ' 
 ' Returns the length of a null delimited Wide-Character String.
 +/
private uint strlenW(in wchar *ws) 
{
    uint uix;

    for (uix = 0; ws[uix] != '\u0000'; uix++) {}
          ; // do nothing until a null is found
    return uix;
}

unittest
{
    debug( stringW ) writefln( "stringw.toStringW( in wchar* ).unittest"w );

    wchar[] ws;

    ws = toStringW(null);
    assert( ws == ""w );

    ws = toStringW(cast(wchar *)"foo\0"w);
    assert( ws == "foo"w );
}

/+
 ' wchar[] doFormatW(...)
 ' 
 ' Formats data and returns it as a Wide-Character String
 +/
wchar[] doFormatW(...)
{
    wchar[] ws;

    void add2Buffer(in dchar dc)
    {
        std.utf.encode(ws, dc);
    }

    std.format.doFormat(&add2Buffer, _arguments, _argptr);
        
    return ws; 
    
} // end wchar[] doFormatW(...)

wchar[] toStringW(in long lValue, in uint uiRadix)
in
{
    assert( uiRadix >= 2U && uiRadix <= 36U );
}
body
{
    if (uiRadix == 10U)
        return toStringW(lValue);    // handle signed cases only for radix 10

    return toStringW( cast(ulong)lValue, uiRadix );

} // end wchar[] toStringW( in long, in uint )

wchar[] toStringW(in ulong lValue, in uint uiRadix)
in
{
    assert( uiRadix >= 2U && uiRadix <= 36U );
}
body
{
    wchar[lValue.sizeof * 8U] wsBuffer;
    uint ui = wsBuffer.length;

    if (lValue < uiRadix && lValue < HEXDIGITS_W.length)
        return HEXDIGITS_W[lValue .. lValue + 1L];

    do
    {
        ubyte ubytC;

        ubytC = lValue % uiRadix;
        lValue = lValue / uiRadix;
        ui--;
        wsBuffer[ui] = (ubytC < 10) ? ubytC + '0' : ubytC + 'A' - 10;

    } while (lValue);

    return wsBuffer[ui .. length].dup;

} // end wchar[] toStringW( in ulong, in uint )

unittest
{
    debug( stringW )  writefln( "stringw.toStringW( in long, in uint ).unittest"w );
    debug( stringW )  writefln( "stringw.toStringW( in ulong, in uint ).unittest"w );

    wchar[] wsr;
    int     i;

    wsr = toStringW(-10L, 10u);
    assert( wsr == "-10"w );

    wsr = toStringW(15L, 2u);
    //writefln( "wsr = '%s'", wsr );
    assert( wsr == "1111"w );

    wsr = toStringW(1L, 2u);
    //writefln( "wsr = '%s'", wsr );
    assert( wsr == "1"w );

    wsr = toStringW(0x1234AFL, 16u);
    //writefln( "wsr = '%s'", wsr );
    assert( wsr == "1234AF"w );
}

/+
 ' wchar[] toStringzW( in wchar[] )
 ' 
 ' Returns a Wide-Character String that is null delimited.
 '
 +/
wchar[] toStringzW(in wchar[] ws)
{  
    wchar[] wsz;
    
    if (ws.length == 0)
        return ""w;
        
    if (ws[0] == '\0')
        return ws;
            
    // Check for a null, if found return the original wchar[]
    // else return a copy of it with an appended null. 
    for (uint uix = 1; uix < ws.length; uix++)
        if (ws[uix] == '\0')
            return ws; 

    wsz = ws.dup;
    return (wsz ~ "\0"w);
            
} // end wchar[] toStringzW( in wchar[] )

/+
 ' Function      : wchar[] toSoundexW( in wchar[] )
 ' Version       : v0.1 Beta
 ' Author        : David L. 'SpottedTiger' Davis
 ' Date Created  : 14.Oct.04 Compiled and Tested with dmd v0.102
 ' Date Modified : (None)
 '               :
 ' Licence       : Pubic Domain
 ' --------------------------------------------------------------
 ' Soundex Rules based on this Web Site:
 ' http://www.archives.gov/research_room/genealogy/census/soundex.html
 ' --------------------------------------------------------------
 '
 ' Return the soundex code for a given 
 ' string based upon its phonetic sound.
 +/
wchar[] toSoundexW(in wchar[] ws) 
{
    wchar[] wsWord = toupperW(stripW(ws));
    wchar[] wsSoundex;
    wchar   wcPrevVal;
    wchar   wcPrevVal2;

    /+ Return a converts (A thru Z) ASCII alphabetic  
     ' characters to their appropriate numeric Soundex
     ' Code value or special character.
     '         A,E,I,O,U,Y     = '/'
     '         H,W             = '*'
     '         B,F,P,V         = '1'
     '         C,G,J,K,Q,S,X,Z = '2'
     '         D,T             = '3'
     '         L               = '4'
     '         M,N             = '5'
     '         R               = '6'
     '         (Ignore these)  = '-'
     +/
    static char getSoundexFormattingCodeW(in wchar wc) 
    {
        switch (wc) 
        {
          case 'A','E','I','O','U','Y': 
              return '/';  
              
          case 'H','W': 
              return '*'; 
              
          case 'B','F','P','V': 
              return '1';
              
          case 'C','G','J','K','Q','S','X','Z':
              return '2';
              
          case 'D','T':
              return '3';
        
          case 'L':
              return '4';
            
          case 'M','N':
              return '5';
            
          case 'R':
              return '6';
            
          default:    
              break;
        }

        return '-'; 
        
    } // end wchar getSoundexFormatingCodeW( in wchar )    
    
    // Keep the first alphabetic character
    wsSoundex.length = 1;
    wsSoundex[0] = wsWord[0];
    
    // Transform the characters to a Soundex Code or 
    // a special processing character.
    for (uint uix = 0; uix < wsWord.length; uix++)
        wsWord[uix] = getSoundexFormattingCodeW(wsWord[uix]);
    
    // Apply a number of the Soundex rules for double characters
    // and side-to-side Soundex code issues
    for (uint uix = 1; uix < wsWord.length; uix++)
    {
        //writef( "%s", wsWord[ uix ] );

        if (wsWord[uix - 1] == wsWord[uix] || wcPrevVal == wsWord[uix])
        {
           wcPrevVal     = wsWord[uix];
           wsWord[uix] = '-';
        }    
        
        if (uix - 2 >= 0)
            if (wcPrevVal2 == '*')
                if (wsWord[uix - 2] == wsWord[uix] )
                    wsWord[uix] = '-';

        if ( wsWord[uix] == '/' && wcPrevVal != '*')
            wcPrevVal = '\u0020';

        wcPrevVal2 = wsWord[uix];
    }	

    // Filter out the unwanted character to create the final Soundex String
    for (uint uix = 1; uix < wsWord.length; uix++)
        if (wsWord[uix] != '/' && wsWord[uix] != '*' && wsWord[uix] != '-')
            wsSoundex ~= wsWord[uix];            

    // If Soundex Code is shorter then four characters append four zeros.
    if (wsSoundex.length < 4)
        wsSoundex ~= "0000"w;
    
    //writefln();
    
    // Return only the first four characters
    return wsSoundex[0 .. 4].dup; 
    
} // end wchar[] toSoundexW( in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.toSoundexW( in wchar[] ).unittest"w );

    // Sound-Alike Names 
    assert( "F626"w == toSoundexW( "fraser"w ) );
    assert( "F626"w == toSoundexW( "frazier"w ) );
    
    assert( "V230"w == toSoundexW( "Vogt"w ) );
    assert( "V230"w == toSoundexW( "Voght"w ) );
    assert( "V230"w == toSoundexW( "Voight"w ) );
    
    assert( "F230"w == toSoundexW( "Fogt"w ) );
    assert( "F230"w == toSoundexW( "Foght"w ) );
    assert( "F230"w == toSoundexW( "Foight"w ) );
    //----------------------------------------
    
    // Normal examples
    assert( "W252"w == toSoundexW( "Washington"w )  );
    assert( "E460"w == toSoundexW( "Euler"w ) );
    assert( "H416"w == toSoundexW( "Hilbert"w ) );
    assert( "K530"w == toSoundexW( "Knuth"w ) );
    assert( "L222"w == toSoundexW( "Lukasiewicz"w ) );
    assert( "V532"w == toSoundexW( "VanDeusen"w ) );
    assert( "N253"w == toSoundexW( "Naesmyth"w ) );  
    assert( "B630"w == toSoundexW( "Baird"w ) );

    // When 2 or more of the same key letters (or their number 
    // equivalents) come together, they should be treated as 
    // ONE key letter.
    assert( "J250"w == toSoundexW( "Jackson"w ) );
    assert( "S163"w == toSoundexW( "Sheppard"w ) );
    assert( "S200"w == toSoundexW( "Sacks"w ) );
    assert( "S640"w == toSoundexW( "Sherrell"w ) );
    assert( "L300"w == toSoundexW( "Lloyd"w ) );
    assert( "G362"w == toSoundexW( "Gutierrez"w ) ); 
    assert( "C450"w == toSoundexW( "Callahan"w ) );
    assert( "G200"w == toSoundexW( "Gauss"w ) );

    // A key letter or its equivalent immediately following an 
    // initial letter of the same value should be disregarded 
    // in coding.
    assert( "P236"w == toSoundexW( "Pfister"w ) );
    assert( "S232"w == toSoundexW( "Schechwitz"w ) );
    assert( "S240"w == toSoundexW( "Szakal"w ) );
    assert( "S460"w == toSoundexW( "Scklar"w ) ); 
    
    // Key letters or their equivalents separated by H or W 
    // are coded as one key letter.   
    assert( "S200"w == toSoundexW( "Sokwzy"w ) ); 
    assert( "S452"w == toSoundexW( "Schkolink"w ) ); 
    assert( "S460"w == toSoundexW( "Schklar"w ) ); 
    assert( "A261"w == toSoundexW( "Ashcraft"w ) );
    assert( "S432"w == toSoundexW( "Schultz"w ) );
    
    // When a repeated key letter or its equivalent is separated 
    // by an A, E, I, O, U or Y, the key letters or their 
    // equivalents are considered separately.
    assert( "T522"w == toSoundexW( "Tymczak"w ) );
    assert( "S335"w == toSoundexW( "Staten"w ) ); 
    assert( "S422"w == toSoundexW( "Salkiewicz"w ) ); 
    assert( "S550"w == toSoundexW( "Simone"w ) ); 
    
    // Some names do not contain any of the six key letters or their 
    // equivalents - that is they have only vowels A, E, I, O, U or 
    // W, H, Y, after the first letter of the name. Such names are 
    // filed in a group by themselves in back of their letter guide.
    assert( "S000"w == toSoundexW( "Shea"w ) ); 
    assert( "L000"w == toSoundexW( "Lee"w ) );
    
    // Testing variations of the Womack surname 
    assert( "W520"w == toSoundexW( "Wamac"w ) ); 
    assert( "W520"w == toSoundexW( "Wamack"w ) ); 
    assert( "W520"w == toSoundexW( "Wammack"w ) ); 
    assert( "W520"w == toSoundexW( "Wammuck"w ) ); 
    assert( "W520"w == toSoundexW( "Womac"w ) ); 
    assert( "W520"w == toSoundexW( "Womach"w ) ); 
    assert( "W520"w == toSoundexW( "Womich"w ) ); 
    assert( "W520"w == toSoundexW( "Womick"w ) ); 
    assert( "W520"w == toSoundexW( "Wommac"w ) ); 
    assert( "W520"w == toSoundexW( "Wommack"w ) );    
    assert( "W520"w == toSoundexW( "Wommuck"w ) );    
    assert( "W520"w == toSoundexW( "Wonack"w ) );    
    assert( "W520"w == toSoundexW( "Woomack"w ) );     
}

/+
 ' wchar[] capwordsW( in wchar[] )
 '
 ' Capitalize all words in string.
 ' Remove leading and trailing whitespace.
 ' Replace all sequences of whitespace with a single space.
 +/
wchar[] capwordsW(in wchar[] ws)
{
    wchar[] wst = stripW(ws).dup;
    wchar[] wsr; 
    wchar   wc;
    bool    bPrevWhiteSpace = false;
    bool    bInWord = false;

    for (uint uix = 0; uix < wst.length; uix++)
    {
        switch (wst[uix])
        {
           case '\u0020':
           case '\u0009':
           case '\u000C':
           case '\u000D':
           case '\u000A':
           case '\u000B':

              if (!bPrevWhiteSpace)
              {
                  wsr ~= '\u0020';
                  bPrevWhiteSpace = true;
              }
              
              break;

           default:

               wsr ~= wst[uix];
               bPrevWhiteSpace = false;
               break;
        }
    }

    // Go back through wsr and capitalize the words
    bInWord = false;

    for (uint uix = 0; uix < wsr.length; uix++)
    {
        wc = wsr[uix];

        // If within ASCII 'A' thru 'Z' range, 
        // convert the character to lowercase.
        if (wc >= '\u0041' && wc <= '\u005A')
        {
            if (bInWord)
            {
                wsr[uix] = tolowerW(wc);
            }

            bInWord = true;
        }
        // If within ASCII 'a' thru 'z' range, 
        // convert the character to uppercase.
        else if (wc >= '\u0061' && wc <= '\u007A')
        {
            if (!bInWord)
            {
                wsr[uix] = toupperW(wc);
            }

            bInWord = true;
        }
        else
            bInWord = false;
    }

    return wsr;

} // end wchar[] capwordsW( in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.capwordsW( in wchar[] ).unittest"w );

    wchar[] ws1 = "\tfoo abc(aD)*  \t  (q PTT  "w;
    wchar[] ws2;

    ws2 = capwordsW(ws1);
    //printf("s2 = '%.*s'\n", s2);
    assert( ws2 == "Foo Abc(Ad)* (Q Ptt"w );
}

/+
 ' wchar[] joinW( in wchar[][], in wchar[] )
 '
 ' Concatenate all the strings together into one
 ' string; use wsSep[] as the separator.
 +/
wchar[] joinW(in wchar[][] wsWords, in wchar[] wsSep)
{
    uint uiLen;
    uint uiSepLen;
    uint uix;
    uint uiy;
    wchar[] wsResult;
    uint uiWordLen;

    if (wsWords.length)
    {
        uiLen = 0;
        for (uix = 0; uix < wsWords.length; uix++)
            uiLen += wsWords[uix].length;

        uiSepLen = wsSep.length;
        uiLen += (wsWords.length - 1) * uiSepLen;

        wsResult = new wchar[uiLen];

        uix = 0;

        while (true)
        {
            uiWordLen = wsWords[uix].length;

            wsResult[uiy .. uiy + uiWordLen] = wsWords[uix];
            uiy += uiWordLen;
            uix++;

            if (uix >= wsWords.length)
                break;

            wsResult[uiy .. uiy + uiSepLen] = wsSep;
            uiy += uiSepLen;
        }
        assert( uiy == uiLen );
    }

    return wsResult;

} // end wchar[] joinW( in wchar[][], in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.joinW( in wchar[], in wchar[] ).unittest"w );

    wchar[] wsWord1 = "peter"w;
    wchar[] wsWord2 = "paul"w;
    wchar[] wsWord3 = "jerry"w;
    wchar[][3] wsWords;
    wchar[] wsR;

    wsWords[ 0 ] = wsWord1;
    wsWords[ 1 ] = wsWord2;
    wsWords[ 2 ] = wsWord3;
    wsR = joinW(wsWords, ","w);
    assert( wsR == "peter,paul,jerry"w );
}

/+
 ' wchar[][] splitW( in wchar[] )
 '
 ' Split ws[] into an array of words,
 ' using whitespace as the delimiter.
 +/
wchar[][] splitW(in wchar[] ws)
{
    uint ui;
    uint uiStart;
    bool bInWord = false;
    wchar[][] wsWords;

    for (ui = 0; ui < ws.length; ui++)
    {
        switch ( ws[ ui ] )
        {
           case '\u0020':
           case '\u0009':
           case '\u000C':
           case '\u000D':
           case '\u000A':
           case '\u000B':
               if (bInWord)
               {
                   wsWords ~= ws[uiStart .. ui];
                   bInWord = false;
               }
               break;

           default:
               if (!bInWord)
               {   
                   uiStart = ui;
                   bInWord = true;
               }
               break;
        }
    }

    if (bInWord == true)
        wsWords ~= ws[uiStart .. ui];

    return wsWords;

} // end wchar[][] splitW( in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.splitW( in wchar[] ).unittest"w );

    wchar[]   ws = " peter paul\tjohn "w;
    wchar[][] wsWords;

    wsWords = splitW(ws);
    assert( wsWords.length == 3 );
    assert( wsWords[ 0 ] == "peter"w );
    assert( wsWords[ 1 ] == "paul"w );
    assert( wsWords[ 2 ] == "john"w );
}

/+
 ' wchar[][] splitlinesW( in wchar[] )
 '
 ' Split s[] into an array of lines,
 ' using CR, LF, or CR-LF as the delimiter.
 ' ('\r' = '\u000D' and '\n' = '\u000A')
 +/
wchar[][] splitlinesW(in wchar[] ws)
{
    uint  ui;
    uint  uiStart;
    uint  uiNumLines;
    wchar[][] wsLines;
    wchar wc;

    uiNumLines = 0;

    for (ui = 0; ui < ws.length; ui++)
    {
        wc = ws[ui];
        if (wc == '\u000D' || wc == '\u000A')
        {
            uiNumLines++;
            uiStart = ui + 1;
            if (wc == '\u000D' && ui + 1 < ws.length && ws[ui + 1] == '\u000A')
            {
                ui++;
                uiStart++;
            }
        }
    }

    if (uiStart != ui)
        uiNumLines++;

    wsLines = new wchar[][uiNumLines];
    uiNumLines = 0;
    uiStart = 0;

    for (ui = 0; ui < ws.length; ui++)
    {
        wc = ws[ui];
        if (wc == '\u000D' || wc == '\u000A')
        {
            wsLines[uiNumLines] = ws[uiStart .. ui];
            uiNumLines++;
            uiStart = ui + 1;
            if (wc == '\u000D' && ui + 1 < ws.length && ws[ui + 1] == '\u000A')
            {
                ui++;
                uiStart++;
            }
        }
    }

    if (uiStart != ui)
    {
        wsLines[uiNumLines] = ws[uiStart .. ui];
        uiNumLines++;
    }

    assert( uiNumLines == wsLines.length );
    return wsLines;

} // end wchar[][] splitlinesW( in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.splitlinesW( in wchar[][] ).unittest"w );

    wchar[] ws = "\rpeter\n\rpaul\r\njohn\n"w;
    wchar[][] wsLines;
    int i;

    wsLines = splitlinesW(ws);
    assert( wsLines.length == 5 );
    assert( wsLines[0].length == 0 );
    assert( wsLines[1] == "peter"w );
    assert( wsLines[2].length == 0 );
    assert( wsLines[3] == "paul"w );
    assert( wsLines[4] == "john"w );

    ws = ws[0 .. ws.length - 1];  // lop off trailing \n
    wsLines = splitlinesW(ws);
    assert( wsLines.length == 5 );
    assert( wsLines[4] == "john"w );
}

/+
 ' wchar[] replaceSliceW( in wchar[], in wchar[], in wchar[] )
 '
 ' Return a string that is wsString[] with wsSlice[] replaced by wsReplacement[].
 +/
wchar[] replaceSliceW(in wchar[] wsString, in wchar[] wsSlice, in wchar[] wsReplacement)
in
{
    // Verify that slice[] really is a slice of wsString[]
    int iSo = cast(wchar *)wsSlice - cast(wchar *)wsString;
    assert( iSo >= 0 );
    //writefln( "string.length = %d, so = %d, slice.length = %d", string.length, so, slice.length );
    assert( wsString.length >= iSo + wsSlice.length );
}
body
{
    wchar[] wsResult;
    int iSo = cast(wchar *)wsSlice - cast(wchar *)wsString;

    wsResult.length = wsString.length - wsSlice.length + wsReplacement.length;

    wsResult[0 .. iSo] = wsString[0 .. iSo];
    wsResult[iSo .. iSo + wsReplacement.length] = wsReplacement;
    wsResult[iSo + wsReplacement.length .. wsResult.length] = wsString[iSo + wsSlice.length .. wsString.length];

    return wsResult;

} // end wchar[] replaceSliceW( in wchar[], in wchar[], in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.replaceSliceW( in wchar[], in wchar[], in wchar[] ).unittest"w );

    wchar[] wsString = "hello"w;
    wchar[] wsSlice = wsString[2 .. 4];

    wchar[] wsR = replaceSliceW(wsString, wsSlice, "bar"w);
    assert( wsR == "hebaro"w );
}

/+
 ' wchar[] insertW( in wchar[], in int, in wchar[] )
 '
 ' Insert sub[] into s[] at location index.
 +/
wchar[] insertW(in wchar[] ws, in int iIndex, in wchar[] wsSub)
in
{
    assert( 0 <= iIndex && iIndex <= ws.length );
}
body
{
    if (wsSub.length == 0)
        return ws;

    if (ws.length == 0)
        return wsSub;

    int iNewLength = ws.length + wsSub.length;
    wchar[] wsResult = new wchar[iNewLength];

    wsResult[0 .. iIndex] = ws[0 .. iIndex];
    wsResult[iIndex .. iIndex + wsSub.length] = wsSub;
    wsResult[iIndex + wsSub.length .. iNewLength] = ws[iIndex .. ws.length];
    return wsResult;

} // end wchar[] insertW( in wchar[], in int, in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.insert( in wchar[], in int, in wcahr[] ).unittest"w );

    wchar[] wsR;

    wsR = insertW("abcd"w, 0, "e"w);
    assert( wsR == "eabcd"w);

    wsR = insertW("abcd"w, 4, "e"w);
    assert( wsR == "abcde"w );

    wsR = insertW("abcd"w, 2, "ef"w);
    assert( wsR == "abefcd"w );

    wsR = insertW(null, 0, "e"w);
    assert( wsR == "e"w );

    wsR = insertW("abcd"w, 0, null);
    assert( wsR == "abcd"w );
}

/+
 ' wchar[] expandtabsW( in wchar[], in int )
 '
 ' Replace tabs with the appropriate number of spaces.
 ' tabsize is the distance between tab stops.
 +/
wchar[] expandtabsW(in wchar[] ws, in int iTabSize)
{
    wchar[] wsR;
    wchar   wc;
    int     i;
    int     iStart;
    int     iTabStop;
    int     iCol;
    int     iN;

    static wchar[8] wsSpaces = "        "w;

    iCol = 0;
    for (i = 0; i < ws.length; i++)
    {
        wc = ws[i];
        if (wc == '\u0009')  //'\t'
        {   
            wsR ~= ws[iStart .. i];
            iStart = i + 1;

            iTabStop = iCol + iTabSize;
            iTabStop -= iTabStop % iTabSize;

            while (iCol < iTabStop)
            {
                iN = iTabStop - iCol;
                if (iN > wsSpaces.length)
                    iN = wsSpaces.length;
                wsR ~= wsSpaces[0 ..  iN];
                iCol += iN;
            }
        }
        else
        {
            iCol++;
        }
    }

    wsR ~= ws[iStart .. i];
    return wsR;

} // end wchar[] expandtabsW( in wchar[], in int )

unittest
{
    debug( stringW ) writefln( "stringw.expandtabsW( in wchar[], in int ).unittest"w );

    wchar[] ws = "This \tis\t a fofof\tof list"w;
    wchar[] wsR;
    int i;

    wsR = expandtabsW(ws, 8);
    assert( wsR == "This    is       a fofof        of list"w );
}

/+
 ' wchar[] maketransW( in wchar[], in wchar[] )
 '
 ' Construct translation table for translate().
 +/
wchar[] maketransW(in wchar[] wsFrom, in wchar[] wsTo)
in
{
    assert( wsFrom.length == wsTo.length );
}
body
{
    wchar[] wsT = new wchar[256];
    int i;

    for (i = 0; i < 256; i++)
        wsT[i] = cast(wchar)i;

    for (i = 0; i < wsFrom.length; i++)
        wsT[wsFrom[i]] = wsTo[i];

    return wsT;

} // end wchar[] maketransW( in wchar[], in wchar[] )

/+
 ' wchar[] translateW( in wchar[], in wchar[], in wchar[] )
 ' 
 ' Translate characters in s[] using table created by maketrans().
 ' Delete chars in delchars[].
 +/
wchar[] translateW(in wchar[] ws, in wchar[] wsTransTab, in wchar[] wsDelChars)
in
{
    assert( wsTransTab.length == 256 );
}
body
{
    wchar[] wsR;
    wchar   wc;
    int     i;
    int     iCount;

    bit[256] bDelTab;

    bDelTab[] = false;

    for (i = 0; i < wsDelChars.length; i++)
    {
        bDelTab[wsDelChars[i]] = true;
    }

    iCount = 0;
    for (i = 0; i < ws.length; i++)
    {
        if ( !bDelTab[ws[i]] )
            iCount++;
        //writefln( "ws[%d]='%c', iCount=%d\n", i, ws[i], iCount);
    }

    wsR = new wchar[iCount];
    iCount = 0;
    for (i = 0; i < ws.length; i++)
    {   
        wc = ws[i];

        if ( !bDelTab[wc] )
        {
            wsR[iCount] = wsTransTab[wc];
            iCount++;
        }
    }

    return wsR;

} // end wchar[] translateW( in wchar[], in wchar[], in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.translateW( in wchar[], in wchar[], in wchar[] ).unittest"w );

    wchar[] wsFrom = "abcdef"w;
    wchar[] wsTo   = "ABCDEF"w;
    wchar[] ws     = "The quick dog fox"w;
    wchar[] wsT;
    wchar[] wsR;

    wsT = maketransW(wsFrom, wsTo);
    wsR = translateW(ws, wsT, "kg"w);
    //writefln( "wsR =%s", wsR );
    assert( wsR == "ThE quiC Do Fox"w );
}

/+
 ' Function      : wchar[] addOrdSuffixW( in ulong )
 ' Author        : David L. 'SpottedTiger' Davis
 ' Date Created  : 25.May.02 in VB6
 ' Date Modified : 15.Nov.04 rewritten in D
 ' Licence       : Pubic Domain
 +/
wchar[] addOrdSuffixW(in ulong ulNumber)
{
    wchar[] wsNumber = toStringW(ulNumber);
    
    switch (wsNumber[wsNumber.length - 1 .. wsNumber.length])
    {
      case "1"w: return wsNumber ~ "st"w;
      case "2"w: return wsNumber ~ "nd"w;
      case "3"w: return wsNumber ~ "rd"w;
      default  : return wsNumber ~ "th"w;
    }
   
} // end wchar[] addOrdSuffixW( in ulong )

unittest
{
    debug( stringW ) writefln( "stringw.addOrdSuffixW( in ulong ).unittest"w );
    assert( addOrdSuffixW(123) == "123rd"w );
    assert( addOrdSuffixW(21) == "21st"w );
    assert( addOrdSuffixW(1022) == "1022nd"w );
    assert( addOrdSuffixW(1024) == "1024th"w );

    assert( addOrdSuffixW(long.max) == "9223372036854775807th"w );
    assert( addOrdSuffixW(ulong.min) == "0th"w );
    assert( addOrdSuffixW(ulong.max) == "18446744073709551615th"w );
}

/+
 ' Function     : bool isDigitsW( in wchar[] )
 '              :           
 '              : Checks a string to see if all it's characters 
 '              : are digits, and if so it returns true.
 '              :
 ' Author       : David L. 'SpottedTiger' Davis
 ' Date Created : 19.Jun.04
 ' Language     : DigitalMars D v0.92
 '              :
 ' Requirements : none
 +/
bool isDigitsW(in wchar[] ws)
{
    foreach (dchar dc; ws)
        if (findW( DIGITS_W, cast(wchar)dc) == -1) 
            return false;
        
    return true;   
     
} // end bool isDigitsW( in wchar[] )

unittest
{
    debug( stringW ) writefln( "stringw.isDigitsW( in wchar[] ).unittest"w );

    assert( isDigitsW("123"w) == true );
    assert( isDigitsW("0x59y"w) == false ); 
}
