/+
 ' Source          : toNumeric.d - Convert a numeric string 
 '                 :               into a boxed numeric value.
 ' Version         : v1.0 
 ' Author          : David L. 'SpottedTiger' Davis
 ' Source Platform : Intel x86 / WinXP SP2 
 ' Target Platform : Should work on any CPU / OS combo that D compiles on.
 '                 :
 ' Created Date    : 02.Jun.05 Function was requested by Andrew 
 '                 :           Fedoniouk for use with the std.boxer
 '                 :           module. Compiled and tested with dmd v0.125.
 ' Modified Date   : 03.Jun.05 Corrected a number of issues, and
 '                 :           added support for the complex floats. 
 '                 : 07.Jun.05 Added a isNumberic() function that returns
 '                 :           flags that help decide how to box the 
 '                 :           numeric datatype. Before the both isNumeric()
 '                 :           and toNumeric() functions were doing double 
 '                 :           checking, which was unnecessary.
 '                 : 09.Jun.05 Recompiled and unittested under D v0.126. 
 '                 : 10.Sep.05 Added updated isnumeric.d and recompiled
 '                 :           and tested dmd v0.131.
 '                 : 26.Oct.05 Recompiled and Tested with dmd v0.137.    
 '                 : 30.Dec.05 Recompiled and Tested with dmd v0.142.   
 '                 : 23.Mar.06 Recompiled and Tested with dmd v0.150. 
 '                 : 28.May.06 Recompiled and Tested with dmd v0.159. 
 '                 : 19.Jul.06 All "Imports" are now private by default 
 '                 :           (which shouldn't be a problem, since I
 '                 :           always use the private keyword) in 
 '                 :           dmd v0.163, recompiled and tested.
 '                 : 23.Oct.06 Recompiled and Tested with dmd v0.172.
 '                 : 09.Dec.06 Because of "improved name mangling" changes done in dmd v0.176,
 '                 : had to change "extern (C) void assert_3std5boxer() { }" to
 '                 : "extern (C) int D3std5boxer8__assertFiZv() { return 0; }", then
 '                 : recompiled and unittested with dmd v0.177.
 '                 : 05.Jan.07 Recompiled and Tested with dmd v1.0.
 '                 : 03.Jan.08 Recompiled and Tested with dmd v1.025. 
 '                 :           Commented "extern (C) int D3std5boxer8__assertFiZv() { return 0; } // dmd v0.176+"
 '                 : 20.Mar.08 Recompiled and Tested with dmd v1.028.
 '                 : 20.Jun.08 Recompiled and Tested with dmd v1.031. 
 '                 : 17.Jul.08 Recompiled and Tested with dmd v1.033. 
 '                 : 13.May.09 Recompiled and Tested with dmd v1.045.
 '                 :
 ' License         : Public Domain
 ' -----------------------------------------------------------
 ' To Debug:   dmd tonumeric.d isnumeric.d -debug=tonumeric 
 ' To Compile: dmd mycode.d tonumeric.d isnumeric.d
 +/
debug( tonumeric ) private import scomp = std.compiler;
private import std.stdio;
private import std.boxer; 
private import std.string;
private import std.conv;
/+
 ' DLD 06.Sep.05
 ' ------------------------------------------------------
 ' Once std.string is updated with the newer version of  
 ' isNumeric(), I'll remove this import as a requirement.
 +/
private import isnumeric;

// Burton Radons <burton-radons[at]smocky[dot]com>
// This entry allows -unittest for testing, and/or no need for the -release switch
// extern (C) void assert_3std5boxer() { } // dmd v0.175 and below
// extern (C) int D3std5boxer8__assertFiZv() { return 0; } // dmd v0.176+

Box toNumeric(in char[] s, in bool bAllowSep = false)
{
    Box    boxx;
    char[] sx        = std.string.tolower(s);
    char[] s1        = "";
    char[] s2        = "";
    int    iLen      = s.length;
    bool   bComplex  = false;
    bool   bNeg      = false;
    bool   bISuffix  = false;
    bool   bInt      = true;
    int    j         = 0;
    char   c;
    long   l         = 0L;
    ulong  ul        = 0UL;
    real   r1        = real.nan;
    real   r2        = real.nan;
    ireal  ir        = ireal.nan; 
    creal  cr        = creal.nan;
    bool[] bflags;

    sx = std.string.strip(sx);
    
    // Check first for NaN and Infinity
    if (sx == "nan")
    { boxx = box(float.nan); return boxx; }
    else if (sx == "inf")
    { boxx = box(float.infinity); return boxx; }
    else if (sx == "-inf")
    { boxx = box(-float.infinity); return boxx; }   
    else if (sx == "nani") // || sx == "0i")
    { boxx = box(ifloat.nan); return boxx; }
    else if (sx == "nan+nani" || sx == "nan+0i")
    { boxx = box(cfloat.re.nan+cfloat.im.nan); return boxx; }

    // If a valid numeric string, set flags before boxing,
    // else throw an Exception if its not a numeric string.
    if (isnumeric.isNumeric(sx, bflags, bAllowSep))
    {
        bInt     = bflags[0];
        bNeg     = bflags[1];
        bISuffix = bflags[2];
        bComplex = bflags[3];
    }
    else
    {
        throw new Exception("Error in toNumeric(in char[], " ~
                            "in bool = false), the first parameter " ~ 
                            "\"" ~ s ~ "\" is not a numeric string!");
    }
   
    // Remove Separators that may exist, 
    // if bAllowSep was set to true
    if (bAllowSep)
    {
        sx = std.string.replace(sx, "_", "");
        sx = std.string.replace(sx, ",", "");
    }

    // Set to sx's current length
    iLen = sx.length;
    
    // Removing the suffixes, even the 'i' before calling
    // the conversion functions, else an error with occur.
    for (int i = sx.length - 2; i < sx.length; i++)
    {
        c = sx[i];
        if (c == 'f' || c == 'l' || 
            c == 'u' || c == 'i')
            iLen--;
    }
    
    //writefln("sx=\"%s\", bNeg=%b, bInt=%b, bISuffix=%b, bComplex=%b", 
    //         sx, bNeg, bInt, bISuffix, bComplex);
    
    // Adjust sx's length to iLen
    sx.length = iLen;

    // Positive Whole Number
    if (bInt && !bNeg)
    {
        //writefln("Unsigned Int");
        ul = std.conv.toUlong(sx);
        if (ul <= ubyte.max)
            boxx = box(cast(ubyte)ul);
        else if (ul <= ushort.max)
            boxx = box(cast(ushort)ul);
        else if (ul <= uint.max)
            boxx = box(cast(uint)ul);
        else
            boxx = box(ul);
    }
    // Negative Whole Numer
    else if (bInt && bNeg)
    {
        //writefln("Signed Int");
        l = std.conv.toLong(sx);
        if (l >= byte.min && l <= byte.max)
            boxx = box(cast(byte)l);
        else if (l >= short.min && l <= short.max)
            boxx = box(cast(short)l);
        else if (l >= int.min && l <= int.max)
            boxx = box(cast(int)l);
        else
            boxx = box(cast(long)l);
    }
    // Regular Floating-Point Number
    else if (!bISuffix && !bComplex)
    {
        //writefln("Regular Float");
        r1 = std.conv.toReal(sx);
        if (r1 >= float.min && r1 <= float.max)
            boxx = box(cast(float)r1);
        else if (r1 >= double.min && r1 <= double.max)
            boxx = box(cast(double)r1);
        else
            boxx = box(r1);
    }
    /****************************************************/
    /*  ** Walter is still working on putting in the ** */
    /*  ** toIfloat() thru toCfloat() functions.     ** */
    /****************************************************/
    // Imaginary Floating-Point Number
    else if (bISuffix && !bComplex)
    {
        //writefln("Imaginary Float");

        //ir = std.conv.toIreal(sx);
        //if (ir >= ifloat.min && ir <= ifloat.max)
        //    boxx = box(cast(ifloat)ir);
        //else if (ir >= idouble.min && ir <= idouble.max)
        //    boxx = box(cast(idouble)ir);
        //else
        //    boxx = box(ir);

        //** A temp fix until the toIreal() is ready in std.conv **
        r1 = std.conv.toReal(sx);
        if (r1 >= float.min && r1 <= float.max)
            boxx = box(cast(ifloat)(r1 * 1.0i));
        else if (r1 >= double.min && r1 <= double.max)
            boxx = box(cast(idouble)(r1 * 1.0i));
        else
            boxx = box(cast(ireal)(r1 * 1.0i));
    }
    // Complex Floating-Point Number
    else if (bComplex)
    {
        //writefln("Complex Float");
        //cr = std.conv.toCreal(sx);
        //boxx = box(cr);
        
        //** A temp fix until the toCreal() is ready in std.conv **
        .splitComplexStrings(sx, s1, s2);
        //writefln("sx=\"%s\", s1=\"%s\", s2=\"%s\"", sx, s1, s2);
        r1 = std.conv.toReal(s1);
        r2 = std.conv.toReal(s2);
        cr = cast(creal)(r1 + (r2 * 1.0i));
        boxx = box(cr);
    }
    else
    {
        //writefln("Invalid numeric string");
        throw new Exception("Error in toNumeric(in char[], " ~
                            "in bool = false), the first parameter " ~ 
                            s ~ " is not a numeric string!");
    }

    return boxx;
}

/+
 ' Splits a complex floats (cfloat, cdouble, and creal) into two workable strings.
 ' Grammar:
 ' (for cfloat, cdouble, and creal)
 ' ['+'|'-']digit(s)[.][digit(s)][[e-|e+]digit(s)][+]
 '         [digit(s)[.][digit(s)][[e-|e+]digit(s)][i|f|L|Li|fi]]
 '      or [nan|nani|nan+nani|inf|-inf]
 '
 ' creal Range: from 3.362103E-4932+3.362103E-4932i (min) 
 '                to 1.189731E+4932+1.189731E+4932i (max)
 +/
bool splitComplexStrings(in char[] s, out char[] s1, out char[] s2)
{
    char[] sx   = s;
    int    iLen = s.length;
    int    plus = 0;
    
    s1 = "";
    s2 = "";
    
    if (!iLen)
        throw new Exception("spiitComplexStrings(): " ~
                            " input string is an empty string.");

    if (iLen >= 2)
    {
        sx = std.string.tolower(s);   
        // When "nan" or "nani" just return them.
        if (sx == "nan" || sx == "nani" ||
            sx == "nan+nani" || sx =="nan+0i")
        { s1 = "nan"; s2 = "nani"; return 1; }
    
        if (sx == "inf")
        { s1 = "inf"; s2 = "infi"; return 1; }    
    
        if (sx == "-inf")
        { s1 = "-inf"; s2 = "-infi"; return 1; }    
    }
    
    // Split the original string out into two strings.
    for (int i = 1; i < iLen; i++)
        if ((sx[i - 1] != 'e' && sx[i - 1] != 'E') && sx[i] == '+')
        {
            s1 = sx[0..i];
            if (i + 1 < iLen - 1)
                s2 = sx[i + 1..iLen]; // - 1];
            else 
                s2 = "0e+0"; //"0e+0i";
            
            break;
        }   

    // Handle the case when there's only a single value 
    // to work with, and set the other string to zero.
    if (!s1.length)
    { s1 = sx; s2 = "0e+0"; } //"0e+0i"; }
 
    //writefln( "splitComplexStrings() s=\"%s\", s1=\"%s\", s2=\"%s\", len=%d", 
    //           s, s1, s2, len );
   
    return true;
}

debug( tonumeric )
{
void main()
{
    Box wx;

    wx = toNumeric("123456");
    writefln("Box wx = toNumeric(\"123456\");");
    writefln("unbox!(uint)(wx)=%d", unbox!(uint)(wx));
    writefln();

    wx = toNumeric("-124599");
    writefln("Box wx = toNumeric(\"-124599\");");
    writefln("unbox!(long)(wx)=%8d", unbox!(long)(wx));
    writefln();

    wx = toNumeric("1.234");
    writefln("Box wx = toNumeric(\"1.234\");");
    writefln("unbox!(double)(wx)=%8.3f", unbox!(double)(wx));
    writefln();

    wx = toNumeric("1.234fi"); 
    writefln("Box wx = toNumeric(\"1.234fi\");");
    writefln("unbox!(ifloat)(wx)=%g", unbox!(ifloat)(wx));
    writefln();

    wx = toNumeric("1.234e+12");
    writefln("Box wx = toNumeric(\"1.234e+12\");");
    writefln("unbox!(float)(wx)=%8.4g", unbox!(float)(wx));
    writefln();

    wx = toNumeric("1.234e+12+345.4e+34i");
    writefln("Box wx = toNumeric(\"1.234e+12+345.4e+34i\");");
    writefln("unbox!(creal)(wx)=%g", unbox!(creal)(wx));
    writefln();
    
    wx = toNumeric("nan+nani");
    writefln("Box wx = toNumeric(\"nan+nani\");");
    writefln("unbox!(creal)(wx)=%g", unbox!(creal)(wx));
    writefln();
    
    writefln("Compiled and tested with: \"%s v%d.%03d\"", scomp.name, scomp.version_major, scomp.version_minor);
    writefln("tonumeric.unittest finished!"c); 
    
    //writefln("cr=%g toString()=\"%s\"", 
    //         cfloat.re.nan+cfloat.im.nan, 
    //         std.string.toString(cfloat.re.nan+cfloat.im.nan));
      
    // cr=inf toString()="-nan" ? There isn't a negative NaN value!        
    //writefln("cr=%g toString()=\"%s\"", 
    //         cfloat.re.infinity+cfloat.im.infinity, 
    //         std.string.toString(cfloat.re.infinity+-float.im.infinity));
}
}

/+
Output:
----------
C:\dmd>dmd tonumeric.d isnumeric.d -release
C:\dmd\bin\..\..\dm\bin\link.exe tonumeric+isnumeric,,,user32+kernel32/noi;

C:\dmd>tonumeric
unbox!(uint)(wx)=123456
unbox!(long)(wx)= -124599
unbox!(double)(wx)=   1.234
unbox!(ifloat)(wx)=1.234
unbox!(float)(wx)=1.234e+12
unbox!(creal)(wx)=1.234e+12+3.454e+36i
unbox!(creal)(wx)=nan+0i

C:\dmd>
+/